theory KMP imports
  "../Coinductive_Languages/Coinductive_Language"
begin

primrec match where
  "match L [] = \<oo> L"
| "match L (x # xs) = (\<oo> L \<or> match (\<dd> L x) xs)"

lemma match_alt: "match L w = (\<exists>p s. w = p @ s \<and> in_language L p)"
  apply (induct w arbitrary: L)
  apply (auto simp: Cons_eq_append_conv intro: exI[of _ "[]"] exI[of _ "_ # _"])
  apply (metis in_language.simps(2))
  done

corec tab where
  "tab xs failure =
    Lang (xs = []) (\<lambda>c. if xs = [] \<or> hd xs \<noteq> c then failure c else tab (tl xs) (\<dd> (failure c)))"

lemma tab_simps [simp]:
  "tab [] failure = Lang True failure"
  "tab (x # xs) failure = Lang False (\<lambda>c. if x \<noteq> c then failure c else tab xs (\<dd> (failure x)))"
by(subst tab.code; simp add: fun_eq_iff; fail)+

lemma mk_table_sel [simp]:
  "\<oo> (tab xs failure) = (xs = [])"
  "\<dd> (tab xs failure) = (\<lambda>c. if xs = [] \<or> hd xs \<noteq> c then failure c else tab (tl xs) (\<dd> (failure c)))"
  by (subst tab.code; simp)+

definition aux :: "'a list \<Rightarrow> ('a \<Rightarrow> 'a language) \<Rightarrow> 'a \<Rightarrow> 'a language"
where "aux xs failure = \<dd> (tab xs failure)"

friend_of_corec aux
where "aux xs failure c = Lang
   (if xs = [] \<or> hd xs \<noteq> c then \<oo> (failure c) else tl xs = [])
   (if xs = [] \<or> hd xs \<noteq> c then \<dd> (failure c) else (aux (tl xs) (\<dd> (failure c))))"
subgoal unfolding aux_def mk_table_sel by(subst tab.code; simp)
apply (auto simp: rel_fun_def rel_prod.simps split: prod.split; metis fst_conv snd_conv)
done

context fixes xs :: "'a list" begin

corec table :: "'a language"
where "table = Lang (xs = []) (aux xs (\<lambda>_. table))"

lemma table: "table = tab xs (\<lambda>_. table)"
by(subst (1) table.code)(rule language.expand; simp add: aux_def)

end

definition is_substring_of :: "'a list \<Rightarrow> 'a list \<Rightarrow> bool"
where "is_substring_of xs ys = match (table xs) ys"

export_code is_substring_of in Haskell module_name KMP file code

end
