section \<open>Generative probabilistic values\<close>

theory Generative_Probabilistic_Value imports
  "~~/src/HOL/Probability/SPMF"
  "../../AFP/MFMC_Countable/Rel_PMF_Characterisation"
  PF_Set
  Resumption
  Generat
  "../../BNF_Corec"
begin

interpretation rel_spmf_characterisation
by(unfold_locales)(rule rel_pmf_measureI)

context includes lifting_syntax begin
lemma rel_fun_curry:
  "(A ===> B ===> C) f g \<longleftrightarrow> (rel_prod A B ===> C) (case_prod f) (case_prod g)"
by(auto simp add: rel_fun_def)
end

lemma tlength_eq_infinity_iff: "tlength xs = \<infinity> \<longleftrightarrow> \<not> tfinite xs"
including tllist.lifting by transfer(simp add: llength_eq_infty_conv_lfinite)

lemma transp_rel_fun: "\<lbrakk> is_equality Q; transp R \<rbrakk> \<Longrightarrow> transp (rel_fun Q R)"
by(rule transpI)(auto dest: transpD rel_funD simp add: is_equality_def)

lemma rel_fun_inf: "inf (rel_fun Q R) (rel_fun Q R') = rel_fun Q (inf R R')"
by(rule antisym)(auto elim: rel_fun_mono dest: rel_funD)

lemma eSuc_SUP: "A \<noteq> {} \<Longrightarrow> eSuc (SUPREMUM A f) = (SUP x:A. eSuc (f x))"
by(subst eSuc_Sup)(simp_all)

lemma (in ccpo) Sup_image_mono:
  assumes ccpo: "class.ccpo luba orda lessa"
  and mono: "monotone orda op \<le> f"
  and chain: "Complete_Partial_Order.chain orda A"
  and "A \<noteq> {}"
  shows "Sup (f ` A) \<le> (f (luba A))"
proof(rule ccpo_Sup_least)
  from chain show "Complete_Partial_Order.chain op \<le> (f ` A)"
    by(rule chain_imageI)(rule monotoneD[OF mono])
  fix x
  assume "x \<in> f ` A"
  then obtain y where "x = f y" "y \<in> A" by blast
  from \<open>y \<in> A\<close> have "orda y (luba A)" by(rule ccpo.ccpo_Sup_upper[OF ccpo chain])
  hence "f y \<le> f (luba A)" by(rule monotoneD[OF mono])
  thus "x \<le> f (luba A)" using \<open>x = f y\<close> by simp
qed

lemma (in ccpo) admissible_le_mono:
  assumes "monotone op \<le> op \<le> f"
  shows "ccpo.admissible Sup op \<le> (\<lambda>x. x \<le> f x)"
proof(rule ccpo.admissibleI)
  fix Y
  assume chain: "Complete_Partial_Order.chain op \<le> Y"
    and Y: "Y \<noteq> {}"
    and le [rule_format]: "\<forall>x\<in>Y. x \<le> f x"
  have "\<Squnion>Y \<le> \<Squnion>(f ` Y)" using chain
    by(rule ccpo_Sup_least)(rule order_trans[OF le]; blast intro!: ccpo_Sup_upper chain_imageI[OF chain] intro: monotoneD[OF assms])
  also have "\<dots> \<le> f (\<Squnion>Y)"
    by(rule Sup_image_mono[OF _ assms chain Y, where lessa="op <"]) unfold_locales
  finally show "\<Squnion>Y \<le> \<dots>" .
qed

lemma (in ccpo) fixp_induct_strong2:
  assumes adm: "ccpo.admissible Sup op \<le> P"
  and mono: "monotone op \<le> op \<le> f"
  and bot: "P (\<Squnion>{})"
  and step: "\<And>x. \<lbrakk> x \<le> ccpo_class.fixp f; x \<le> f x; P x \<rbrakk> \<Longrightarrow> P (f x)"
  shows "P (ccpo_class.fixp f)"
proof(rule fixp_strong_induct[where P="\<lambda>x. x \<le> f x \<and> P x", THEN conjunct2])
  show "ccpo.admissible Sup op \<le> (\<lambda>x. x \<le> f x \<and> P x)"
    using admissible_le_mono adm by(rule admissible_conj)(rule mono)
next
  show "\<Squnion>{} \<le> f (\<Squnion>{}) \<and> P (\<Squnion>{})"
    by(auto simp add: bot chain_empty intro: ccpo_Sup_least)
next
  fix x
  assume "x \<le> ccpo_class.fixp f" "x \<le> f x \<and> P x"
  thus "f x \<le> f (f x) \<and> P (f x)"
    by(auto dest: monotoneD[OF mono] intro: step)
qed(rule mono)

context partial_function_definitions begin

lemma fixp_induct_strong2_uc:
  fixes F :: "'c \<Rightarrow> 'c"
    and U :: "'c \<Rightarrow> 'b \<Rightarrow> 'a"
    and C :: "('b \<Rightarrow> 'a) \<Rightarrow> 'c"
    and P :: "('b \<Rightarrow> 'a) \<Rightarrow> bool"
  assumes mono: "\<And>x. mono_body (\<lambda>f. U (F (C f)) x)"
    and eq: "f \<equiv> C (fixp_fun (\<lambda>f. U (F (C f))))"
    and inverse: "\<And>f. U (C f) = f"
    and adm: "ccpo.admissible lub_fun le_fun P"
    and bot: "P (\<lambda>_. lub {})"
    and step: "\<And>f'. \<lbrakk> le_fun (U f') (U f); le_fun (U f') (U (F f')); P (U f') \<rbrakk> \<Longrightarrow> P (U (F f'))"
  shows "P (U f)"
unfolding eq inverse
apply (rule ccpo.fixp_induct_strong2[OF ccpo adm])
apply (insert mono, auto simp: monotone_def fun_ord_def bot fun_lub_def)[2]
apply (rule_tac f'5="C x" in step)
apply (simp_all add: inverse eq)
done

end


hide_const (open) Done

subsection {* Type definition *}

context notes [[quick_and_dirty]] begin (* TODO: remove when codatatype is fixed *)
codatatype (results'_gpv: 'a, outs'_gpv: 'out, 'in) gpv
  = GPV (the_gpv: "('a, 'out, 'in \<Rightarrow> ('a, 'out, 'in) gpv) generat spmf")
end

declare gpv.rel_eq [relator_eq]

text {* Reactive values are like generative, except that they take an input first. *}

type_synonym ('a, 'out, 'in) rpv = "'in \<Rightarrow> ('a, 'out, 'in) gpv"
print_translation -- \<open>pretty printing for @{typ "('a, 'out, 'in) rpv"}\<close> \<open>
  let
    fun tr' [in1, Const (@{type_syntax gpv}, _) $ a $ out $ in2] =
      if in1 = in2 then Syntax.const @{type_syntax rpv} $ a $ out $ in1
      else raise Match;
  in [(@{type_syntax "fun"}, K tr')]
  end
\<close>
typ "('a, 'out, 'in) rpv"
text \<open>
  Effectively, @{typ "('a, 'out, 'in) gpv"} and @{typ "('a, 'out, 'in) rpv"} are mutually recursive.
\<close>

lemma eq_GPV_iff: "f = GPV g \<longleftrightarrow> the_gpv f = g"
by(cases f) auto

declare gpv.set[simp del]

lemma rel_gpv_def':
  "rel_gpv A B gpv gpv' \<longleftrightarrow>
  (\<exists>gpv''. (\<forall>(x, y) \<in> results'_gpv gpv''. A x y) \<and> (\<forall>(x, y) \<in> outs'_gpv gpv''. B x y) \<and>
           map_gpv fst fst gpv'' = gpv \<and> map_gpv snd snd gpv'' = gpv')"
unfolding rel_gpv_def by(auto simp add: BNF_Def.Grp_def)

definition results'_rpv :: "('a, 'out, 'in) rpv \<Rightarrow> 'a set"
where "results'_rpv rpv = range rpv \<bind> results'_gpv"

definition outs'_rpv :: "('a, 'out, 'in) rpv \<Rightarrow> 'out set"
where "outs'_rpv rpv = range rpv \<bind> outs'_gpv"

abbreviation rel_rpv
  :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> ('out \<Rightarrow> 'out' \<Rightarrow> bool)
  \<Rightarrow> ('in \<Rightarrow> ('a, 'out, 'in) gpv) \<Rightarrow> ('in \<Rightarrow> ('b, 'out', 'in) gpv) \<Rightarrow> bool"
where "rel_rpv A B \<equiv> rel_fun op = (rel_gpv A B)"

lemma in_results'_rpv [iff]: "x \<in> results'_rpv rpv \<longleftrightarrow> (\<exists>input. x \<in> results'_gpv (rpv input))"
by(simp add: results'_rpv_def)

lemma in_outs_rpv [iff]: "out \<in> outs'_rpv rpv \<longleftrightarrow> (\<exists>input. out \<in> outs'_gpv (rpv input))"
by(simp add: outs'_rpv_def)

lemma results'_GPV [simp]:
  "results'_gpv (GPV r) =
   (set_spmf r \<bind> generat_pures) \<union> 
   ((set_spmf r \<bind> generat_conts) \<bind> results'_rpv)"
by(auto simp add: gpv.set bind_UNION set_spmf_def)

lemma outs'_GPV [simp]:
  "outs'_gpv (GPV r) =
   (set_spmf r \<bind> generat_outs) \<union> 
   ((set_spmf r \<bind> generat_conts) \<bind> outs'_rpv)"
by(auto simp add: gpv.set bind_UNION set_spmf_def)

lemma outs'_gpv_unfold:
  "outs'_gpv r =
   (set_spmf (the_gpv r) \<bind> generat_outs) \<union> 
   ((set_spmf (the_gpv r) \<bind> generat_conts) \<bind> outs'_rpv)"
by(cases r) simp

lemma outs'_gpv_induct [consumes 1, case_names Out Cont, induct set: outs'_gpv]:
  assumes x: "x \<in> outs'_gpv gpv"
  and Out: "\<And>generat gpv. \<lbrakk> generat \<in> set_spmf (the_gpv gpv); x \<in> generat_outs generat \<rbrakk> \<Longrightarrow> P gpv"
  and Cont: "\<And>generat gpv c input.
    \<lbrakk> generat \<in> set_spmf (the_gpv gpv); c \<in> generat_conts generat; x \<in> outs'_gpv (c input); P (c input) \<rbrakk> \<Longrightarrow> P gpv"
  shows "P gpv"
using x
apply(induction y\<equiv>"x" gpv)
 apply(rule Out, simp add: in_set_spmf, simp)
apply(erule imageE, rule Cont, simp add: in_set_spmf, simp, simp, simp)
.

lemma outs'_gpv_cases [consumes 1, case_names Out Cont, cases set: outs'_gpv]:
  assumes "x \<in> outs'_gpv gpv"
  obtains (Out) generat where "generat \<in> set_spmf (the_gpv gpv)" "x \<in> generat_outs generat"
    | (Cont) generat c input where "generat \<in> set_spmf (the_gpv gpv)" "c \<in> generat_conts generat" "x \<in> outs'_gpv (c input)"
using assms by cases(auto simp add: in_set_spmf)

lemma outs'_gpvI [intro?]:
  shows outs'_gpv_Out: "\<lbrakk> generat \<in> set_spmf (the_gpv gpv); x \<in> generat_outs generat \<rbrakk> \<Longrightarrow> x \<in> outs'_gpv gpv"
  and outs'_gpv_Cont: "\<lbrakk> generat \<in> set_spmf (the_gpv gpv); c \<in> generat_conts generat; x \<in> outs'_gpv (c input) \<rbrakk> \<Longrightarrow> x \<in> outs'_gpv gpv"
by(auto intro: gpv.set_sel simp add: in_set_spmf)

lemma results'_gpv_induct [consumes 1, case_names Pure Cont, induct set: results'_gpv]:
  assumes x: "x \<in> results'_gpv gpv"
  and Pure: "\<And>generat gpv. \<lbrakk> generat \<in> set_spmf (the_gpv gpv); x \<in> generat_pures generat \<rbrakk> \<Longrightarrow> P gpv"
  and Cont: "\<And>generat gpv c input.
    \<lbrakk> generat \<in> set_spmf (the_gpv gpv); c \<in> generat_conts generat; x \<in> results'_gpv (c input); P (c input) \<rbrakk> \<Longrightarrow> P gpv"
  shows "P gpv"
using x
apply(induction y\<equiv>"x" gpv)
 apply(rule Pure; simp add: in_set_spmf)
apply(erule imageE, rule Cont, simp add: in_set_spmf, simp, simp, simp)
.

lemma results'_gpv_cases [consumes 1, case_names Pure Cont, cases set: results'_gpv]:
  assumes "x \<in> results'_gpv gpv"
  obtains (Pure) generat where "generat \<in> set_spmf (the_gpv gpv)" "x \<in> generat_pures generat"
    | (Cont) generat c input where "generat \<in> set_spmf (the_gpv gpv)" "c \<in> generat_conts generat" "x \<in> results'_gpv (c input)"
using assms by cases(auto simp add: in_set_spmf)

lemma results'_gpvI [intro?]:
  shows results'_gpv_Pure: "\<lbrakk> generat \<in> set_spmf (the_gpv gpv); x \<in> generat_pures generat \<rbrakk> \<Longrightarrow> x \<in> results'_gpv gpv"
  and results'_gpv_Cont: "\<lbrakk> generat \<in> set_spmf (the_gpv gpv); c \<in> generat_conts generat; x \<in> results'_gpv (c input) \<rbrakk> \<Longrightarrow> x \<in> results'_gpv gpv"
by(auto intro: gpv.set_sel simp add: in_set_spmf)

lemma left_unique_rel_gpv [transfer_rule]:
  "\<lbrakk> left_unique A; left_unique B \<rbrakk> \<Longrightarrow> left_unique (rel_gpv A B)"
unfolding left_unique_alt_def gpv.rel_conversep[symmetric] gpv.rel_compp[symmetric]
by(subst gpv.rel_eq[symmetric])(rule gpv.rel_mono)

lemma right_unique_rel_gpv [transfer_rule]:
  "\<lbrakk> right_unique A; right_unique B \<rbrakk> \<Longrightarrow> right_unique (rel_gpv A B)"
unfolding right_unique_alt_def gpv.rel_conversep[symmetric] gpv.rel_compp[symmetric]
by(subst gpv.rel_eq[symmetric])(rule gpv.rel_mono)

lemma bi_unique_rel_gpv [transfer_rule]:
  "\<lbrakk> bi_unique A; bi_unique B \<rbrakk> \<Longrightarrow> bi_unique (rel_gpv A B)"
unfolding bi_unique_alt_def by(simp add: left_unique_rel_gpv right_unique_rel_gpv)

lemma left_total_rel_gpv [transfer_rule]:
  "\<lbrakk> left_total A; left_total B \<rbrakk> \<Longrightarrow> left_total (rel_gpv A B)"
unfolding left_total_alt_def gpv.rel_conversep[symmetric] gpv.rel_compp[symmetric]
by(subst gpv.rel_eq[symmetric])(rule gpv.rel_mono)

lemma right_total_rel_gpv [transfer_rule]:
  "\<lbrakk> right_total A; right_total B \<rbrakk> \<Longrightarrow> right_total (rel_gpv A B)"
unfolding right_total_alt_def gpv.rel_conversep[symmetric] gpv.rel_compp[symmetric]
by(subst gpv.rel_eq[symmetric])(rule gpv.rel_mono)

lemma bi_total_rel_gpv [transfer_rule]: "\<lbrakk> bi_total A; bi_total B \<rbrakk> \<Longrightarrow> bi_total (rel_gpv A B)"
unfolding bi_total_alt_def by(simp add: left_total_rel_gpv right_total_rel_gpv)

declare gpv.map_transfer[transfer_rule]

lemma if_distrib_map_gpv [if_distribs]:
  "map_gpv f g (if b then gpv else gpv') = (if b then map_gpv f g gpv else map_gpv f g gpv')"
by simp

subsection {* Simple, derived operations *}

primcorec Done :: "'a \<Rightarrow> ('a, 'out, 'in) gpv"
where "the_gpv (Done a) = return_spmf (Pure a)"

primcorec Pause :: "'out \<Rightarrow> ('in \<Rightarrow> ('a, 'out, 'in) gpv) \<Rightarrow> ('a, 'out, 'in) gpv"
where "the_gpv (Pause out c) = return_spmf (IO out c)"

primcorec lift_spmf :: "'a spmf \<Rightarrow> ('a, 'out, 'in) gpv"
where "the_gpv (lift_spmf p) = map_spmf Pure p"

definition Fail :: "('a, 'out, 'in) gpv"
where "Fail = GPV (return_pmf None)"

definition React :: "('in \<Rightarrow> 'out \<times> ('a, 'out, 'in) rpv) \<Rightarrow> ('a, 'out, 'in) rpv"
where "React f input = case_prod Pause (f input)"

definition rFail :: "('a, 'out, 'in) rpv"
where "rFail = (\<lambda>_. Fail)"

lemma Done_inject [simp]: "Done x = Done y \<longleftrightarrow> x = y"
by(simp add: Done.ctr)

lemma Pause_inject [simp]: "Pause out c = Pause out' c' \<longleftrightarrow> out = out' \<and> c = c'"
by(simp add: Pause.ctr)

lemma [simp]:
  shows Done_neq_Pause: "Done x \<noteq> Pause out c"
  and Pause_neq_Done: "Pause out c \<noteq> Done x"
by(simp_all add: Done.ctr Pause.ctr)

lemma outs'_gpv_Done [simp]: "outs'_gpv (Done x) = {}"
by(auto elim: outs'_gpv_cases)

lemma results'_gpv_Done [simp]: "results'_gpv (Done x) = {x}"
by(auto intro: results'_gpvI elim: results'_gpv_cases)

lemma outs'_gpv_Pause [simp]: "outs'_gpv (Pause out c) = insert out (\<Union>input. outs'_gpv (c input))"
by(auto 4 4 intro: outs'_gpvI elim: outs'_gpv_cases)

lemma results'_gpv_Pause [simp]: "results'_gpv (Pause out rpv) = results'_rpv rpv"
by(auto 4 4 intro: results'_gpvI elim: results'_gpv_cases)

lemma lift_spmf_return [simp]: "lift_spmf (return_spmf x) = Done x"
by(simp add: lift_spmf.ctr Done.ctr)

lemma lift_spmf_None [simp]: "lift_spmf (return_pmf None) = Fail"
by(rule gpv.expand)(simp add: Fail_def)

lemma the_gpv_lift_spmf [simp]: "the_gpv (lift_spmf r) = map_spmf Pure r"
by(simp)

lemma outs'_gpv_lift_spmf [simp]: "outs'_gpv (lift_spmf p) = {}"
by(auto 4 3 elim: outs'_gpv_cases)

lemma results'_gpv_lift_spmf [simp]: "results'_gpv (lift_spmf p) = set_spmf p"
by(auto 4 3 elim: results'_gpv_cases intro: results'_gpvI)

lemma lift_spmf_inject [simp]: "lift_spmf p = lift_spmf q \<longleftrightarrow> p = q"
by(auto simp add: lift_spmf.code dest!: pmf.inj_map_strong[rotated] option.inj_map_strong[rotated])

lemma map_lift_spmf: "map_gpv f g (lift_spmf p) = lift_spmf (map_spmf f p)"
by(rule gpv.expand)(simp add: gpv.map_sel spmf.map_comp o_def)

lemma [simp]:
  shows Fail_neq_Pause: "Fail \<noteq> Pause out c"
  and Pause_neq_Fail: "Pause out c \<noteq> Fail"
  and Fail_neq_Done: "Fail \<noteq> Done x"
  and Done_neq_Fail: "Done x \<noteq> Fail"
by(simp_all add: Fail_def Pause.ctr Done.ctr)

text {* Add @{typ unit} closure to circumvent SML value restriction *}

definition Fail' :: "unit \<Rightarrow> ('a, 'out, 'in) gpv"
where [code del]: "Fail' _ = Fail"

lemma Fail_code [code_unfold]: "Fail = Fail' ()"
by(simp add: Fail'_def)

lemma Fail'_code [code]:
  "Fail' x = GPV (return_pmf None)"
by(simp add: Fail'_def Fail_def)

lemma Fail_sel [simp]:
  "the_gpv Fail = return_pmf None"
by(simp add: Fail_def)

lemma Fail_eq_GPV_iff [simp]: "Fail = GPV f \<longleftrightarrow> f = return_pmf None"
by(auto simp add: Fail_def)

lemma outs'_gpv_Fail [simp]: "outs'_gpv Fail = {}"
by(auto elim: outs'_gpv_cases)

lemma results'_gpv_Fail [simp]: "results'_gpv Fail = {}"
by(auto elim: results'_gpv_cases)

lemma React_inject [iff]: "React f = React f' \<longleftrightarrow> f = f'"
by(auto simp add: React_def fun_eq_iff split_def intro: prod.expand)

lemma React_apply [simp]: "f input = (out, c) \<Longrightarrow> React f input = Pause out c"
by(simp add: React_def)

lemma rFail_apply [simp]: "rFail input = Fail"
by(simp add: rFail_def)

lemma [simp]:
  shows rFail_neq_React: "rFail \<noteq> React f"
  and React_neq_rFail: "React f \<noteq> rFail"
by(simp_all add: React_def fun_eq_iff split_beta)

lemma rel_gpv_FailI [simp]: "rel_gpv A C Fail Fail"
by(subst gpv.rel_sel) simp


lemma rel_gpv_Done [iff]: "rel_gpv A C (Done x) (Done y) \<longleftrightarrow> A x y"
by(subst gpv.rel_sel) simp

lemma rel_gpv_Pause [iff]: "rel_gpv A C (Pause out c) (Pause out' c') \<longleftrightarrow> C out out' \<and> (\<forall>x. rel_gpv A C (c x) (c' x))"
by(subst gpv.rel_sel)(simp add: rel_fun_def)

lemma rel_gpv_lift_spmf [iff]: "rel_gpv A C (lift_spmf p) (lift_spmf q) \<longleftrightarrow> rel_spmf A p q"
by(subst gpv.rel_sel)(simp add: spmf_rel_map)

context includes lifting_syntax begin
lemma Fail_parametric [transfer_rule]: "rel_gpv A C Fail Fail"
by(fact rel_gpv_FailI)

lemma Done_parametric [transfer_rule]: "(A ===> rel_gpv A C) Done Done"
by(rule rel_funI) simp

lemma Pause_parametric [transfer_rule]:
  "(C ===> (op = ===> rel_gpv A C) ===> rel_gpv A C) Pause Pause"
by(simp add: rel_fun_def)

lemma lift_spmf_parametric [transfer_rule]:
  "(rel_spmf A ===> rel_gpv A C) lift_spmf lift_spmf"
by(simp add: rel_fun_def)
end

lemma map_gpv_Done [simp]: "map_gpv f g (Done x) = Done (f x)"
by(simp add: Done.code)

lemma map_gpv_Pause [simp]: "map_gpv f g (Pause x c) = Pause (g x) (map_gpv f g \<circ> c)"
by(simp add: Pause.code)

lemma map_gpv_Fail [simp]: "map_gpv f g Fail = Fail"
by(simp add: Fail_def)

subsection {* Monad structure *}

primcorec bind_gpv :: "('a, 'out, 'in) gpv \<Rightarrow> ('a \<Rightarrow> ('b, 'out, 'in) gpv) \<Rightarrow> ('b, 'out, 'in) gpv"
where
  "the_gpv (bind_gpv r f) =
   map_spmf (map_generat id id (op \<circ> (case_sum id (\<lambda>r. bind_gpv r f))))
     (the_gpv r \<bind>
      (case_generat
        (\<lambda>x. map_spmf (map_generat id id (op \<circ> Inl)) (the_gpv (f x)))
        (\<lambda>out c. return_spmf (IO out (\<lambda>input. Inr (c input))))))"

declare bind_gpv.sel [simp del]

adhoc_overloading Monad_Syntax.bind bind_gpv

lemma bind_gpv_unfold [code]:
  "r \<bind> f = GPV (
   do {
     generat \<leftarrow> the_gpv r;
     case generat of Pure x \<Rightarrow> the_gpv (f x)
       | IO out c \<Rightarrow> return_spmf (IO out (\<lambda>input. c input \<bind> f))
   })"
unfolding bind_gpv_def
apply(rule gpv.expand)
apply(simp add: map_spmf_bind_spmf)
apply(rule arg_cong[where f="bind_spmf (the_gpv r)"])
apply(auto split: generat.split simp add: map_spmf_bind_spmf fun_eq_iff spmf.map_comp o_def generat.map_comp id_def[symmetric] generat.map_id pmf.map_id option.map_id)
done

lemma bind_gpv_code_cong: "f = f' \<Longrightarrow> bind_gpv f g = bind_gpv f' g" by simp
setup {* Code_Simp.map_ss (Simplifier.add_cong @{thm bind_gpv_code_cong}) *}

lemma bind_gpv_sel:
  "the_gpv (r \<bind> f) =
   do {
     generat \<leftarrow> the_gpv r;
     case generat of Pure x \<Rightarrow> the_gpv (f x)
       | IO out c \<Rightarrow> return_spmf (IO out (\<lambda>input. bind_gpv (c input) f))
   }"
by(subst bind_gpv_unfold) simp

lemma bind_gpv_sel' [simp]:
  "the_gpv (r \<bind> f) =
   do {
     generat \<leftarrow> the_gpv r;
     if is_Pure generat then the_gpv (f (result generat))
     else return_spmf (IO (output generat) (\<lambda>input. bind_gpv (continuation generat input) f))
   }"
unfolding bind_gpv_sel
by(rule arg_cong[where f="bind_spmf (the_gpv r)"])(simp add: fun_eq_iff split: generat.split)

lemma Done_bind_gpv [simp]: "Done a \<bind> f = f a"
by(rule gpv.expand)(simp)

lemma bind_gpv_Done [simp]: "f \<bind> Done = f"
proof(coinduction arbitrary: f rule: gpv.coinduct)
  case (Eq_gpv f)
  have *: "the_gpv f \<bind> (case_generat (\<lambda>x. return_spmf (Pure x)) (\<lambda>out c. return_spmf (IO out (\<lambda>input. Inr (c input))))) =
           map_spmf (map_generat id id (op \<circ> Inr)) (bind_spmf (the_gpv f) return_spmf)"
    unfolding map_spmf_bind_spmf
    by(rule arg_cong2[where f=bind_spmf])(auto simp add: fun_eq_iff split: generat.split)
  show ?case
    by(auto simp add: * bind_gpv.simps pmf.rel_map option.rel_map[abs_def] generat.rel_map[abs_def] simp del: bind_gpv_sel' intro!: rel_generatI rel_spmf_reflI)
qed

lemma if_distrib_bind_gpv2 [if_distribs]:
  "bind_gpv gpv (\<lambda>y. if b then f y else g y) = (if b then bind_gpv gpv f else bind_gpv gpv g)"
by simp

lemma lift_spmf_bind: "lift_spmf r \<bind> f = GPV (r \<bind> the_gpv \<circ> f)"
by(coinduction arbitrary: r f rule: gpv.coinduct_strong)(auto simp add: bind_map_spmf o_def intro: rel_pmf_reflI rel_optionI rel_generatI)

lemma the_gpv_bind_gpv_lift_spmf [simp]:
  "the_gpv (bind_gpv (lift_spmf p) f) = bind_spmf p (the_gpv \<circ> f)"
by(simp add: bind_map_spmf o_def)

lemma lift_spmf_bind_spmf: "lift_spmf (p \<bind> f) = lift_spmf p \<bind> (\<lambda>x. lift_spmf (f x))"
by(rule gpv.expand)(simp add: lift_spmf_bind o_def map_spmf_bind_spmf)

lemma GPV_bind:
  "GPV f \<bind> g = 
   GPV (f \<bind> (\<lambda>generat. case generat of Pure x \<Rightarrow> the_gpv (g x) | IO out c \<Rightarrow> return_spmf (IO out (\<lambda>input. c input \<bind> g))))"
by(subst bind_gpv_unfold) simp

lemma GPV_bind':
  "GPV f \<bind> g = GPV (f \<bind> (\<lambda>generat. if is_Pure generat then the_gpv (g (result generat)) else return_spmf (IO (output generat) (\<lambda>input. continuation generat input \<bind> g))))"
unfolding GPV_bind gpv.inject
by(rule arg_cong[where f="bind_spmf f"])(simp add: fun_eq_iff split: generat.split)

lemma bind_gpv_assoc:
  fixes f :: "('a, 'out, 'in) gpv"
  shows "(f \<bind> g) \<bind> h = f \<bind> (\<lambda>x. g x \<bind> h)"
proof(coinduction arbitrary: f g h rule: gpv.coinduct_strong)
  case (Eq_gpv f g h)
  show ?case
    apply(simp cong del: if_weak_cong)
    apply(rule rel_spmf_bindI[where R="op ="])
     apply(simp add: option.rel_eq pmf.rel_eq)
    apply(fastforce intro: rel_pmf_return_pmfI rel_generatI rel_spmf_reflI)
    done
qed

lemma map_gpv_bind_gpv: "map_gpv f g (bind_gpv gpv h) = bind_gpv (map_gpv id g gpv) (\<lambda>x. map_gpv f g (h x))"
apply(coinduction arbitrary: gpv rule: gpv.coinduct_strong)
apply(simp add: bind_gpv.sel gpv.map_sel spmf_rel_map generat.rel_map o_def bind_map_spmf del: bind_gpv_sel')
apply(rule rel_spmf_bind_reflI)
apply(auto simp add: spmf_rel_map generat.rel_map split: generat.split intro!: rel_spmf_reflI generat.rel_refl rel_funI)
done

lemma map_gpv_id_bind_gpv: "map_gpv f id (bind_gpv gpv g) = bind_gpv gpv (map_gpv f id \<circ> g)"
by(simp add: map_gpv_bind_gpv gpv.map_id o_def)

lemma map_gpv_conv_bind:
  "map_gpv f (\<lambda>x. x) x = bind_gpv x (\<lambda>x. Done (f x))"
using map_gpv_bind_gpv[of f "\<lambda>x. x" x Done] by(simp add: id_def[symmetric] gpv.map_id)

lemma bind_map_gpv: "bind_gpv (map_gpv f id gpv) g = bind_gpv gpv (g \<circ> f)"
by(simp add: map_gpv_conv_bind id_def bind_gpv_assoc o_def)

lemma bind_gpv_Fail [simp]: "Fail \<bind> f = Fail"
by(subst bind_gpv_unfold)(simp add: Fail_def)

context includes lifting_syntax begin

lemma bind_gpv_parametric [transfer_rule]:
  "(rel_gpv A C ===> (A ===> rel_gpv B C) ===> rel_gpv B C) bind_gpv bind_gpv"
unfolding bind_gpv_def by transfer_prover

end

lemma rel_gpv_bindI:
  "\<lbrakk> rel_gpv A C gpv gpv'; \<And>x y. A x y \<Longrightarrow> rel_gpv B C (f x) (g y) \<rbrakk>
  \<Longrightarrow> rel_gpv B C (bind_gpv gpv f) (bind_gpv gpv' g)"
by(fact bind_gpv_parametric[THEN rel_funD, THEN rel_funD, OF _ rel_funI])

lemma bind_gpv_cong:
  "\<lbrakk> gpv = gpv'; \<And>x. x \<in> results'_gpv gpv' \<Longrightarrow> f x = g x \<rbrakk> \<Longrightarrow> bind_gpv gpv f = bind_gpv gpv' g"
apply(subst gpv.rel_eq[symmetric])
apply(rule rel_gpv_bindI[where A="eq_onp (\<lambda>x. x \<in> results'_gpv gpv')"])
 apply(subst (asm) gpv.rel_eq[symmetric])
 apply(erule gpv.rel_mono_strong)
  apply(simp add: eq_onp_def)
 apply simp
apply(clarsimp simp add: gpv.rel_eq eq_onp_def)
done

definition bind_rpv :: "('a, 'in, 'out) rpv \<Rightarrow> ('a \<Rightarrow> ('b, 'in, 'out) gpv) \<Rightarrow> ('b, 'in, 'out) rpv"
where "bind_rpv rpv f = (\<lambda>input. bind_gpv (rpv input) f)"

lemma bind_rpv_apply [simp]: "bind_rpv rpv f input = bind_gpv (rpv input) f"
by(simp add: bind_rpv_def fun_eq_iff)

adhoc_overloading Monad_Syntax.bind bind_rpv

lemma bind_rpv_code_cong: "rpv = rpv' \<Longrightarrow> bind_rpv rpv f = bind_rpv rpv' f" by simp
setup {* Code_Simp.map_ss (Simplifier.add_cong @{thm bind_rpv_code_cong}) *}

lemma bind_rpv_rDone [simp]: "bind_rpv rpv Done = rpv"
by(simp add: bind_rpv_def)

lemma bind_gpv_Pause [simp]: "bind_gpv (Pause out rpv) f = Pause out (bind_rpv rpv f)"
by(rule gpv.expand)(simp add: fun_eq_iff)

lemma bind_rpv_React [simp]: "bind_rpv (React f) g = React (apsnd (\<lambda>rpv. bind_rpv rpv g) \<circ> f)"
by(simp add: React_def split_beta fun_eq_iff)

lemma bind_rpv_assoc: "bind_rpv (bind_rpv rpv f) g = bind_rpv rpv ((\<lambda>gpv. bind_gpv gpv g) \<circ> f)"
by(simp add: fun_eq_iff bind_gpv_assoc o_def)

lemma bind_rpv_Done [simp]: "bind_rpv Done f = f"
by(simp add: bind_rpv_def)

subsection \<open> Embedding @{typ "'a spmf"} as a monad \<close>

context includes lifting_syntax begin
lemma neg_fun_distr3:
assumes 1: "left_unique R" "right_total R"
assumes 2: "right_unique S" "left_total S"
shows "(R OO R' ===> S OO S') \<le> ((R ===> S) OO (R' ===> S'))"
  using functional_relation[OF 2] functional_converse_relation[OF 1]
  unfolding rel_fun_def OO_def
  apply clarify
  apply (subst all_comm)
  apply (subst all_conj_distrib[symmetric])
  apply (intro choice)
  by metis
end

locale spmf_to_gpv begin

text \<open>
  The lifting package cannot handle free term variables in the merging of transfer rules,
  so for the embedding we define a specialised relator @{text "rel_gpv'"}
  which acts only on the returned values.
\<close>

definition rel_gpv' :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> ('a, 'out, 'in) gpv \<Rightarrow> ('b, 'out, 'in) gpv \<Rightarrow> bool"
where "rel_gpv' A = rel_gpv A op ="

lemma rel_gpv'_eq [relator_eq]: "rel_gpv' op = = op ="
unfolding rel_gpv'_def gpv.rel_eq ..

lemma rel_gpv'_mono [relator_mono]: "A \<le> B \<Longrightarrow> rel_gpv' A \<le> rel_gpv' B"
unfolding rel_gpv'_def by(rule gpv.rel_mono; simp)

lemma rel_gpv'_distr [relator_distr]: "rel_gpv' A OO rel_gpv' B = rel_gpv' (A OO B)"
unfolding rel_gpv'_def by (metis OO_eq gpv.rel_compp) 

lemma left_unique_rel_gpv' [transfer_rule]: "left_unique A \<Longrightarrow> left_unique (rel_gpv' A)"
unfolding rel_gpv'_def by(simp add: left_unique_rel_gpv left_unique_eq)

lemma right_unique_rel_gpv' [transfer_rule]: "right_unique A \<Longrightarrow> right_unique (rel_gpv' A)"
unfolding rel_gpv'_def by(simp add: right_unique_rel_gpv right_unique_eq)

lemma bi_unique_rel_gpv' [transfer_rule]: "bi_unique A \<Longrightarrow> bi_unique (rel_gpv' A)"
unfolding rel_gpv'_def by(simp add: bi_unique_rel_gpv bi_unique_eq)

lemma left_total_rel_gpv' [transfer_rule]: "left_total A \<Longrightarrow> left_total (rel_gpv' A)"
unfolding rel_gpv'_def by(simp add: left_total_rel_gpv left_total_eq)

lemma right_total_rel_gpv' [transfer_rule]: "right_total A \<Longrightarrow> right_total (rel_gpv' A)"
unfolding rel_gpv'_def by(simp add: right_total_rel_gpv right_total_eq)

lemma bi_total_rel_gpv' [transfer_rule]: "bi_total A \<Longrightarrow> bi_total (rel_gpv' A)"
unfolding rel_gpv'_def by(simp add: bi_total_rel_gpv bi_total_eq)


text \<open>
  We cannot use @{text setup_lifting} because @{typ "('a, 'out, 'in) gpv"} contains
  type variables which do not appear in @{typ "'a spmf"}.
\<close>

definition cr_spmf_gpv :: "'a spmf \<Rightarrow> ('a, 'out, 'in) gpv \<Rightarrow> bool"
where "cr_spmf_gpv p gpv \<longleftrightarrow> gpv = lift_spmf p"

definition spmf_of_gpv :: "('a, 'out, 'in) gpv \<Rightarrow> 'a spmf"
where "spmf_of_gpv gpv = (THE p. gpv = lift_spmf p)"

lemma spmf_of_gpv_lift_spmf [simp]: "spmf_of_gpv (lift_spmf p) = p"
unfolding spmf_of_gpv_def by auto

lemma rel_spmf_setD2:
  "\<lbrakk> rel_spmf A p q; y \<in> set_spmf q \<rbrakk> \<Longrightarrow> \<exists>x\<in>set_spmf p. A x y"
by(erule rel_spmfE) force

lemma rel_gpv_lift_spmf1: "rel_gpv A B (lift_spmf p) gpv \<longleftrightarrow> (\<exists>q. gpv = lift_spmf q \<and> rel_spmf A p q)"
apply(subst gpv.rel_sel)
apply(simp add: spmf_rel_map rel_generat_Pure1)
apply safe
 apply(rule exI[where x="map_spmf result (the_gpv gpv)"])
 apply(clarsimp simp add: spmf_rel_map)
 apply(rule conjI)
  apply(rule gpv.expand)
  apply(simp add: spmf.map_comp)
  apply(subst map_spmf_cong[OF refl, where g=id])
   apply(drule (1) rel_spmf_setD2)
   apply clarsimp
  apply simp
 apply(erule rel_spmf_mono)
 apply clarsimp
apply(clarsimp simp add: spmf_rel_map)
done

lemma rel_gpv_lift_spmf2: "rel_gpv A B gpv (lift_spmf q) \<longleftrightarrow> (\<exists>p. gpv = lift_spmf p \<and> rel_spmf A p q)"
by(subst gpv.rel_flip[symmetric])(simp add: rel_gpv_lift_spmf1 pmf.rel_flip option.rel_conversep)

definition pcr_spmf_gpv :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> 'a spmf \<Rightarrow> ('b, 'out, 'in) gpv \<Rightarrow> bool"
where "pcr_spmf_gpv A = cr_spmf_gpv OO rel_gpv A op ="

lemma pcr_cr_eq_spmf_gpv: "pcr_spmf_gpv op = = cr_spmf_gpv"
by(simp add: pcr_spmf_gpv_def gpv.rel_eq OO_eq)

lemma left_unique_cr_spmf_gpv: "left_unique cr_spmf_gpv"
by(rule left_uniqueI)(simp add: cr_spmf_gpv_def)

lemma left_unique_pcr_spmf_gpv [transfer_rule]:
  "left_unique A \<Longrightarrow> left_unique (pcr_spmf_gpv A)"
unfolding pcr_spmf_gpv_def by(intro left_unique_OO left_unique_cr_spmf_gpv left_unique_rel_gpv left_unique_eq)

lemma right_unique_cr_spmf_gpv: "right_unique cr_spmf_gpv"
by(rule right_uniqueI)(simp add: cr_spmf_gpv_def)

lemma right_unique_pcr_spmf_gpv [transfer_rule]:
  "right_unique A \<Longrightarrow> right_unique (pcr_spmf_gpv A)"
unfolding pcr_spmf_gpv_def by(intro right_unique_OO right_unique_cr_spmf_gpv right_unique_rel_gpv right_unique_eq)

lemma bi_unique_cr_spmf_gpv: "bi_unique cr_spmf_gpv"
by(simp add: bi_unique_alt_def left_unique_cr_spmf_gpv right_unique_cr_spmf_gpv)

lemma bi_unique_pcr_spmf_gpv [transfer_rule]: "bi_unique A \<Longrightarrow> bi_unique (pcr_spmf_gpv A)"
by(simp add: bi_unique_alt_def left_unique_pcr_spmf_gpv right_unique_pcr_spmf_gpv)

lemma left_total_cr_spmf_gpv: "left_total cr_spmf_gpv"
by(rule left_totalI)(simp add: cr_spmf_gpv_def)

lemma left_total_pcr_spmf_gpv [transfer_rule]: "left_total A ==> left_total (pcr_spmf_gpv A)"
unfolding pcr_spmf_gpv_def by(intro left_total_OO left_total_cr_spmf_gpv left_total_rel_gpv left_total_eq)

unbundle lifting_syntax

lemma return_spmf_gpv_transfer':
  "(op = ===> cr_spmf_gpv) return_spmf Done"
by(rule rel_funI)(simp add: cr_spmf_gpv_def)

lemma return_spmf_gpv_transfer [transfer_rule]:
  "(A ===> pcr_spmf_gpv A) return_spmf Done"
unfolding pcr_spmf_gpv_def
apply(rewrite in "(\<hole> ===> _) _ _" eq_OO[symmetric])
apply(rule pos_fun_distr[THEN le_funD, THEN le_funD, THEN le_boolD, THEN mp])
apply(rule relcomppI)
 apply(rule return_spmf_gpv_transfer')
apply transfer_prover
done

lemma bind_spmf_gpv_transfer':
  "(cr_spmf_gpv ===> (op = ===> cr_spmf_gpv) ===> cr_spmf_gpv) bind_spmf bind_gpv"
apply(clarsimp simp add: rel_fun_def cr_spmf_gpv_def)
apply(rule gpv.expand)
apply(simp add: bind_map_spmf map_spmf_bind_spmf o_def)
done

lemma bind_spmf_gpv_transfer [transfer_rule]:
  "(pcr_spmf_gpv A ===> (A ===> pcr_spmf_gpv B) ===> pcr_spmf_gpv B) bind_spmf bind_gpv"
unfolding pcr_spmf_gpv_def
apply(rewrite in "(_ ===> (\<hole> ===> _) ===> _) _ _" eq_OO[symmetric])
apply(rule fun_mono[THEN le_funD, THEN le_funD, THEN le_boolD, THEN mp])
  apply(rule order.refl)
 apply(rule fun_mono)
  apply(rule neg_fun_distr3[OF left_unique_eq right_total_eq right_unique_cr_spmf_gpv left_total_cr_spmf_gpv])
 apply(rule order.refl)
apply(rule fun_mono[THEN le_funD, THEN le_funD, THEN le_boolD, THEN mp])
  apply(rule order.refl)
 apply(rule pos_fun_distr)
apply(rule pos_fun_distr[THEN le_funD, THEN le_funD, THEN le_boolD, THEN mp])
apply(rule relcomppI)
 apply(rule bind_spmf_gpv_transfer')
apply transfer_prover
done

lemma lift_spmf_gpv_transfer':
  "(op = ===> cr_spmf_gpv) (\<lambda>x. x) lift_spmf"
by(simp add: rel_fun_def cr_spmf_gpv_def)

lemma lift_spmf_gpv_transfer [transfer_rule]:
  "(rel_spmf A ===> pcr_spmf_gpv A) (\<lambda>x. x) lift_spmf"
unfolding pcr_spmf_gpv_def
apply(rewrite in "(\<hole> ===> _) _ _" eq_OO[symmetric])
apply(rule pos_fun_distr[THEN le_funD, THEN le_funD, THEN le_boolD, THEN mp])
apply(rule relcomppI)
 apply(rule lift_spmf_gpv_transfer')
apply transfer_prover
done

lemma fail_spmf_gpv_transfer': "cr_spmf_gpv (return_pmf None) Fail"
by(simp add: cr_spmf_gpv_def)

lemma fail_spmf_gpv_transfer [transfer_rule]: "pcr_spmf_gpv A (return_pmf None) Fail"
unfolding pcr_spmf_gpv_def
apply(rule relcomppI)
 apply(rule fail_spmf_gpv_transfer')
apply transfer_prover
done

lemma map_spmf_gpv_transfer':
  "(op = ===> R ===> cr_spmf_gpv ===> cr_spmf_gpv) (\<lambda>f g. map_spmf f) map_gpv"
by(simp add: rel_fun_def cr_spmf_gpv_def map_lift_spmf)

lemma map_spmf_gpv_transfer [transfer_rule]:
  "((A ===> B) ===> R ===> pcr_spmf_gpv A ===> pcr_spmf_gpv B) (\<lambda>f g. map_spmf f) map_gpv"
unfolding pcr_spmf_gpv_def
apply(rewrite in "((\<hole> ===> _) ===> _)  _ _" eq_OO[symmetric])
apply(rewrite in "((_ ===> \<hole>) ===> _)  _ _" eq_OO[symmetric])
apply(rewrite in "(_ ===> \<hole> ===> _)  _ _" OO_eq[symmetric])
apply(rule fun_mono[THEN le_funD, THEN le_funD, THEN le_boolD, THEN mp])
  apply(rule neg_fun_distr3[OF left_unique_eq right_total_eq right_unique_eq left_total_eq])
 apply(rule fun_mono[OF order.refl])
 apply(rule pos_fun_distr)
apply(rule fun_mono[THEN le_funD, THEN le_funD, THEN le_boolD, THEN mp])
  apply(rule order.refl)
 apply(rule pos_fun_distr)
apply(rule pos_fun_distr[THEN le_funD, THEN le_funD, THEN le_boolD, THEN mp])
apply(rule relcomppI)
 apply(unfold rel_fun_eq)
 apply(rule map_spmf_gpv_transfer')
apply(unfold rel_fun_eq[symmetric])
apply transfer_prover
done

end

subsection {* Embedding resumptions *}

primcorec lift_resumption :: "('a, 'out, 'in) resumption \<Rightarrow> ('a, 'out, 'in) gpv"
where
  "the_gpv (lift_resumption r) = 
  (case r of resumption.Done None \<Rightarrow> return_pmf None
    | resumption.Done (Some x') => return_spmf (Pure x')
    | resumption.Pause out c => map_spmf (map_generat id id (op \<circ> lift_resumption)) (return_spmf (IO out c)))"

lemma the_gpv_lift_resumption:
  "the_gpv (lift_resumption r) = 
   (if is_Done r then if Option.is_none (resumption.result r) then return_pmf None else return_spmf (Pure (the (resumption.result r)))
    else return_spmf (IO (resumption.output r) (lift_resumption \<circ> resume r)))"
by(simp split: option.split resumption.split)

declare lift_resumption.simps [simp del]

lemma lift_resumption_Done [code]:
  "lift_resumption (resumption.Done x) = (case x of None \<Rightarrow> Fail | Some x' \<Rightarrow> Done x')"
by(rule gpv.expand)(simp add: the_gpv_lift_resumption split: option.split)

lemma lift_resumption_DONE [simp]:
  "lift_resumption (DONE x) = Done x"
by(simp add: DONE_def lift_resumption_Done)

lemma lift_resumption_ABORT [simp]:
  "lift_resumption ABORT = Fail"
by(simp add: ABORT_def lift_resumption_Done)

lemma lift_resumption_Pause [simp, code]:
  "lift_resumption (resumption.Pause out c) = Pause out (lift_resumption \<circ> c)"
by(rule gpv.expand)(simp add: the_gpv_lift_resumption)

subsection {* Assertions *}

definition assert_gpv :: "bool \<Rightarrow> (unit, 'out, 'in) gpv"
where "assert_gpv b = (if b then Done () else Fail)"

lemma assert_gpv_simps [simp]:
  "assert_gpv True = Done ()"
  "assert_gpv False = Fail"
by(simp_all add: assert_gpv_def)

lemma [simp]:
  shows assert_gpv_eq_Done: "assert_gpv b = Done x \<longleftrightarrow> b"
  and Done_eq_assert_gpv: "Done x = assert_gpv b \<longleftrightarrow> b"
  and Pause_neq_assert_gpv: "Pause out rpv \<noteq> assert_gpv b"
  and assert_gpv_neq_Pause: "assert_gpv b \<noteq> Pause out rpv"
  and assert_gpv_eq_Fail: "assert_gpv b = Fail \<longleftrightarrow> \<not> b"
  and Fail_eq_assert_gpv: "Fail = assert_gpv b \<longleftrightarrow> \<not> b"
by(simp_all add: assert_gpv_def)

lemma assert_gpv_inject [simp]: "assert_gpv b = assert_gpv b' \<longleftrightarrow> b = b'"
by(simp add: assert_gpv_def)

lemma assert_gpv_sel [simp]:
  "the_gpv (assert_gpv b) = map_spmf Pure (assert_spmf b)"
by(simp add: assert_gpv_def)

lemma the_gpv_bind_assert [simp]:
  "the_gpv (bind_gpv (assert_gpv b) f) =
   bind_spmf (assert_spmf b) (the_gpv \<circ> f)"
by(cases b) simp_all

primcorec try_gpv :: "('a, 'call, 'ret) gpv \<Rightarrow> ('a, 'call, 'ret) gpv \<Rightarrow> ('a, 'call, 'ret) gpv" ("TRY _ ELSE _" [0,60] 59)
where
  "the_gpv (TRY gpv ELSE gpv') = 
   map_spmf (map_generat id id (\<lambda>c input. case c input of Inl gpv \<Rightarrow> try_gpv gpv gpv' | Inr gpv' \<Rightarrow> gpv'))
     (try_spmf (map_spmf (map_generat id id (map_fun id Inl)) (the_gpv gpv))
               (map_spmf (map_generat id id (map_fun id Inr)) (the_gpv gpv')))"

lemma try_gpv_sel:
  "the_gpv (TRY gpv ELSE gpv') =
   TRY map_spmf (map_generat id id (\<lambda>c input. TRY c input ELSE gpv')) (the_gpv gpv) ELSE the_gpv gpv'"
by(simp add: try_gpv_def map_try_spmf spmf.map_comp o_def generat.map_comp generat.map_ident id_def)

lemma try_gpv_Done [simp]: "TRY Done x ELSE gpv' = Done x"
by(rule gpv.expand)(simp)

lemma try_gpv_Fail [simp]: "TRY Fail ELSE gpv' = gpv'"
by(rule gpv.expand)(simp add: spmf.map_comp o_def generat.map_comp generat.map_ident)

lemma try_gpv_Pause [simp]: "TRY Pause out c ELSE gpv' = Pause out (\<lambda>input. TRY c input ELSE gpv')"
by(rule gpv.expand) simp

lemma try_gpv_Fail2 [simp]: "TRY gpv ELSE Fail = gpv"
by(coinduction arbitrary: gpv rule: gpv.coinduct_strong)
  (auto simp add: spmf_rel_map generat.rel_map intro!: rel_spmf_reflI generat.rel_refl)

lemma lift_try_spmf: "lift_spmf (TRY p ELSE q) = TRY lift_spmf p ELSE lift_spmf q" 
by(rule gpv.expand)(simp add: map_try_spmf spmf.map_comp o_def)

lemma try_assert_gpv: "TRY assert_gpv b ELSE gpv' = (if b then Done () else gpv')"
by(simp)

context includes lifting_syntax begin
lemma try_gpv_parametric [transfer_rule]:
  "(rel_gpv A C ===> rel_gpv A C ===> rel_gpv A C) try_gpv try_gpv"
unfolding try_gpv_def by transfer_prover
end

lemma map_try_gpv: "map_gpv f g (TRY gpv ELSE gpv') = TRY map_gpv f g gpv ELSE map_gpv f g gpv'"
by(simp add: gpv.rel_map try_gpv_parametric[THEN rel_funD, THEN rel_funD] gpv.rel_refl gpv.rel_eq[symmetric])

lemma try_bind_assert_gpv:
  "TRY (assert_gpv b \<bind> f) ELSE gpv = (if b then TRY (f ()) ELSE gpv else gpv)"
by(simp)

subsection {* Executing gpv: possibilistic trace semantics *}

type_synonym ('a, 'out, 'in) raw_trace = "(('out, 'in) event, 'a option) tllist"

context IO_execution_base begin

-- {* Work around the limitation that \isacommand{inductive\_set} does not support instantiated parameters *}
coinductive tracesp_gpv :: "('a, 'out, 'in) gpv \<Rightarrow> ('a, 'out, 'in) raw_trace \<Rightarrow> bool"
where
  Fail: "None \<in> set_pmf (the_gpv r) \<Longrightarrow> tracesp_gpv r (TNil None)"
| Pure: "Pure x \<in> set_spmf (the_gpv r) \<Longrightarrow> tracesp_gpv r (TNil (Some x))"
| IO: 
  "\<lbrakk> IO out c \<in> set_spmf (the_gpv r); input \<in> wf_response out; tracesp_gpv (c input) tr \<rbrakk>
  \<Longrightarrow> tracesp_gpv r (TCons (Event out input) tr)"

definition traces_gpv :: "('a, 'out, 'in) gpv \<Rightarrow> ('a, 'out, 'in) raw_trace set"
where "traces_gpv r \<equiv> {tr. tracesp_gpv r tr}"
lemma tracesp_gpv_traces_gpv_eq [pred_set_conv]: "tracesp_gpv r = (\<lambda>tr. tr \<in> traces_gpv r)" by(simp add: traces_gpv_def)
context begin
local_setup {* Local_Theory.map_background_naming (Name_Space.mandatory_path "traces_gpv") *}
(* name space prefix does not work *)
lemmas intros [intro?] = tracesp_gpv.intros[to_set]
  and coinduct [consumes 1, case_names Fail Pure IO, coinduct set: traces_gpv] = tracesp_gpv.coinduct[to_set]
  and cases [consumes 1, case_names Fail Pure IO, cases set: traces_gpv] = tracesp_gpv.cases[to_set]
  and simps = tracesp_gpv.simps[to_set]
end
lemmas traces_gpv_intros [intro?] = tracesp_gpv.intros[to_set]
  and traces_gpv_cases [consumes 1, case_names Fail Pure IO, cases set: traces_gpv] = tracesp_gpv.cases[to_set]
  and traces_gpv_simps = tracesp_gpv.simps[to_set]

inductive_simps TNil_None_in_traces_gpv [to_set, simp]: "tracesp_gpv r (TNil None)"
inductive_simps TNil_Some_in_traces_gpv [to_set, simp]: "tracesp_gpv r (TNil (Some x))"
inductive_simps TCons_in_traces_gpv [to_set, simp]: "tracesp_gpv r (TCons (Event out input) tr)"

lemma traces_gpv_coinduct [consumes 1, case_names Fail Pure IO, case_conclusion IO wf_response ttl, coinduct set: traces_gpv]:
  assumes "X r tr"
  and Fail: "\<And>r tr. \<lbrakk> X r tr; tr = TNil None \<rbrakk> \<Longrightarrow> None \<in> set_pmf (the_gpv r)"
  and Pure: "\<And>r tr x. \<lbrakk> X r tr; tr = TNil (Some x) \<rbrakk> \<Longrightarrow> Pure x \<in> set_spmf (the_gpv r)"
  and IO: "\<And>r out input tr tr'. \<lbrakk> X r tr; tr = TCons (Event out input) tr' \<rbrakk>
    \<Longrightarrow> input \<in> wf_response out \<and> (\<exists>c. IO out c \<in> set_spmf (the_gpv r) \<and> (X (c input) tr' \<or> tr' \<in> traces_gpv (c input)))"
  shows "tr \<in> traces_gpv r"
using \<open>X r tr\<close>
proof(coinduct rule: tracesp_gpv.coinduct[to_set])
  case (tracesp_gpv r tr)
  show ?case
  proof(cases tr)
    case (TNil xo)
    show ?thesis
    proof(cases xo)
      case None
      hence ?Fail using TNil tracesp_gpv by(simp add: Fail)
      thus ?thesis ..
    next
      case (Some x)
      hence ?Pure using TNil tracesp_gpv by(simp add: Pure)
      thus ?thesis by simp
    qed
  next
    case (TCons event tr')
    obtain out input where "event = Event out input" by(cases event)
    with IO[OF \<open>X r tr\<close>] TCons tracesp_gpv have ?IO by auto
    thus ?thesis by simp
  qed
qed

lemma traces_gpv_coinduct' [consumes 1, case_names Fail Pure IO, case_conclusion IO wf_response ttl]:
  assumes "X r tr"
  and Fail: "\<And>r tr. \<lbrakk> X r tr; is_TNil tr; Option.is_none (terminal tr) \<rbrakk> \<Longrightarrow> None \<in> set_pmf (the_gpv r)"
  and Pure: "\<And>r tr. \<lbrakk> X r tr; is_TNil tr; \<not> Option.is_none (terminal tr) \<rbrakk> \<Longrightarrow> Pure (the (terminal tr)) \<in> set_spmf (the_gpv r)"
  and IO: "\<And>r tr. \<lbrakk> X r tr; \<not> is_TNil tr \<rbrakk>
    \<Longrightarrow> event_in (thd tr) \<in> wf_response (event_out (thd tr)) \<and> (\<exists>c. IO (event_out (thd tr)) c \<in> set_spmf (the_gpv r) \<and> (X (c (event_in (thd tr))) (ttl tr) \<or> ttl tr \<in> traces_gpv (c (event_in (thd tr)))))"
  shows "tr \<in> traces_gpv r"
using \<open>X r tr\<close>
apply(coinduct)
  apply(drule Fail, simp, simp, simp)
 apply(drule Pure, simp, simp, simp)
apply(drule IO, simp, simp)
done

lemma traces_gpv_cases':
  assumes "tr \<in> traces_gpv r"
  obtains (Fail) "is_TNil tr" "Option.is_none (terminal tr)" "None \<in> set_pmf (the_gpv r)"
  | (Pure) "is_TNil tr" "\<not> Option.is_none (terminal tr)" "Pure (the (terminal tr)) \<in> set_spmf (the_gpv r)"
  | (IO) out c where "\<not> is_TNil tr" "out = event_out (thd tr)" "IO out c \<in> set_spmf (the_gpv r)" 
    "event_in (thd tr) \<in> wf_response out" "ttl tr \<in> traces_gpv (c (event_in (thd tr)))"
using assms by cases auto

end

subsection {* Order for @{typ "('a, 'out, 'in) gpv"} *}

coinductive ord_gpv :: "('a, 'out, 'in) gpv \<Rightarrow> ('a, 'out, 'in) gpv \<Rightarrow> bool"
where
  "ord_spmf (rel_generat op = op = (rel_fun op = ord_gpv)) f g \<Longrightarrow> ord_gpv (GPV f) (GPV g)"

inductive_simps ord_gpv_simps [simp]:
  "ord_gpv (GPV f) (GPV g)"

lemma ord_gpv_coinduct [consumes 1, case_names ord_gpv, coinduct pred: ord_gpv]:
  assumes "X f g"
  and step: "\<And>f g. X f g \<Longrightarrow> ord_spmf (rel_generat op = op = (rel_fun op = X)) (the_gpv f) (the_gpv g)"
  shows "ord_gpv f g"
using `X f g`
by(coinduct)(auto dest: step simp add: eq_GPV_iff intro: ord_spmf_mono rel_generat_mono rel_fun_mono)

lemma ord_gpv_the_gpvD:
  "ord_gpv f g \<Longrightarrow> ord_spmf (rel_generat op = op = (rel_fun op = ord_gpv)) (the_gpv f) (the_gpv g)"
by(erule ord_gpv.cases) simp

lemma reflp_equality: "reflp op ="
by(simp add: reflp_def)

lemma ord_gpv_reflI [simp]: "ord_gpv f f"
by(coinduction arbitrary: f)(auto intro: ord_spmf_reflI simp add: rel_generat_same rel_fun_def)

lemma reflp_ord_gpv: "reflp ord_gpv"
by(rule reflpI)(rule ord_gpv_reflI)

lemma ord_gpv_trans:
  assumes "ord_gpv f g" "ord_gpv g h"
  shows "ord_gpv f h"
using assms
proof(coinduction arbitrary: f g h)
  case (ord_gpv f g h)
  have *: "ord_spmf (rel_generat op = op = (rel_fun op = (\<lambda>f h. \<exists>g. ord_gpv f g \<and> ord_gpv g h))) (the_gpv f) (the_gpv h) =
    ord_spmf (rel_generat (op = OO op =) (op = OO op =) (rel_fun op = (ord_gpv OO ord_gpv))) (the_gpv f) (the_gpv h)"
    by(simp add: relcompp.simps[abs_def])
  then show ?case using ord_gpv
    by(auto elim!: ord_gpv.cases simp add: generat.rel_compp ord_spmf_compp fun.rel_compp)
qed

lemma ord_gpv_compp: "(ord_gpv OO ord_gpv) = ord_gpv"
by(auto simp add: fun_eq_iff intro: ord_gpv_trans)

lemma transp_ord_gpv [simp]: "transp ord_gpv"
by(blast intro: transpI ord_gpv_trans)

lemma ord_gpv_antisym:
  "\<lbrakk> ord_gpv f g; ord_gpv g f \<rbrakk> \<Longrightarrow> f = g"
proof(coinduction arbitrary: f g)
  case (Eq_gpv f g)
  let ?R = "rel_generat op = op = (rel_fun op = ord_gpv)"
  from \<open>ord_gpv f g\<close> have "ord_spmf ?R (the_gpv f) (the_gpv g)" by cases simp
  moreover
  from \<open>ord_gpv g f\<close> have "ord_spmf ?R (the_gpv g) (the_gpv f)" by cases simp
  ultimately have "rel_spmf (inf ?R ?R\<inverse>\<inverse>) (the_gpv f) (the_gpv g)"
    by(rule rel_spmf_inf)(auto 4 3 intro: transp_rel_generatI transp_ord_gpv reflp_ord_gpv reflp_equality reflp_fun1 is_equality_eq transp_rel_fun)
  also have "inf ?R ?R\<inverse>\<inverse> = rel_generat (inf op = op =) (inf op = op =) (rel_fun op = (inf ord_gpv ord_gpv\<inverse>\<inverse>))"
    unfolding rel_generat_inf[symmetric] rel_fun_inf[symmetric]
    by(simp add: generat.rel_conversep[symmetric] fun.rel_conversep)
  finally show ?case by(simp add: inf_fun_def)
qed

lemma RFail_least [simp]: "ord_gpv Fail f"
by(coinduction arbitrary: f)(simp add: eq_GPV_iff)

subsection {* Bounds on interaction *}

context notes monotone_SUP[partial_function_mono] begin
declaration {* Partial_Function.init "lfp_strong" @{term lfp.fixp_fun} @{term lfp.mono_body}
  @{thm lfp.fixp_rule_uc} @{thm lfp.fixp_strong_induct_uc} NONE *}

partial_function (lfp_strong) interaction_bound :: "('a, 'out, 'in) gpv \<Rightarrow> enat"
where
  "interaction_bound gpv =
  (SUP generat:set_spmf (the_gpv gpv). case generat of Pure _ \<Rightarrow> 0 | IO out c \<Rightarrow> eSuc (SUP input. interaction_bound (c input)))"
end

lemma interaction_bound_fixp_induct [case_names adm bottom step]:
  "\<lbrakk> ccpo.admissible (fun_lub Sup) (fun_ord op \<le>) P;
     P (\<lambda>_. 0);
    \<And>interaction_bound'. \<lbrakk> P interaction_bound'; \<And>gpv. interaction_bound' gpv \<le> interaction_bound gpv \<rbrakk>
      \<Longrightarrow> P (\<lambda>gpv. \<Squnion>generat\<in>set_spmf (the_gpv gpv). case generat of Pure x \<Rightarrow> 0
         | IO out c \<Rightarrow> eSuc (\<Squnion>input. interaction_bound' (c input))) \<rbrakk>
   \<Longrightarrow> P interaction_bound"
by(erule interaction_bound.fixp_induct)(simp_all add: bot_enat_def fun_ord_def)

lemma interaction_bound_coinduct [consumes 1, case_names interaction_bound]:
  assumes X: "X gpv n"
  and *: "\<And>gpv n out c input. \<lbrakk> X gpv n; IO out c \<in> set_spmf (the_gpv gpv) \<rbrakk> 
    \<Longrightarrow> \<exists>n'. (X (c input) n' \<or> interaction_bound (c input) \<le> n') \<and> eSuc n' \<le> n"
  shows "interaction_bound gpv \<le> n"
using X
proof(induction arbitrary: gpv n rule: interaction_bound_fixp_induct)
  case adm show ?case by(intro cont_intro)
  case bottom show ?case by simp
next
  case (step interaction_bound')
  { fix out c
    assume IO: "IO out c \<in> set_spmf (the_gpv gpv)"
    from *[OF step.prems IO] obtain n' where n: "n = eSuc n'"
      by(cases n rule: co.enat.exhaust) auto
    moreover 
    { fix input
      have "\<exists>n''. (X (c input) n'' \<or> interaction_bound (c input) \<le> n'') \<and> eSuc n'' \<le> n"
        using step.prems IO \<open>n = eSuc n'\<close> by(auto 4 3 dest: *)
      then have "interaction_bound' (c input) \<le> n'" using n
        by(auto dest: step.IH intro: step.hyps[THEN order_trans] elim!: order_trans simp add: neq_zero_conv_eSuc) }
    ultimately have "eSuc (\<Squnion>input. interaction_bound' (c input)) \<le> n"
      by(auto intro: SUP_least) }
  then show ?case by(auto intro!: SUP_least split: generat.split)
qed

lemma interaction_bound_IO:
   "IO out c \<in> set_spmf (the_gpv gpv)
   \<Longrightarrow> eSuc (interaction_bound (c input)) \<le> interaction_bound gpv"
by(rewrite in "_ \<le> \<hole>" interaction_bound.simps)(auto intro!: SUP_upper2)

lemma interaction_bound_Done [simp]: "interaction_bound (Done x) = 0"
by(simp add: interaction_bound.simps)

lemma interaction_bound_Fail [simp]: "interaction_bound Fail = 0"
by(simp add: interaction_bound.simps bot_enat_def)

lemma interaction_bound_Pause [simp]: "interaction_bound (Pause out c) = eSuc (SUP input. interaction_bound (c input))"
by(simp add: interaction_bound.simps)

lemma interaction_bound_lift_spmf [simp]: "interaction_bound (lift_spmf p) = 0"
by(simp add: interaction_bound.simps SUP_constant bot_enat_def)

lemma interaction_bound_assert_gpv [simp]: "interaction_bound (assert_gpv b) = 0"
by(cases b) simp_all

lemma interaction_bound_bind:
  defines "ib1 \<equiv> interaction_bound" and "ib2 \<equiv> interaction_bound"
  shows "interaction_bound (p \<bind> f) \<le> ib1 p + (SUP x:results'_gpv p. ib2 (f x))"
proof(induction arbitrary: p rule: interaction_bound_fixp_induct)
  case adm show ?case by simp
  case bottom show ?case by simp
  case (step interaction_bound') show ?case unfolding ib1_def ib2_def
  proof(rule SUP_least)
    fix generat'
    assume "generat' \<in> set_spmf (the_gpv (p \<bind> f))"
    then obtain generat where generat: "generat \<in> set_spmf (the_gpv p)"
      and *: "case generat of Pure x \<Rightarrow> generat' \<in> set_spmf (the_gpv (f x)) 
           | IO out c \<Rightarrow> generat' = IO out (\<lambda>input. c input \<bind> f)"
      by(clarsimp simp add: set_bind_spmf bind_gpv.sel simp del: bind_gpv_sel')
        (clarsimp split: generat.split_asm simp add: generat.map_comp o_def generat.map_id[unfolded id_def])
    show "(case generat' of Pure x \<Rightarrow> 0 | IO out c \<Rightarrow> eSuc (\<Squnion>input. interaction_bound' (c input)))
           \<le> interaction_bound p + (\<Squnion>x\<in>results'_gpv p. interaction_bound (f x))"
      (is "?lhs \<le> ?p + ?f")
    proof(cases generat)
      case (Pure x)
      have "?lhs \<le> (case generat' of Pure x \<Rightarrow> 0 | IO out c \<Rightarrow> eSuc (\<Squnion>input. interaction_bound (c input)))"
        by(cases generat')(auto intro: step.hyps SUP_mono)
      also have "\<dots> \<le> (SUP generat':set_spmf (the_gpv (f x)). (case generat' of Pure x \<Rightarrow> 0 | IO out c \<Rightarrow> eSuc (\<Squnion>input. interaction_bound (c input))))"
        using * Pure by(auto intro: SUP_upper)
      also have "\<dots> \<le> 0 + ?f" using generat Pure
        by(rewrite in "_ \<le> \<hole>" interaction_bound.simps)(auto 4 3 intro: SUP_upper results'_gpv_Pure)
      also have "\<dots> \<le> ?p + ?f" by simp
      finally show ?thesis .
    next
      case (IO out c)
      with * have "?lhs = eSuc (SUP input. interaction_bound' (c input \<bind> f))" by simp
      also have "\<dots> \<le> eSuc (SUP input. interaction_bound (c input) + (SUP x:results'_gpv (c input). interaction_bound (f x)))"
        by(auto intro: SUP_mono step.IH[unfolded ib1_def ib2_def])
      also have "\<dots> \<le> (case IO out c of Pure (x :: 'a) \<Rightarrow> 0 | IO out c \<Rightarrow> eSuc (SUP input. interaction_bound (c input))) + (SUP input. SUP x:results'_gpv (c input). interaction_bound (f x))"
        by(simp add: iadd_Suc)(metis (mono_tags, lifting) SUP_le_iff add_mono_thms_linordered_semiring(1) dual_order.refl)
      also have "\<dots> \<le> ?p + ?f"
        apply(rewrite in "_ \<le> \<hole>" interaction_bound.simps)
        apply(rule add_mono SUP_least SUP_upper generat[unfolded IO])+
        apply(auto intro: results'_gpv_Cont[OF generat]simp add: IO)
        done
      finally show ?thesis .
    qed
  qed
qed

lemma interaction_bound_bind_lift_spmf [simp]:
  "interaction_bound (lift_spmf p \<bind> f) = (SUP x:set_spmf p. interaction_bound (f x))"
by(subst (1 2) interaction_bound.simps)(simp add: set_bind_spmf bind_UNION SUP_UNION)

text \<open>
  There is no nice @{const interaction_bound} equation for @{const bind_gpv}, as it computes
  an exact bound, but we only need an upper bound.
  As @{typ enat} is hard to work with (and @{term \<infinity>} does not constrain a gpv in any way),
  we work with @{typ nat}.
\<close>

inductive interaction_bounded_by :: "('a, 'out, 'in) gpv \<Rightarrow> nat \<Rightarrow> bool"
for gpv n where
  interaction_bounded_by: "\<lbrakk> interaction_bound gpv \<le> enat n \<rbrakk> \<Longrightarrow> interaction_bounded_by gpv n"

lemmas interaction_bounded_byI = interaction_bounded_by
hide_fact (open) interaction_bounded_by

lemma interaction_bounded_by_mono:
  "\<lbrakk> interaction_bounded_by gpv n; n \<le> m \<rbrakk> \<Longrightarrow> interaction_bounded_by gpv m"
unfolding interaction_bounded_by.simps by(erule order_trans) simp

lemma interaction_bounded_by_contD:
  "\<lbrakk> interaction_bounded_by gpv n; IO out c \<in> set_spmf (the_gpv gpv) \<rbrakk>
  \<Longrightarrow> n > 0 \<and> interaction_bounded_by (c input) (n - 1)"
unfolding interaction_bounded_by.simps
by(subst (asm) interaction_bound.simps)(auto simp add: SUP_le_iff eSuc_le_iff enat_eSuc_iff dest!: bspec)

lemma interaction_bounded_byI_epred:
  assumes "\<And>out c. IO out c \<in> set_spmf (the_gpv gpv) \<Longrightarrow> n \<noteq> 0 \<and> (\<forall>input. interaction_bounded_by (c input) (n - 1))"
  shows "interaction_bounded_by gpv n"
unfolding interaction_bounded_by.simps
by(subst interaction_bound.simps)(auto intro!: SUP_least split: generat.split dest!: assms simp add: eSuc_le_iff enat_eSuc_iff gr0_conv_Suc neq_zero_conv_eSuc interaction_bounded_by.simps)

lemma interaction_bounded_by_IO:
  "\<lbrakk> IO out c \<in> set_spmf (the_gpv gpv); interaction_bounded_by gpv n \<rbrakk>
  \<Longrightarrow> n \<noteq> 0 \<and> interaction_bounded_by (c input) (n - 1)"
by(drule interaction_bound_IO[where input=input])(auto simp add: interaction_bounded_by.simps epred_conv_minus eSuc_le_iff enat_eSuc_iff)

named_theorems interaction_bound

lemma interaction_bounded_by_start:
  "\<lbrakk> TERM gpv; interaction_bounded_by gpv m; interaction_bounded_by gpv m \<Longrightarrow> thesis \<rbrakk> \<Longrightarrow> thesis"
by simp

lemma interaction_bounded_by_Done [iff]: "interaction_bounded_by (Done x) n"
by(simp add: interaction_bounded_by.simps)

lemma interaction_bounded_by_DoneI [interaction_bound]: "interaction_bounded_by (Done x) 0"
by simp

lemma interaction_bounded_by_Fail [iff]: "interaction_bounded_by Fail n"
by(simp add: interaction_bounded_by.simps)

lemma interaction_bounded_by_FailI [interaction_bound]: "interaction_bounded_by Fail 0"
by simp

lemma interaction_bounded_by_lift_spmf [iff]: "interaction_bounded_by (lift_spmf p) n"
by(simp add: interaction_bounded_by.simps)

lemma interaction_bounded_by_lift_spmfI [interaction_bound]: "interaction_bounded_by (lift_spmf p) 0"
by simp

lemma interaction_bounded_by_assert_gpv [simp]: "interaction_bounded_by (assert_gpv b) n"
by(cases b) simp_all

lemma interaction_bounded_by_assert_gpvI [interaction_bound]: "interaction_bounded_by (assert_gpv b) 0"
by simp

lemma interaction_bounded_by_Pause [simp]:
  "interaction_bounded_by (Pause out c) n \<longleftrightarrow> 0 < n \<and> (\<forall>input. interaction_bounded_by (c input) (n - 1))"
by(auto 4 3 simp add: interaction_bounded_by.simps eSuc_le_iff enat_eSuc_iff gr0_conv_Suc intro: SUP_least dest: order_trans[OF SUP_upper, rotated])

lemma interaction_bounded_by_bind_PauseI [interaction_bound]:
  "(\<And>input. interaction_bounded_by (c input \<bind> f) n)
  \<Longrightarrow> interaction_bounded_by (Pause out c \<bind> f) (n + 1)"
by(simp add: plus_1_eSuc)

lemma interaction_bounded_by_bindI [interaction_bound]:
  "\<lbrakk> interaction_bounded_by gpv n; \<And>x. x \<in> results'_gpv gpv \<Longrightarrow> interaction_bounded_by (f x) m \<rbrakk>
  \<Longrightarrow> interaction_bounded_by (gpv \<bind> f) (n + m)"
unfolding interaction_bounded_by.simps plus_enat_simps(1)[symmetric]
by(rule order_trans[OF interaction_bound_bind])(erule add_mono, simp add: SUP_least)

lemma interaction_bounded_by_bind_lift_spmf [iff]:
  "interaction_bounded_by (lift_spmf p \<bind> f) n \<longleftrightarrow> (\<forall>x\<in>set_spmf p. interaction_bounded_by (f x) n)"
by(simp add: interaction_bounded_by.simps SUP_le_iff)

lemma interaction_bounded_by_bind_lift_spmfI [interaction_bound]:
  "(\<And>x. x \<in> set_spmf p \<Longrightarrow> interaction_bounded_by (f x) n)
  \<Longrightarrow> interaction_bounded_by (lift_spmf p \<bind> f) n"
by(simp)

lemma interaction_bounded_by_bind_DoneI [interaction_bound]:
  "interaction_bounded_by (f x) n \<Longrightarrow> interaction_bounded_by (Done x \<bind> f) n"
by(simp)

lemma interaction_bounded_by_if [interaction_bound]:
  "\<lbrakk> b \<Longrightarrow> interaction_bounded_by gpv1 n; \<not> b \<Longrightarrow> interaction_bounded_by gpv2 m \<rbrakk>
  \<Longrightarrow> interaction_bounded_by (if b then gpv1 else gpv2) (max n m)"
by(auto 4 3 simp add: max_def not_le elim: interaction_bounded_by_mono)


subsection {* Typing *}

subsubsection {* Interface between gpvs and rpvs *}

type_synonym ('call, 'ret) \<I> = "'call set \<times> ('call \<Rightarrow> 'ret set)"

fun outs_\<I> :: "('call, 'ret) \<I> \<Rightarrow> 'call set"
where "outs_\<I> (outs, c) = outs"

lemma outs_\<I>_conv: "outs_\<I> = fst"
by(simp add: fun_eq_iff)

fun responses_\<I> :: "('call, 'ret) \<I> \<Rightarrow> 'call \<Rightarrow> 'ret set"
where "responses_\<I> (outs, c) = c"

lemma responses_\<I>_conv: "responses_\<I> = snd"
by(simp add: fun_eq_iff)

abbreviation \<I>_top :: "('call, 'ret) \<I>"
where "\<I>_top \<equiv> (UNIV, \<lambda>_. UNIV)"


context begin
qualified inductive resultsp_gpv :: "('out, 'in) \<I> \<Rightarrow> 'a \<Rightarrow> ('a, 'out, 'in) gpv \<Rightarrow> bool"
  for \<Gamma> x
where
  Pure: "Pure x \<in> set_spmf (the_gpv gpv) \<Longrightarrow> resultsp_gpv \<Gamma> x gpv"
| IO:
  "\<lbrakk> IO out c \<in> set_spmf (the_gpv gpv); input \<in> responses_\<I> \<Gamma> out; resultsp_gpv \<Gamma> x (c input) \<rbrakk>
  \<Longrightarrow> resultsp_gpv \<Gamma> x gpv"

definition results_gpv :: "('out, 'in) \<I> \<Rightarrow> ('a, 'out, 'in) gpv \<Rightarrow> 'a set"
where "results_gpv \<Gamma> gpv \<equiv> {x. resultsp_gpv \<Gamma> x gpv}"

lemma resultsp_gpv_results_gpv_eq [pred_set_conv]: "resultsp_gpv \<Gamma> x gpv \<longleftrightarrow> x \<in> results_gpv \<Gamma> gpv"
by(simp add: results_gpv_def)

context begin
local_setup {* Local_Theory.map_background_naming (Name_Space.mandatory_path "results_gpv") *}

lemmas intros [intro?] = resultsp_gpv.intros[to_set]
  and Pure = Pure[to_set]
  and IO = IO[to_set]
  and induct [consumes 1, case_names Pure IO, induct set: results_gpv] = resultsp_gpv.induct[to_set]
  and cases [consumes 1, case_names Pure IO, cases set: results_gpv] = resultsp_gpv.cases[to_set]
  and simps = resultsp_gpv.simps[to_set]
end

inductive_simps results_gpv_GPV [to_set, simp]: "resultsp_gpv \<Gamma> x (GPV gpv)"

end

lemma results_gpv_Done [iff]: "results_gpv \<Gamma> (Done x) = {x}"
by(auto simp add: Done.ctr)

lemma results_gpv_Fail [iff]: "results_gpv \<Gamma> Fail = {}"
by(auto simp add: Fail_def)

lemma results_gpv_Pause [simp]:
  "results_gpv \<Gamma> (Pause out c) = (\<Union>input\<in>responses_\<I> \<Gamma> out. results_gpv \<Gamma> (c input))"
by(auto simp add: Pause.ctr)

lemma results_gpv_lift_spmf [iff]: "results_gpv \<Gamma> (lift_spmf p) = set_spmf p"
by(auto simp add: lift_spmf.ctr)

lemma results_gpv_assert_gpv [simp]: "results_gpv \<Gamma> (assert_gpv b) = (if b then {()} else {})"
by auto

lemma results_gpv_bind_gpv [simp]:
  "results_gpv \<Gamma> (gpv \<bind> f) = (\<Union>x\<in>results_gpv \<Gamma> gpv. results_gpv \<Gamma> (f x))"
  (is "?lhs = ?rhs")
proof(intro set_eqI iffI)
  fix x
  assume "x \<in> ?lhs"
  then show "x \<in> ?rhs"
  proof(induction gpv'\<equiv>"gpv \<bind> f" arbitrary: gpv)
    case Pure thus ?case
      by(auto 4 3 simp add: set_bind_spmf split: if_split_asm intro: results_gpv.intros rev_bexI)
  next
    case (IO out c input)
    from \<open>IO out c \<in> _\<close>
    obtain generat where generat: "generat \<in> set_spmf (the_gpv gpv)"
      and *: "IO out c \<in> set_spmf (if is_Pure generat then the_gpv (f (result generat))
                                   else return_spmf (IO (output generat) (\<lambda>input. continuation generat input \<bind> f)))"
      by(auto simp add: set_bind_spmf)
    thus ?case
    proof(cases generat)
      case (Pure y)
      with generat have "y \<in> results_gpv \<Gamma> gpv" by(auto intro: results_gpv.intros)
      thus ?thesis using * Pure \<open>input \<in> responses_\<I> \<Gamma> out\<close> \<open>x \<in> results_gpv \<Gamma> (c input)\<close>
        by(auto intro: results_gpv.IO)
    next
      case (IO out' c')
      hence [simp]: "out' = out"
        and c: "\<And>input. c input = bind_gpv (c' input) f" using * by simp_all
      from IO.hyps(4)[OF c] obtain y where y: "y \<in> results_gpv \<Gamma> (c' input)"
        and "x \<in> results_gpv \<Gamma> (f y)" by blast
      from y IO generat have "y \<in> results_gpv \<Gamma> gpv" using `input \<in> responses_\<I> \<Gamma> out`
        by(auto intro: results_gpv.IO)
      with \<open>x \<in> results_gpv \<Gamma> (f y)\<close> show ?thesis by blast
    qed
  qed
next
  fix x
  assume "x \<in> ?rhs"
  then obtain y where y: "y \<in> results_gpv \<Gamma> gpv"
    and x: "x \<in> results_gpv \<Gamma> (f y)" by blast
  from y show "x \<in> ?lhs"
  proof(induction)
    case (Pure gpv)
    with x show ?case
      by cases(auto 4 4 intro: results_gpv.intros rev_bexI simp add: set_bind_spmf)
  qed(auto 4 4 simp add: set_bind_spmf intro: rev_bexI results_gpv.IO)
qed

lemma results_gpv_\<I>_top: "results_gpv \<I>_top = results'_gpv"
proof(intro ext set_eqI iffI)
  fix x gpv
  assume "x \<in> results_gpv \<I>_top gpv"
  then show "x \<in> results'_gpv gpv"
    by induction(auto intro: results'_gpvI)
next
  fix x gpv
  assume "x \<in> results'_gpv gpv"
  then show "x \<in> results_gpv \<I>_top gpv"
    by induction(auto intro: results_gpv.intros elim!: generat.set_cases)
qed

lemma results'_bind_gpv [simp]:
  "results'_gpv (bind_gpv gpv f) = (\<Union>x\<in>results'_gpv gpv. results'_gpv (f x))"
unfolding results_gpv_\<I>_top[symmetric] by simp

subsubsection {* Type judgements *}

coinductive WT_gpv :: "('out, 'in) \<I> \<Rightarrow> ('a, 'out, 'in) gpv \<Rightarrow> bool" ("((_)/ \<turnstile>g (_) \<surd>)" [100, 0] 99)
  for \<Gamma>
where
  "(\<And>out c. IO out c \<in> set_spmf gpv \<Longrightarrow> out \<in> outs_\<I> \<Gamma> \<and> (\<forall>input\<in>responses_\<I> \<Gamma> out. \<Gamma> \<turnstile>g c input \<surd>))
  \<Longrightarrow> \<Gamma> \<turnstile>g GPV gpv \<surd>"

lemma WT_gpv_coinduct [consumes 1, case_names WT_gpv, case_conclusion WT_gpv out cont, coinduct pred: WT_gpv]:
  assumes *: "X gpv"
  and step: "\<And>gpv out c.
    \<lbrakk> X gpv; IO out c \<in> set_spmf (the_gpv gpv) \<rbrakk>
    \<Longrightarrow> out \<in> outs_\<I> \<Gamma> \<and> (\<forall>input \<in> responses_\<I> \<Gamma> out. X (c input) \<or> \<Gamma> \<turnstile>g c input \<surd>)"
  shows "\<Gamma> \<turnstile>g gpv \<surd>"
using * by(coinduct)(auto dest: step simp add: eq_GPV_iff)

lemma WT_gpv_simps:
  "\<Gamma> \<turnstile>g GPV gpv \<surd> \<longleftrightarrow>
   (\<forall>out c. IO out c \<in> set_spmf gpv \<longrightarrow> out \<in> outs_\<I> \<Gamma> \<and> (\<forall>input\<in>responses_\<I> \<Gamma> out. \<Gamma> \<turnstile>g c input \<surd>))"
by(subst WT_gpv.simps) simp

lemma WT_gpvI:
  "(\<And>out c. IO out c \<in> set_spmf (the_gpv gpv) \<Longrightarrow> out \<in> outs_\<I> \<Gamma> \<and> (\<forall>input\<in>responses_\<I> \<Gamma> out. \<Gamma> \<turnstile>g c input \<surd>))
  \<Longrightarrow> \<Gamma> \<turnstile>g gpv \<surd>"
by(cases gpv)(simp add: WT_gpv_simps)

lemma WT_gpvD:
  assumes "\<Gamma> \<turnstile>g gpv \<surd>"
  shows WT_gpv_OutD: "IO out c \<in> set_spmf (the_gpv gpv) \<Longrightarrow> out \<in> outs_\<I> \<Gamma>"
  and WT_gpv_ContD: "\<lbrakk> IO out c \<in> set_spmf (the_gpv gpv); input \<in> responses_\<I> \<Gamma> out \<rbrakk> \<Longrightarrow> \<Gamma> \<turnstile>g c input \<surd>"
using assms by(cases, fastforce)+

lemma WT_gpv_Done [iff]: "\<Gamma> \<turnstile>g Done x \<surd>"
by(rule WT_gpvI) simp_all

lemma WT_gpv_Fail [iff]: "\<Gamma> \<turnstile>g Fail \<surd>"
by(rule WT_gpvI) simp_all

lemma WT_gpv_PauseI: 
  "\<lbrakk> out \<in> outs_\<I> \<Gamma>; \<And>input. input \<in> responses_\<I> \<Gamma> out \<Longrightarrow> \<Gamma> \<turnstile>g c input \<surd> \<rbrakk>
   \<Longrightarrow> \<Gamma> \<turnstile>g Pause out c \<surd>"
by(rule WT_gpvI) simp_all

lemma WT_gpv_Pause [iff]:
  "\<Gamma> \<turnstile>g Pause out c \<surd> \<longleftrightarrow> out \<in> outs_\<I> \<Gamma> \<and> (\<forall>input \<in> responses_\<I> \<Gamma> out. \<Gamma> \<turnstile>g c input \<surd>)"
by(auto intro: WT_gpv_PauseI dest: WT_gpvD)

lemma WT_gpv_bindI:
  "\<lbrakk> \<Gamma> \<turnstile>g gpv \<surd>; \<And>x. x \<in> results_gpv \<Gamma> gpv \<Longrightarrow> \<Gamma> \<turnstile>g f x \<surd> \<rbrakk>
  \<Longrightarrow> \<Gamma> \<turnstile>g gpv \<bind> f \<surd>"
proof(coinduction arbitrary: gpv)
  case [rule_format]: (WT_gpv out c gpv)
  from \<open>IO out c \<in> _\<close>
  obtain generat where generat: "generat \<in> set_spmf (the_gpv gpv)"
    and *: "IO out c \<in> set_spmf (if is_Pure generat then the_gpv (f (result generat))
                                 else return_spmf (IO (output generat) (\<lambda>input. continuation generat input \<bind> f)))"
    by(auto simp add: set_bind_spmf)
  show ?case
  proof(cases generat)
    case (Pure y)
    with generat have "y \<in> results_gpv \<Gamma> gpv" by(auto intro: results_gpv.Pure)
    hence "\<Gamma> \<turnstile>g f y \<surd>" by(rule WT_gpv)
    with * Pure show ?thesis by(auto dest: WT_gpvD) 
  next
    case (IO out' c')
    hence [simp]: "out' = out"
      and c: "\<And>input. c input = bind_gpv (c' input) f" using * by simp_all
    from generat IO have **: "IO out c' \<in> set_spmf (the_gpv gpv)" by simp
    with \<open>\<Gamma> \<turnstile>g gpv \<surd>\<close> have ?out by(auto dest: WT_gpvD)
    moreover {
      fix input
      assume input: "input \<in> responses_\<I> \<Gamma> out"
      with \<open>\<Gamma> \<turnstile>g gpv \<surd>\<close> ** have "\<Gamma> \<turnstile>g c' input \<surd>" by(rule WT_gpvD)
      moreover {
        fix y
        assume "y \<in> results_gpv \<Gamma> (c' input)"
        with ** input have "y \<in> results_gpv \<Gamma> gpv" by(rule results_gpv.IO)
        hence "\<Gamma> \<turnstile>g f y \<surd>" by(rule WT_gpv) }
      moreover note calculation }
    hence ?cont using c by blast
    ultimately show ?thesis ..
  qed
qed

lemma WT_gpv_bindD2:
  assumes WT: "\<Gamma> \<turnstile>g gpv \<bind> f \<surd>"
  and x: "x \<in> results_gpv \<Gamma> gpv"
  shows "\<Gamma> \<turnstile>g f x \<surd>"
using x WT
proof induction
  case (Pure gpv)
  show ?case
  proof(rule WT_gpvI)
    fix out c
    assume "IO out c \<in> set_spmf (the_gpv (f x))"
    with Pure have "IO out c \<in> set_spmf (the_gpv (gpv \<bind> f))"
      by(auto simp add: set_bind_spmf intro: rev_bexI)
    with \<open>\<Gamma> \<turnstile>g gpv \<bind> f \<surd>\<close> show "out \<in> outs_\<I> \<Gamma> \<and> (\<forall>input\<in>responses_\<I> \<Gamma> out. \<Gamma> \<turnstile>g c input \<surd>)"
      by(auto dest: WT_gpvD)
  qed
next
  case (IO out c gpv input)
  from \<open>IO out c \<in> _\<close>
  have "IO out (\<lambda>input. bind_gpv (c input) f) \<in> set_spmf (the_gpv (gpv \<bind> f))"
    by(auto simp add: set_bind_spmf intro: rev_bexI)
  with IO.prems have "\<Gamma> \<turnstile>g c input \<bind> f \<surd>" using \<open>input \<in> _\<close> by(rule WT_gpv_ContD)
  thus ?case by(rule IO.IH)
qed

lemma WT_gpv_bindD1: "\<Gamma> \<turnstile>g gpv \<bind> f \<surd> \<Longrightarrow> \<Gamma> \<turnstile>g gpv \<surd>"
proof(coinduction arbitrary: gpv)
  case (WT_gpv out c gpv)
  from \<open>IO out c \<in> _\<close>
  have "IO out (\<lambda>input. bind_gpv (c input) f) \<in> set_spmf (the_gpv (gpv \<bind> f))"
    by(auto simp add: set_bind_spmf intro: rev_bexI)
  with \<open>\<Gamma> \<turnstile>g gpv \<bind> f \<surd>\<close> show ?case
    by(auto simp del: bind_gpv_sel' dest: WT_gpvD)
qed

lemma WT_gpv_bind [simp]: "\<Gamma> \<turnstile>g gpv \<bind> f \<surd> \<longleftrightarrow> \<Gamma> \<turnstile>g gpv \<surd> \<and> (\<forall>x\<in>results_gpv \<Gamma> gpv. \<Gamma> \<turnstile>g f x \<surd>)"
by(blast intro: WT_gpv_bindI dest: WT_gpv_bindD1 WT_gpv_bindD2)

lemma WT_gpv_top [simp, intro!]: "\<I>_top \<turnstile>g gpv \<surd>"
by(coinduction arbitrary: gpv)(auto)

lemma WT_gpv_lift_spmf [simp, intro!]: "\<I> \<turnstile>g lift_spmf p \<surd>"
by(rule WT_gpvI) auto

lemma WT_gpv_coinduct_bind [consumes 1, case_names WT_gpv, case_conclusion WT_gpv out cont]:
  assumes *: "X gpv"
  and step: "\<And>gpv out c. \<lbrakk> X gpv; IO out c \<in> set_spmf (the_gpv gpv) \<rbrakk>
    \<Longrightarrow> out \<in> outs_\<I> \<I> \<and> (\<forall>input\<in>responses_\<I> \<I> out.
            X (c input) \<or>
            \<I> \<turnstile>g c input \<surd> \<or>
            (\<exists>(gpv' :: ('b, 'call, 'ret) gpv) f. c input = gpv' \<bind> f \<and> \<I> \<turnstile>g gpv' \<surd> \<and> (\<forall>x\<in>results_gpv \<I> gpv'. X (f x))))"
  shows "\<I> \<turnstile>g gpv \<surd>"
proof -
  fix x
  def gpv' \<equiv> "Done x :: ('b, 'call, 'ret) gpv"
    and f \<equiv> "(\<lambda>_. gpv) :: 'b \<Rightarrow> ('a, 'call, 'ret) gpv" 
  with * have "\<I> \<turnstile>g gpv' \<surd>" and "\<And>x. x \<in> results_gpv \<I> gpv' \<Longrightarrow> X (f x)" by simp_all
  then have "\<I> \<turnstile>g gpv' \<bind> f \<surd>"
  proof(coinduction arbitrary: gpv' f rule: WT_gpv_coinduct)
    case [rule_format]: (WT_gpv out c gpv')
    from \<open>IO out c \<in> _\<close>
    obtain generat where generat: "generat \<in> set_spmf (the_gpv gpv')"
      and *: "IO out c \<in> set_spmf (if is_Pure generat
        then the_gpv (f (result generat))
        else return_spmf (IO (output generat) (\<lambda>input. continuation generat input \<bind> f)))"
      by(clarsimp simp add: set_bind_spmf)
    show ?case
    proof(cases generat)
      case (Pure x)
      from Pure * have IO: "IO out c \<in> set_spmf (the_gpv (f x))" by simp
      from generat Pure have "x \<in> results_gpv \<I> gpv'" by (simp add: results_gpv.Pure)
      then have "X (f x)" by(rule WT_gpv)
      from step[OF this IO] show ?thesis by(auto 4 4 intro: exI[where x="Done _"])
    next
      case (IO out' c')
      with * have [simp]: "out' = out"
        and c: "c = (\<lambda>input. c' input \<bind> f)" by simp_all
      from IO generat have IO: "IO out c' \<in> set_spmf (the_gpv gpv')" by simp
      then have "\<And>input. input \<in> responses_\<I> \<I> out \<Longrightarrow> results_gpv \<I> (c' input) \<subseteq> results_gpv \<I> gpv'"
        by(auto intro: results_gpv.IO)
      with WT_gpvD[OF \<open>\<I> \<turnstile>g gpv' \<surd>\<close> IO] show ?thesis unfolding c using WT_gpv(2) by blast
    qed
  qed
  then show ?thesis unfolding gpv'_def f_def by simp
qed

subsection {* Sub-gpvs *}

context begin
qualified inductive sub_gpvsp :: "('out, 'in) \<I> \<Rightarrow> ('a, 'out, 'in) gpv \<Rightarrow> ('a, 'out, 'in) gpv \<Rightarrow> bool"
  for \<I> x
where
  base:
  "\<lbrakk> IO out c \<in> set_spmf (the_gpv gpv); input \<in> responses_\<I> \<I> out; x = c input \<rbrakk> 
  \<Longrightarrow> sub_gpvsp \<I> x gpv"
| cont: 
  "\<lbrakk> IO out c \<in> set_spmf (the_gpv gpv); input \<in> responses_\<I> \<I> out; sub_gpvsp \<I> x (c input) \<rbrakk>
  \<Longrightarrow> sub_gpvsp \<I> x gpv"

qualified lemma sub_gpvsp_base:
  "\<lbrakk> IO out c \<in> set_spmf (the_gpv gpv); input \<in> responses_\<I> \<I> out \<rbrakk> 
  \<Longrightarrow> sub_gpvsp \<I> (c input) gpv"
by(rule base) simp_all

definition sub_gpvs :: "('out, 'in) \<I> \<Rightarrow>  ('a, 'out, 'in) gpv \<Rightarrow> ('a, 'out, 'in) gpv set"
where "sub_gpvs \<I> gpv \<equiv> {x. sub_gpvsp \<I> x gpv}"

lemma sub_gpvsp_sub_gpvs_eq [pred_set_conv]: "sub_gpvsp \<I> x gpv \<longleftrightarrow> x \<in> sub_gpvs \<I> gpv"
by(simp add: sub_gpvs_def)

context begin
local_setup {* Local_Theory.map_background_naming (Name_Space.mandatory_path "sub_gpvs") *}

lemmas intros [intro?] = sub_gpvsp.intros[to_set]
  and base = sub_gpvsp_base[to_set]
  and cont = cont[to_set]
  and induct [consumes 1, case_names Pure IO, induct set: sub_gpvs] = sub_gpvsp.induct[to_set]
  and cases [consumes 1, case_names Pure IO, cases set: sub_gpvs] = sub_gpvsp.cases[to_set]
  and simps = sub_gpvsp.simps[to_set]
end
end

lemma WT_sub_gpvsD:
  assumes "\<I> \<turnstile>g gpv \<surd>" and "gpv' \<in> sub_gpvs \<I> gpv"
  shows "\<I> \<turnstile>g gpv' \<surd>"
using assms(2,1) by(induction)(auto dest: WT_gpvD)

lemma WT_sub_gpvsI:
  "\<lbrakk> \<And>out c. IO out c \<in> set_spmf (the_gpv gpv) \<Longrightarrow> out \<in> outs_\<I> \<Gamma>; 
     \<And>gpv'. gpv' \<in> sub_gpvs \<Gamma> gpv \<Longrightarrow> \<Gamma> \<turnstile>g gpv' \<surd> \<rbrakk>
  \<Longrightarrow> \<Gamma> \<turnstile>g gpv \<surd>"
by(rule WT_gpvI)(auto intro: sub_gpvs.base)

subsection {* Losslessness *}

text {* A gpv is lossless iff we are guaranteed to get a result after a finite number of interactions
  that respect the interface *}

inductive lossless_gpv :: "('out, 'in) \<I> \<Rightarrow> ('a, 'out, 'in) gpv \<Rightarrow> bool"
  for \<I>
where
  lossless_gpv:
  "\<lbrakk> lossless_spmf p;
     \<And>out c input. \<lbrakk> IO out c \<in> set_spmf p; input \<in> responses_\<I> \<I> out \<rbrakk>
     \<Longrightarrow> lossless_gpv \<I> (c input) \<rbrakk>
  \<Longrightarrow> lossless_gpv \<I> (GPV p)"

hide_fact lossless_gpv

lemma lossless_gpvI [intro?]:
  "\<lbrakk> lossless_spmf (the_gpv gpv);
     \<And>out c input. \<lbrakk> IO out c \<in> set_spmf (the_gpv gpv); input \<in> responses_\<I> \<I> out \<rbrakk>
     \<Longrightarrow> lossless_gpv \<I> (c input) \<rbrakk>
  \<Longrightarrow> lossless_gpv \<I> gpv"
by(cases gpv)(auto intro: lossless_gpv.intros)

lemma lossless_gpvD:
  assumes "lossless_gpv \<I> gpv"
  shows lossless_gpv_lossless_spmfD: "lossless_spmf (the_gpv gpv)"
  and lossless_gpv_continuationD:
  "\<lbrakk> IO out c \<in> set_spmf (the_gpv gpv); input \<in> responses_\<I> \<I> out \<rbrakk> \<Longrightarrow> lossless_gpv \<I> (c input)"
using assms by(auto elim: lossless_gpv.cases)

lemma lossless_gpv_Done [iff]: "lossless_gpv \<I> (Done x)"
by(rule lossless_gpvI) auto

lemma lossless_gpv_Fail [iff]: "\<not> lossless_gpv \<I> Fail"
by(auto dest: lossless_gpvD)

lemma lossless_gpv_Pause [simp]:
  "lossless_gpv \<I> (Pause out c) \<longleftrightarrow> (\<forall>input \<in> responses_\<I> \<I> out. lossless_gpv \<I> (c input))"
by(auto dest: lossless_gpvD intro: lossless_gpvI)

lemma lossless_gpv_assert_gpv [iff]: "lossless_gpv \<I> (assert_gpv b) \<longleftrightarrow> b"
by(cases b) simp_all

lemma lossless_gpv_lift_spmf [iff]: "lossless_gpv \<I> (lift_spmf p) \<longleftrightarrow> lossless_spmf p"
by(auto dest: lossless_gpvD intro: lossless_gpvI)

lemma lossless_bind_gpvI:
  "\<lbrakk> lossless_gpv \<I> gpv; \<And>x. x \<in> results_gpv \<I> gpv \<Longrightarrow> lossless_gpv \<I> (f x) \<rbrakk>
  \<Longrightarrow> lossless_gpv \<I> (gpv \<bind> f)"
proof(induction rule: lossless_gpv.induct)
  case (lossless_gpv p)
  { fix x
    assume "Pure x \<in> set_spmf p"
    hence "x \<in> results_gpv \<I> (GPV p)" by simp
    hence "lossless_gpv \<I> (f x)" by(rule lossless_gpv.prems) }
  with \<open>lossless_spmf p\<close> show ?case unfolding GPV_bind
    apply(intro lossless_gpv.intros)
     apply(fastforce dest: lossless_gpvD split: generat.split)
    apply(clarsimp simp add: set_bind_spmf; split generat.split_asm)
    apply(auto dest: lossless_gpvD intro!: lossless_gpv)
    done
qed

lemma lossless_bind_gpvD1: "lossless_gpv \<I> (gpv \<bind> f) \<Longrightarrow> lossless_gpv \<I> gpv"
proof(induction gpv'\<equiv>"gpv \<bind> f" arbitrary: gpv rule: lossless_gpv.induct)
  case (lossless_gpv p)
  obtain p' where gpv: "gpv = GPV p'" by(cases gpv)
  from lossless_gpv.hyps gpv have "lossless_spmf p'" by(simp add: GPV_bind)
  then show ?case unfolding gpv
  proof(rule lossless_gpv.intros)
    fix out c input
    assume "IO out c \<in> set_spmf p'" "input \<in> responses_\<I> \<I> out"
    hence "IO out (\<lambda>input. c input \<bind> f) \<in> set_spmf p" using lossless_gpv.hyps gpv
      by(auto simp add: GPV_bind set_bind_spmf intro: rev_bexI)
    thus "lossless_gpv \<I> (c input)" using \<open>input \<in> _\<close> by(rule lossless_gpv.hyps) simp
  qed
qed

lemma lossless_bind_gpvD2:
  assumes "lossless_gpv \<Gamma> (gpv \<bind> f)"
  and "x \<in> results_gpv \<Gamma> gpv"
  shows "lossless_gpv \<Gamma> (f x)"
using assms(2,1)
proof(induction)
  case (Pure gpv)
  thus ?case
    by -(rule lossless_gpvI, auto 4 4 dest: lossless_gpvD simp add: set_bind_spmf intro: rev_bexI)
qed(auto 4 4 dest: lossless_gpvD simp add: set_bind_spmf intro: rev_bexI)

lemma lossless_bind_gpv [simp]:
  "lossless_gpv \<Gamma> (gpv \<bind> f) \<longleftrightarrow> lossless_gpv \<Gamma> gpv \<and> (\<forall>x\<in>results_gpv \<Gamma> gpv. lossless_gpv \<Gamma> (f x))"
by(blast intro: lossless_bind_gpvI dest: lossless_bind_gpvD1 lossless_bind_gpvD2)

lemma lossless_WT_gpv_induct [consumes 2, case_names lossless_gpv]:
  assumes lossless: "lossless_gpv \<I> gpv"
  and WT: "\<I> \<turnstile>g gpv \<surd>"
  and step: "\<And>p. \<lbrakk>
       lossless_spmf p;
       \<And>out c. IO out c \<in> set_spmf p \<Longrightarrow> out \<in> outs_\<I> \<I>;
       \<And>out c input. \<lbrakk>IO out c \<in> set_spmf p; out \<in> outs_\<I> \<I> \<Longrightarrow> input \<in> responses_\<I> \<I> out\<rbrakk> \<Longrightarrow> lossless_gpv \<I> (c input);
       \<And>out c input. \<lbrakk>IO out c \<in> set_spmf p; out \<in> outs_\<I> \<I> \<Longrightarrow> input \<in> responses_\<I> \<I> out\<rbrakk> \<Longrightarrow> \<I> \<turnstile>g c input \<surd>;
       \<And>out c input. \<lbrakk>IO out c \<in> set_spmf p; out \<in> outs_\<I> \<I> \<Longrightarrow> input \<in> responses_\<I> \<I> out\<rbrakk> \<Longrightarrow> P (c input)\<rbrakk>
      \<Longrightarrow> P (GPV p)"
  shows "P gpv"
using lossless WT
apply(induction)
apply(erule step)
apply(auto elim: WT_gpvD simp add: WT_gpv_simps)
done

lemma lossless_gpv_induct_strong [consumes 1, case_names lossless_gpv]:
  assumes gpv: "lossless_gpv \<I> gpv"
  and step:
  "\<And>p. \<lbrakk> lossless_spmf p;
          \<And>gpv. gpv \<in> sub_gpvs \<I> (GPV p) \<Longrightarrow> lossless_gpv \<I> gpv;
          \<And>gpv. gpv \<in> sub_gpvs \<I> (GPV p) \<Longrightarrow> P gpv \<rbrakk>
       \<Longrightarrow> P (GPV p)"
  shows "P gpv"
proof -
  def gpv' \<equiv> gpv
  then have "gpv' \<in> insert gpv (sub_gpvs \<I> gpv)" by simp
  with gpv have "lossless_gpv \<I> gpv' \<and> P gpv'"
  proof(induction arbitrary: gpv')
    case (lossless_gpv p)
    from \<open>gpv' \<in> insert (GPV p) _\<close> show ?case
    proof(rule insertE)
      assume "gpv' = GPV p"
      moreover have "lossless_gpv \<I> (GPV p)"
        by(auto 4 3 intro: lossless_gpvI lossless_gpv.hyps)
      moreover have "P (GPV p)" using lossless_gpv.hyps(1)
        by(rule step)(fastforce elim: sub_gpvs.cases lossless_gpv.IH[THEN conjunct1] lossless_gpv.IH[THEN conjunct2])+
      ultimately show ?case by simp
    qed(fastforce elim: sub_gpvs.cases lossless_gpv.IH[THEN conjunct1] lossless_gpv.IH[THEN conjunct2])
  qed
  thus ?thesis by(simp add: gpv'_def)
qed

lemma lossless_sub_gpvsI:
  assumes spmf: "lossless_spmf (the_gpv gpv)"
  and sub: "\<And>gpv'. gpv' \<in> sub_gpvs \<I> gpv \<Longrightarrow> lossless_gpv \<I> gpv'"
  shows "lossless_gpv \<I> gpv"
using spmf by(rule lossless_gpvI)(rule sub[OF sub_gpvs.base])

lemma lossless_sub_gpvsD:
  assumes "lossless_gpv \<I> gpv" "gpv' \<in> sub_gpvs \<I> gpv"
  shows "lossless_gpv \<I> gpv'"
using assms(2,1) by(induction)(auto dest: lossless_gpvD)

lemma lossless_WT_gpv_induct_strong [consumes 2, case_names lossless_gpv]:
  assumes lossless: "lossless_gpv \<I> gpv"
  and WT: "\<I> \<turnstile>g gpv \<surd>"
  and step: "\<And>p. \<lbrakk> lossless_spmf p;
       \<And>out c. IO out c \<in> set_spmf p \<Longrightarrow> out \<in> outs_\<I> \<I>;
       \<And>gpv. gpv \<in> sub_gpvs \<I> (GPV p) \<Longrightarrow> lossless_gpv \<I> gpv;
       \<And>gpv. gpv \<in> sub_gpvs \<I> (GPV p) \<Longrightarrow> \<I> \<turnstile>g gpv \<surd>;
       \<And>gpv. gpv \<in> sub_gpvs \<I> (GPV p) \<Longrightarrow> P gpv \<rbrakk>
      \<Longrightarrow> P (GPV p)"
  shows "P gpv"
using lossless WT
apply(induction rule: lossless_gpv_induct_strong)
apply(erule step)
apply(auto elim: WT_gpvD dest: WT_sub_gpvsD)
done

lemma try_gpv_lossless [simp]:
  fixes gpv :: "('a, 'call, 'ret) gpv"
  assumes "lossless_gpv \<I>_top gpv" -- \<open>TODO: generalise to arbitrary typings ?\<close>
  shows "TRY gpv ELSE gpv' = gpv"
using assms
apply induction
apply(rule gpv.expand)
apply(simp add: spmf.map_comp generat.map_comp o_def)
apply(rule map_spmf_idI)
subgoal for p generat by(cases generat) simp_all
done

lemma try_gpv_bind_lossless:
  assumes "lossless_gpv \<I>_top gpv" -- \<open>TODO: generalise to arbitrary typings?\<close>
  shows "TRY bind_gpv gpv f ELSE gpv' = bind_gpv gpv (\<lambda>x. TRY f x ELSE gpv')"
using assms
apply induction
apply(rule gpv.expand)
apply(simp add: map_try_spmf spmf.map_comp o_def generat.map_comp map_spmf_bind_spmf bind_gpv.sel try_gpv_sel del: bind_gpv_sel' try_gpv.sel)
apply(subst try_spmf_bind_spmf_lossless, simp)
apply(rule bind_spmf_cong, simp)
apply(clarsimp simp add: spmf.map_comp o_def map_try_spmf generat.map_comp generat.map_ident[abs_def] split: generat.split)
done

subsection {* Inlining *}

lemma gpv_coinduct_bind [consumes 1, case_names Eq_gpv]:
  fixes gpv gpv' :: "('a, 'call, 'ret) gpv"
  assumes *: "R gpv gpv'"
  and step: "\<And>gpv gpv'. R gpv gpv' 
    \<Longrightarrow> rel_spmf (rel_generat op = op = (rel_fun op = (\<lambda>gpv gpv'. R gpv gpv' \<or> gpv = gpv' \<or> 
      (\<exists>gpv2 :: ('b, 'call, 'ret) gpv. \<exists>gpv2' :: ('c, 'call, 'ret) gpv. \<exists>f f'. gpv = bind_gpv gpv2 f \<and> gpv' = bind_gpv gpv2' f' \<and> 
        rel_gpv (\<lambda>x y. R (f x) (f' y)) op = gpv2 gpv2'))))
      (the_gpv gpv) (the_gpv gpv')"
  shows "gpv = gpv'"
proof -
  fix x y
  def gpv1 \<equiv> "Done x :: ('b, 'call, 'ret) gpv"
    and f \<equiv> "(\<lambda>_. gpv) :: 'b \<Rightarrow> ('a, 'call, 'ret) gpv"
    and gpv1' \<equiv> "Done y :: ('c, 'call, 'ret) gpv"
    and f' \<equiv> "(\<lambda>_. gpv') :: 'c \<Rightarrow> ('a, 'call, 'ret) gpv"
  from * have "rel_gpv (\<lambda>x y. R (f x) (f' y)) op = gpv1 gpv1'"
    by(simp add: gpv1_def gpv1'_def f_def f'_def)
  then have "gpv1 \<bind> f = gpv1' \<bind> f'"
  proof(coinduction arbitrary: gpv1 gpv1' f f' rule: gpv.coinduct_strong)
    case (Eq_gpv gpv1 gpv1' f f')
    from Eq_gpv[simplified gpv.rel_sel] show ?case unfolding bind_gpv.sel spmf_rel_map
      apply(rule rel_spmf_bindI)
      subgoal for generat generat'
        apply(cases generat generat' rule: generat.exhaust[case_product generat.exhaust]; clarsimp simp add: o_def spmf_rel_map generat.rel_map)
        subgoal premises Pure for x y
          using step[OF \<open>R (f x) (f' y)\<close>] apply -
          apply(assumption | rule rel_spmf_mono rel_generat_mono rel_fun_mono refl)+
          apply(fastforce intro: exI[where x="Done _"])+
          done
        subgoal by(fastforce simp add: rel_fun_def)
        done
      done
  qed
  thus ?thesis by(simp add: gpv1_def gpv1'_def f_def f'_def)
qed

text {*
  Inlining one gpv into another. This may throw out arbitrarily many
  interactions between the two gpvs if the inlined one does not call its callee.
  So we define it as the coiteration of a least-fixpoint search operator.
*}

context
  fixes callee :: "'s \<Rightarrow> 'call \<Rightarrow> ('ret \<times> 's, 'call', 'ret') gpv"
  notes [[function_internals]]
begin

partial_function (spmf) inline1
  :: "('a, 'call, 'ret) gpv \<Rightarrow> 's
  \<Rightarrow> ('a \<times> 's + 'call' \<times> ('ret \<times> 's, 'call', 'ret') rpv \<times> ('a, 'call, 'ret) rpv) spmf"
where
  "inline1 gpv s =
   the_gpv gpv \<bind>
   case_generat (\<lambda>x. return_spmf (Inl (x, s)))
     (\<lambda>out rpv. the_gpv (callee s out) \<bind>
         case_generat (\<lambda>(x, y). inline1 (rpv x) y)
          (\<lambda>out rpv'. return_spmf (Inr (out, rpv', rpv))))"

lemma inline1_unfold:
  "inline1 gpv s =
   the_gpv gpv \<bind>
   case_generat (\<lambda>x. return_spmf (Inl (x, s)))
     (\<lambda>out rpv. the_gpv (callee s out) \<bind>
         case_generat (\<lambda>(x, y). inline1 (rpv x) y)
          (\<lambda>out rpv'. return_spmf (Inr (out, rpv', rpv))))"
by(fact inline1.simps)

lemma inline1_fixp_induct [case_names adm bottom step]:
  assumes "ccpo.admissible (fun_lub lub_spmf) (fun_ord (ord_spmf op =)) (\<lambda>inline1'. P (\<lambda>gpv s. inline1' (gpv, s)))"
  and "P (\<lambda>_ _. return_pmf None)"
  and "\<And>inline1'. P inline1' \<Longrightarrow> P (\<lambda>gpv s. the_gpv gpv \<bind> case_generat (\<lambda>x. return_spmf (Inl (x, s))) (\<lambda>out rpv. the_gpv (callee s out) \<bind> case_generat (\<lambda>(x, y). inline1' (rpv x) y) (\<lambda>out rpv'. return_spmf (Inr (out, rpv', rpv)))))"
  shows "P inline1"
using assms by(rule inline1.fixp_induct[unfolded curry_conv[abs_def]])

lemma inline1_fixp_induct_strong [case_names adm bottom step]:
  assumes "ccpo.admissible (fun_lub lub_spmf) (fun_ord (ord_spmf op =)) (\<lambda>inline1'. P (\<lambda>gpv s. inline1' (gpv, s)))"
  and "P (\<lambda>_ _. return_pmf None)"
  and "\<And>inline1'. \<lbrakk> \<And>gpv s. ord_spmf op = (inline1' gpv s) (inline1 gpv s); P inline1' \<rbrakk>
    \<Longrightarrow> P (\<lambda>gpv s. the_gpv gpv \<bind> case_generat (\<lambda>x. return_spmf (Inl (x, s))) (\<lambda>out rpv. the_gpv (callee s out) \<bind> case_generat (\<lambda>(x, y). inline1' (rpv x) y) (\<lambda>out rpv'. return_spmf (Inr (out, rpv', rpv)))))"
  shows "P inline1"
using assms by(rule spmf.fixp_strong_induct_uc[where P="\<lambda>f. P (curry f)" and U=case_prod and C=curry, OF inline1.mono inline1_def, simplified curry_case_prod, simplified curry_conv[abs_def] fun_ord_def split_paired_All prod.case case_prod_eta, OF refl]) blast+

lemma inline1_fixp_induct_strong2 [case_names adm bottom step]:
  assumes "ccpo.admissible (fun_lub lub_spmf) (fun_ord (ord_spmf op =)) (\<lambda>inline1'. P (\<lambda>gpv s. inline1' (gpv, s)))"
  and "P (\<lambda>_ _. return_pmf None)"
  and "\<And>inline1'. 
    \<lbrakk> \<And>gpv s. ord_spmf op = (inline1' gpv s) (inline1 gpv s); 
      \<And>gpv s. ord_spmf op = (inline1' gpv s) (the_gpv gpv \<bind> case_generat (\<lambda>x. return_spmf (Inl (x, s))) (\<lambda>out rpv. the_gpv (callee s out) \<bind> case_generat (\<lambda>(x, y). inline1' (rpv x) y) (\<lambda>out rpv'. return_spmf (Inr (out, rpv', rpv)))));
      P inline1' \<rbrakk>
    \<Longrightarrow> P (\<lambda>gpv s. the_gpv gpv \<bind> case_generat (\<lambda>x. return_spmf (Inl (x, s))) (\<lambda>out rpv. the_gpv (callee s out) \<bind> case_generat (\<lambda>(x, y). inline1' (rpv x) y) (\<lambda>out rpv'. return_spmf (Inr (out, rpv', rpv)))))"
  shows "P inline1"
using assms
by(rule spmf.fixp_induct_strong2_uc[where P="\<lambda>f. P (curry f)" and U=case_prod and C=curry, OF inline1.mono inline1_def, simplified curry_case_prod, simplified curry_conv[abs_def] fun_ord_def split_paired_All prod.case case_prod_eta, OF refl]) blast+

lemma inline1_Fail [simp]: "inline1 Fail s = return_pmf None"
by(rewrite inline1.simps) simp

lemma inline1_Done [simp]: "inline1 (Done x) s = return_spmf (Inl (x, s))"
by(rewrite inline1.simps) simp

lemma inline1_lift_spmf [simp]: "inline1 (lift_spmf p) s = map_spmf (\<lambda>x. Inl (x, s)) p"
by(rewrite inline1.simps)(simp add: bind_map_spmf o_def map_spmf_conv_bind_spmf)

lemma inline1_Pause:
  "inline1 (Pause out c) s = 
  the_gpv (callee s out) \<bind> (\<lambda>react. case react of Pure (x, s') \<Rightarrow> inline1 (c x) s' | IO out' c' \<Rightarrow> return_spmf (Inr (out', c', c)))"
by(rewrite inline1.simps) simp

lemma inline1_bind_gpv:
  fixes gpv f s
  defines [simp]: "inline11 \<equiv> inline1" and [simp]: "inline12 \<equiv> inline1" and [simp]: "inline13 \<equiv> inline1"
  shows "inline11 (bind_gpv gpv f) s = bind_spmf (inline12 gpv s) 
    (\<lambda>res. case res of Inl (x, s') \<Rightarrow> inline13 (f x) s' | Inr (out, rpv', rpv) \<Rightarrow> return_spmf (Inr (out, rpv', bind_rpv rpv f)))"
  (is "?lhs = ?rhs")
proof(rule spmf.leq_antisym)
  note [intro!] = ord_spmf_bind_reflI and [split] = generat.split
  show "ord_spmf op = ?lhs ?rhs" unfolding inline11_def
  proof(induction arbitrary: gpv s f rule: inline1_fixp_induct)
    case adm show ?case by simp
    case bottom show ?case by simp
    case (step inline1')
    show ?case unfolding inline12_def
      apply(rewrite inline1.simps; clarsimp simp add: bind_rpv_def)
      apply(rule conjI; clarsimp)
      subgoal premises Pure for x
        apply(rewrite inline1.simps; clarsimp)
        subgoal for out c ret s' using step.IH[of "Done x" "\<lambda>_. c ret" s'] by simp
        done
      subgoal for out c ret s' using step.IH[of "c ret" f s'] by(simp cong del: sum.case_cong_weak)
      done
  qed
  show "ord_spmf op = ?rhs ?lhs" unfolding inline12_def
  proof(induction arbitrary: gpv s rule: inline1_fixp_induct)
    case adm show ?case by simp
    case bottom show ?case by simp
    case (step inline1')
    show ?case unfolding inline11_def
      apply(rewrite inline1.simps; clarsimp simp add: bind_rpv_def)
      apply(rule conjI; clarsimp)
      subgoal by(rewrite inline1.simps; simp)
      subgoal for out c ret s' using step.IH[of "c ret" s'] by(simp cong del: sum.case_cong_weak)
      done
  qed
qed

end

declare [[bnf_internals]]
context notes [[quick_and_dirty]] begin (* TODO: remove when codatatype is fixed *)
codatatype ('a, 'call, 'ret, 'x) gpv' = GPV' (the_gpv': "('a, 'call, 'ret \<Rightarrow> ('a, 'call, 'ret, 'x) gpv') generat spmf")
end

primcorec gpv_of_gpv' :: "('a, 'call, 'ret, 'x) gpv' \<Rightarrow> ('a, 'call, 'ret) gpv"
where "the_gpv (gpv_of_gpv' gpv') = map_spmf (map_generat id id (op \<circ> gpv_of_gpv')) (the_gpv' gpv')"
primcorec gpv'_of_gpv :: "('a, 'call, 'ret) gpv \<Rightarrow> ('a, 'call, 'ret, 'x) gpv'"
where "the_gpv' (gpv'_of_gpv gpv') = map_spmf (map_generat id id (op \<circ> gpv'_of_gpv)) (the_gpv gpv')"

lemma gpv_of_gpv'_inverse [simp]: "gpv'_of_gpv (gpv_of_gpv' gpv') = gpv'"
apply(coinduction arbitrary: gpv')
apply(auto simp add: spmf_rel_map generat.rel_map rel_fun_def intro!: rel_spmf_reflI generat.rel_refl)
done

lemma gpv'_of_gpv_inverse [simp]: "gpv_of_gpv' (gpv'_of_gpv gpv) = gpv"
apply(coinduction arbitrary: gpv)
apply(auto simp add: spmf_rel_map generat.rel_map rel_fun_def intro!: rel_spmf_reflI generat.rel_refl)
done

lemma td_gpv': "type_definition gpv_of_gpv' gpv'_of_gpv UNIV"
by unfold_locales simp_all

setup_lifting td_gpv'

lift_definition bind_gpv' :: "('a, 'call, 'ret) gpv \<Rightarrow> ('a \<Rightarrow> ('b, 'call, 'ret, 'a) gpv') \<Rightarrow> ('b, 'call, 'ret, 'a) gpv'" is bind_gpv .

thm bind_gpv'.rep_eq

lemma gpv_of_gpv'_GPV' [simp]: "gpv_of_gpv' (GPV' p) = GPV (map_spmf (map_generat id id (op \<circ> gpv_of_gpv')) p)"
by(rule gpv.expand) simp

lemma gpv'_of_gpv_GPV [simp]: "gpv'_of_gpv (GPV p) = GPV' (map_spmf (map_generat id id (op \<circ> gpv'_of_gpv)) p)"
by(rule gpv'.expand) simp

abbreviation cast_gpv' :: "('a, 'call, 'ret, 'b) gpv' \<Rightarrow> ('a, 'call, 'ret, 'c) gpv'"
where "cast_gpv' gpv' \<equiv> gpv'_of_gpv (gpv_of_gpv' gpv')"

context includes lifting_syntax begin
lemma GPV'_transfer [transfer_rule]:
  "(rel_spmf (rel_generat op = op = (op = ===> cr_gpv')) ===> cr_gpv') GPV GPV'"
by(simp add: cr_gpv'_def gpv.rel_eq[symmetric] rel_fun_def spmf_rel_map generat.rel_map)

lemma the_gpv'_transfer [transfer_rule]:
  "(cr_gpv' ===> rel_spmf (rel_generat op = op = (op = ===> cr_gpv'))) the_gpv the_gpv'"
by(auto simp add: cr_gpv'_def spmf_rel_map generat.rel_map rel_fun_def intro!: rel_spmf_reflI generat.rel_refl)
end


lemma bind_gpv'_simps:
  "bind_gpv' r f =
   GPV' (map_spmf (map_generat id id (op \<circ> (case_sum id (\<lambda>r. bind_gpv' r f))))
           (bind_spmf (the_gpv r)
             (case_generat
               (\<lambda>x. map_spmf (map_generat id id (op \<circ> Inl)) (the_gpv' (f x)))
               (\<lambda>out c. return_spmf (IO out (\<lambda>input. Inr (c input)))))))"
by transfer(rule bind_gpv.ctr)

lemma bind_gpv'_sel:
  "the_gpv' (bind_gpv' r f) =
   map_spmf (map_generat id id (\<lambda>rpv x. case_sum id (\<lambda>r. bind_gpv' r f) (rpv x)))
           (bind_spmf (the_gpv r)
             (case_generat
               (\<lambda>x. map_spmf (map_generat id id (\<lambda>r x. Inl (r x))) (the_gpv' (f x)))
               (\<lambda>out c. return_spmf (IO out (\<lambda>input. Inr (c input))))))"
by(subst bind_gpv'_simps)(simp add: o_def[abs_def])

lemma gpv_of_gpv'_bind_gpv' [simp]: "gpv_of_gpv' (bind_gpv' gpv f) = bind_gpv gpv (gpv_of_gpv' \<circ> f)"
by transfer(simp add: o_def)

lemma gpv'_of_gpv_bind_gpv [simp]: "gpv'_of_gpv (bind_gpv gpv f) = bind_gpv' gpv (gpv'_of_gpv \<circ> f)"
by transfer (simp add: o_def)

friend_of_corec bind_gpv'
where "bind_gpv' r f =
GPV' (map_spmf (map_generat id id (op \<circ> (case_sum id (\<lambda>r. bind_gpv' r f))))
      (bind_spmf (the_gpv r)
        (case_generat (\<lambda>x. map_spmf (map_generat id id (op \<circ> Inl)) (the_gpv' (f x)))
          (\<lambda>out c. return_spmf (IO out (\<lambda>input. Inr (c input)))))))"
subgoal by(rule bind_gpv'_simps)
apply(fold relator_eq)
apply transfer_prover+
done

lift_definition inline1'
  :: "('s \<Rightarrow> 'call \<Rightarrow> ('ret \<times> 's, 'call', 'ret') gpv) \<Rightarrow> ('a, 'call, 'ret, 'ret \<times> 's) gpv' \<Rightarrow> 's
  \<Rightarrow> ('a \<times> 's + 'call' \<times> ('ret \<times> 's, 'call', 'ret') rpv \<times> ('ret \<Rightarrow> ('a, 'call, 'ret, 'ret \<times> 's) gpv')) spmf"
is inline1 .

lemma inline1_conv_inline1':
  "inline1 oracle gpv' s = map_spmf (map_sum id (map_prod id (map_prod id (map_fun id gpv_of_gpv')))) (inline1' oracle (gpv'_of_gpv gpv') s)"
by transfer(simp add: id_def[symmetric] map_fun_id prod.map_id0 sum.map_id0)

corec inline' :: "('s \<Rightarrow> 'call \<Rightarrow> ('ret \<times> 's, 'call', 'ret') gpv) \<Rightarrow> ('a, 'call, 'ret, 'ret \<times> 's) gpv' \<Rightarrow> 's \<Rightarrow> ('a \<times> 's, 'call', 'ret', 'ret \<times> 's) gpv'"
where
  "inline' oracle c s = GPV' (
   map_spmf (\<lambda>result. case result of Inl xs \<Rightarrow> Pure xs
                       | Inr (out, oracle', rpv) \<Rightarrow> IO out (\<lambda>input. bind_gpv' (oracle' input) (\<lambda>(x, s'). inline' oracle (rpv x) s'))) (inline1' oracle c s))"

lemma inline'_sel:
  "the_gpv' (inline' oracle c s) = 
   map_spmf (\<lambda>result. case result of Inl xs \<Rightarrow> Pure xs
                       | Inr (out, oracle', rpv) \<Rightarrow> IO out (\<lambda>input. bind_gpv' (oracle' input) (\<lambda>(x, s'). inline' oracle (rpv x) s'))) (inline1' oracle c s)"
by(subst inline'.code) simp

definition inline :: "('s \<Rightarrow> 'call \<Rightarrow> ('ret \<times> 's, 'call', 'ret') gpv) \<Rightarrow> ('a, 'call, 'ret) gpv \<Rightarrow> 's \<Rightarrow> ('a \<times> 's, 'call', 'ret') gpv"
where "inline oracle c s = gpv_of_gpv' (inline' oracle (gpv'_of_gpv c) s)"

lemma inline_sel:
  "the_gpv (inline oracle c s) = 
   map_spmf (\<lambda>result. case result of Inl xs \<Rightarrow> Pure xs
                       | Inr (out, oracle', rpv) \<Rightarrow> IO out (\<lambda>input. bind_gpv (oracle' input) (\<lambda>(x, s'). inline oracle (rpv x) s'))) (inline1 oracle c s)"
by(auto simp add: inline_def inline'_sel spmf.map_comp o_def inline1_conv_inline1' split_def intro!: arg_cong2[where f=map_spmf] ext split: sum.split)

lemma inline_Fail [simp]: "inline oracle Fail s = Fail"
by(rule gpv.expand)(simp add: inline_sel)

lemma inline_Done [simp]: "inline oracle (Done x) s = Done (x, s)"
by(rule gpv.expand)(simp add: inline_sel)

lemma inline1'_bind_gpv:
  "inline1' callee (gpv'_of_gpv (bind_gpv gpv f)) s = 
   bind_spmf (inline1 callee gpv s) (\<lambda>res. case res of Inl (x, s') \<Rightarrow> inline1' callee (gpv'_of_gpv (f x)) s' | Inr (out, rpv', rpv) \<Rightarrow> return_spmf (Inr (out, rpv', \<lambda>input. gpv'_of_gpv (bind_gpv (rpv input) f))))"
by transfer(simp add: inline1_bind_gpv bind_rpv_def cong del: sum.case_cong)

lemma inline'_bind_gpv [simp]:
  "inline' oracle (gpv'_of_gpv (bind_gpv gpv f)) s = cast_gpv' (bind_gpv' (inline oracle gpv s) (\<lambda>(x, s'). cast_gpv' (inline' oracle (gpv'_of_gpv (f x)) s')))"
by(coinduction arbitrary: gpv s rule: inline'.coinduct)
  (auto 4 4 simp add: bind_gpv_assoc inline'_sel inline_sel inline1_conv_inline1' bind_gpv'_sel spmf_rel_map spmf.map_comp o_def generat.map_comp map_spmf_bind_spmf inline1'_bind_gpv bind_map_spmf intro!: rel_spmf_bind_reflI rel_spmf_reflI rel_funI gpv'.cong_bind_gpv'[OF refl] gpv'.cong_refl[OF refl] split: sum.split intro: gpv'.cong_base)

lemma inline_bind_gpv [simp]:
  "inline oracle (bind_gpv gpv f) s = bind_gpv (inline oracle gpv s) (\<lambda>(x, s'). inline oracle  (f x) s')"
unfolding inline_def inline'_bind_gpv by(simp add: o_def split_def)

lemma set_inline1_lift_spmf1: "set_spmf (inline1 (\<lambda>s x. lift_spmf (p s x)) gpv s) \<subseteq> range Inl"
apply(induction arbitrary: gpv s rule: inline1_fixp_induct)
subgoal by(rule cont_intro ccpo_class.admissible_leI)+
apply(auto simp add: set_bind_spmf o_def bind_UNION split: generat.split_asm)+
done

lemma in_set_inline1_lift_spmf1: "y \<in> set_spmf (inline1 (\<lambda>s x. lift_spmf (p s x)) gpv s) \<Longrightarrow> \<exists>r s'. y = Inl (r, s')"
by(drule set_inline1_lift_spmf1[THEN subsetD]) auto

lemma inline_lift_spmf1:
  fixes p defines "callee \<equiv> \<lambda>s c. lift_spmf (p s c)"
  shows "inline callee gpv s = lift_spmf (map_spmf projl (inline1 callee gpv s))"
by(rule gpv.expand)(auto simp add: inline_sel spmf.map_comp callee_def intro!: map_spmf_cong[OF refl] dest: in_set_inline1_lift_spmf1)

context includes lifting_syntax begin
lemma inline'_transfer [transfer_rule]:
  "(op = ===> cr_gpv' ===> op = ===> cr_gpv') inline inline'"
by(simp add: rel_fun_def cr_gpv'_def inline_def)

lemma inline1_parametric [transfer_rule]:
  "((S ===> C ===> rel_gpv (rel_prod op = S) C') ===> rel_gpv A C ===> S
   ===> rel_spmf (rel_sum (rel_prod A S) (rel_prod C' (rel_prod (rel_rpv (rel_prod op = S) C') (rel_rpv A C)))))
  inline1 inline1"
  (is "(_ ===> ?R) _ _")
proof(rule rel_funI)
  show "?R (inline1 callee) (inline1 callee')" 
    if [transfer_rule]: "(S ===> C ===> rel_gpv (rel_prod op = S) C') callee callee'"
    for callee callee'
    unfolding inline1_def
    by(unfold rel_fun_curry case_prod_curry)(rule fixp_spmf_parametric[OF inline1.mono inline1.mono]; transfer_prover)
qed

lemma rel_gpv'_transfer [transfer_rule]: "(op = ===> op = ===> R ===> cr_gpv' ===> cr_gpv' ===> op =) (\<lambda>A B _. rel_gpv A B) rel_gpv'"
apply(intro rel_funI; clarsimp simp add: cr_gpv'_def; rule iffI)
subgoal for A C r r' gpv gpv'
  apply(coinduction arbitrary: gpv gpv')
  subgoal for gpv gpv' by(cases gpv gpv' rule: gpv'.exhaust[case_product gpv'.exhaust])(auto simp add: spmf_rel_map generat.rel_map rel_fun_def elim!: rel_spmf_mono rel_generat_mono)
  done
subgoal for A C r r' gpv gpv'
  by(coinduction arbitrary: gpv gpv')(auto 4 4 elim!: gpv'.rel_cases simp add: spmf_rel_map generat.rel_map rel_fun_def elim!: rel_spmf_mono rel_generat_mono)
done

lemma gpv_of_gpv'_parametric [transfer_rule]: "(rel_gpv' A C R ===> rel_gpv A C) gpv_of_gpv' gpv_of_gpv'"
by(rule rel_funI) transfer

lemma gpv'_of_gpv_parametric [transfer_rule]: "(rel_gpv A C ===> rel_gpv' A C R) gpv'_of_gpv gpv'_of_gpv"
by(rule rel_funI) transfer

lifting_update gpv'.lifting
lifting_forget gpv'.lifting

lemma gpv'_of_gpv_map_gpv:
  "gpv'_of_gpv (map_gpv f g gpv) = map_gpv' f g h (gpv'_of_gpv gpv)"
unfolding gpv'.rel_eq[symmetric] gpv'.rel_map
by(rule gpv'_of_gpv_parametric[THEN rel_funD])(simp add: gpv.rel_map gpv.rel_refl)

lemma gpv_of_gpv'_map_gpv':
  "gpv_of_gpv' (map_gpv' f g h gpv') = map_gpv f g (gpv_of_gpv' gpv')"
unfolding gpv.rel_eq[symmetric] gpv.rel_map
by(rule gpv_of_gpv'_parametric[where R="\<lambda>_ _. True", THEN rel_funD])(simp add: gpv'.rel_map gpv'.rel_refl)

lemma inline1'_parametric [transfer_rule]:
  "((S ===> C ===> rel_gpv (rel_prod op = S) C') ===> rel_gpv' A C R ===> S
   ===> rel_spmf (rel_sum (rel_prod A S) (rel_prod C' (rel_prod (rel_rpv (rel_prod op = S) C') (op = ===> rel_gpv' A C R')))))
  inline1' inline1'"
unfolding inline1'_def map_fun_def o_def by transfer_prover

declare gpv'.rel_eq [relator_eq]
declare gpv'.dtor_unfold_transfer[transfer_rule]

lemma bind_gpv'_parametric [transfer_rule]:
  "(rel_gpv A C ===> (A ===> rel_gpv' B C R) ===> rel_gpv' B C R') (\<lambda>gpv f. cast_gpv' (bind_gpv' gpv f)) bind_gpv'"
including gpv'.lifting unfolding rel_fun_def
by transfer(rule bind_gpv_parametric[unfolded rel_fun_def])

lemma map_gpv'_id_bind_gpv':
  "map_gpv' f id h (bind_gpv' gpv g) = cast_gpv' (bind_gpv' gpv (map_gpv' f id id \<circ> g))"
apply(rule sym)
apply(unfold gpv'.rel_eq[symmetric] gpv'.rel_map)
apply(rule bind_gpv'_parametric[where A="op =" and R="op =", THEN rel_funD, THEN rel_funD])
apply(simp_all add: gpv'.rel_map gpv'.rel_refl rel_fun_def gpv.rel_refl)
done

end

context
  fixes callee1 :: "'s1 \<Rightarrow> 'c1 \<Rightarrow> ('r1 \<times> 's1, 'c, 'r) gpv"
  and callee2 :: "'s2 \<Rightarrow> 'c2 \<Rightarrow> ('r2 \<times> 's2, 'c1, 'r1) gpv"
begin

partial_function (spmf) inline2 :: "('a, 'c2, 'r2) gpv \<Rightarrow> 's2 \<Rightarrow> 's1
  \<Rightarrow> ('a \<times> ('s2 \<times> 's1) + 'c \<times> ('r1 \<times> 's1, 'c, 'r) rpv \<times> ('r2 \<times> 's2, 'c1, 'r1) rpv \<times> ('a, 'c2, 'r2) rpv) spmf"
where
  "inline2 gpv s2 s1 =
  bind_spmf (the_gpv gpv)
   (case_generat (\<lambda>x. return_spmf (Inl (x, s2, s1)))
     (\<lambda>out rpv. bind_spmf (inline1 callee1 (callee2 s2 out) s1)
       (case_sum (\<lambda>((r2, s2), s1). inline2 (rpv r2) s2 s1)
         (\<lambda>(x, rpv'', rpv'). return_spmf (Inr (x, rpv'', rpv', rpv))))))"

lemma inline2_fixp_induct [case_names adm bottom step]:
  assumes "ccpo.admissible (fun_lub lub_spmf) (fun_ord (ord_spmf op =)) (\<lambda>inline2. P (\<lambda>gpv s2 s1. inline2 ((gpv, s2), s1)))"
  and "P (\<lambda>_ _ _. return_pmf None)"
  and "\<And>inline2'. P inline2' \<Longrightarrow>
       P (\<lambda>gpv s2 s1. bind_spmf (the_gpv gpv) (\<lambda>generat. case generat of
           Pure x \<Rightarrow> return_spmf (Inl (x, s2, s1))
         | IO out rpv \<Rightarrow> bind_spmf (inline1 callee1 (callee2 s2 out) s1) (\<lambda>lr. case lr of
              Inl ((r2, s2), c) \<Rightarrow> inline2' (rpv r2) s2 c
           | Inr (x, rpv'', rpv') \<Rightarrow> return_spmf (Inr (x, rpv'', rpv', rpv)))))"
  shows "P inline2"
using assms unfolding split_def by(rule inline2.fixp_induct[unfolded curry_conv[abs_def] split_def])

lemma inline1_inline_conv_inline2:
  fixes gpv' :: "('r2 \<times> 's2, 'c1, 'r1) gpv"
  shows "inline1 callee1 (inline callee2 gpv s2) s1 = 
  map_spmf (map_sum (\<lambda>(x, (s2, s1)). ((x, s2), s1))
    (\<lambda>(x, rpv'', rpv', rpv). (x, rpv'', \<lambda>r1. rpv' r1 \<bind> (\<lambda>(r2, s2). inline callee2 (rpv r2) s2))))
  (inline2 gpv s2 s1)"
  (is "?lhs = ?rhs")
proof(rule spmf.leq_antisym)
  def inline1_1 \<equiv> "inline1 :: ('s1 \<Rightarrow> 'c1 \<Rightarrow> ('r1 \<times> 's1, 'c, 'r) gpv) \<Rightarrow> ('r2 \<times> 's2, 'c1, 'r1) gpv \<Rightarrow> 's1 \<Rightarrow> _"
  have "ord_spmf op = ?lhs ?rhs"
    -- \<open> We need in the inductive step that the approximation behaves well with @{const bind_gpv}
         because the corecursion goes through @{const bind_gpv}. So we have to thread it through the induction
         and do one half of the proof from @{thm [source] inline1_bind_gpv} again. We cannot inline
         @{thm [source] inline1_bind_gpv} in this proof here because the types are too specific.
       \<close>
    and "ord_spmf op = (inline1 callee1 (gpv' \<bind> f) s1') 
      (do {
      res \<leftarrow> inline1_1 callee1 gpv' s1';
      case res of Inl (x, s') \<Rightarrow> inline1 callee1 (f x) s'
      | Inr (out, rpv', rpv) \<Rightarrow> return_spmf (Inr (out, rpv', rpv \<bind> f))
    })" for gpv' and f :: "_ \<Rightarrow> ('a \<times> 's2, 'c1, 'r1) gpv" and s1'
  proof(induction arbitrary: gpv s2 s1 gpv' f s1' rule: inline1_fixp_induct_strong2)
    case adm thus ?case
      apply(rule cont_intro)+
      subgoal for a b c d by(cases d; clarsimp)
      done

    case (step inline1')
    note step_IH = step.IH[unfolded inline1_1_def] and step_hyps = step.hyps[unfolded inline1_1_def]
    { case 1
      have inline1: "ord_spmf op =
         (inline1 callee2 gpv s2 \<bind> (\<lambda>lr. case lr of Inl as2 \<Rightarrow> return_spmf (Inl (as2, s1))
            | Inr (out1, rpv', rpv) \<Rightarrow> the_gpv (callee1 s1 out1) \<bind> (\<lambda>generat. case generat of
                Pure (r1, s1) \<Rightarrow> inline1' (bind_gpv (rpv' r1) (\<lambda>(r2, s2). inline callee2 (rpv r2) s2)) s1
              | IO out rpv'' \<Rightarrow> return_spmf (Inr (out, rpv'', \<lambda>r1. bind_gpv (rpv' r1) (\<lambda>(r2, s2). inline callee2 (rpv r2) s2)) ))))
         (the_gpv gpv \<bind> (\<lambda>generat. case generat of Pure x \<Rightarrow> return_spmf (Inl ((x, s2), s1))
            | IO out2 rpv \<Rightarrow> inline1_1 callee1 (callee2 s2 out2) s1 \<bind> (\<lambda>lr. case lr of
                Inl ((r2, s2), s1) \<Rightarrow>
                   map_spmf (map_sum (\<lambda>(x, s2, s1). ((x, s2), s1)) (\<lambda>(x, rpv'', rpv', rpv). (x, rpv'', \<lambda>r1. bind_gpv (rpv' r1) (\<lambda>(r2, s2). inline callee2 (rpv r2) s2))))
                     (inline2 (rpv r2) s2 s1)
              | Inr (out, rpv'', rpv') \<Rightarrow>
                   return_spmf (Inr (out, rpv'', \<lambda>r1. bind_gpv (rpv' r1) (\<lambda>(r2, s2). inline callee2 (rpv r2) s2))))))"
      proof(induction arbitrary: gpv s2 s1 rule: inline1_fixp_induct)
        case step2: (step inline1'')
        note step2_IH = step2.IH[unfolded inline1_1_def]

        show ?case unfolding inline1_1_def
          apply(rewrite in "ord_spmf _ _ \<hole>" inline1.simps)
          apply(clarsimp intro!: ord_spmf_bind_reflI split: generat.split)
          apply(rule conjI)
          subgoal by(rewrite in "ord_spmf _ _ \<hole>" inline2.simps)(clarsimp simp add: map_spmf_bind_spmf o_def split: generat.split sum.split intro!: ord_spmf_bind_reflI spmf.leq_trans[OF step2_IH])
          subgoal by(clarsimp intro!: ord_spmf_bind_reflI step_IH[THEN spmf.leq_trans] split: generat.split sum.split simp add: bind_rpv_def)
          done
      qed simp_all
      show ?case
        apply(rewrite in "ord_spmf _ \<hole> _" inline_sel)
        apply(rewrite in "ord_spmf _ _ \<hole>" inline2.simps)
        apply(clarsimp simp add: map_spmf_bind_spmf bind_map_spmf o_def intro!: ord_spmf_bind_reflI split: generat.split)
        apply(rule spmf.leq_trans[OF spmf.leq_trans, OF _ inline1])
        apply(auto intro!: ord_spmf_bind_reflI split: sum.split generat.split simp add: inline1_1_def map_spmf_bind_spmf)
        done }
    { case 2
      show ?case unfolding inline1_1_def
        by(rewrite inline1.simps)(auto simp del: bind_gpv_sel' simp add: bind_gpv.sel map_spmf_bind_spmf bind_map_spmf o_def bind_rpv_def intro!: ord_spmf_bind_reflI step_IH(2)[THEN spmf.leq_trans] step_hyps(2) split: generat.split sum.split) }
  qed simp_all
  thus "ord_spmf op = ?lhs ?rhs" by -

  show "ord_spmf op = ?rhs ?lhs"
  proof(induction arbitrary: gpv s2 s1 rule: inline2_fixp_induct)
    case adm show ?case by simp
    case bottom show ?case by simp
    case (step inline2')
    show ?case
      apply(rewrite in "ord_spmf _ _ \<hole>" inline1.simps)
      apply(rewrite inline_sel)
      apply(rewrite in "ord_spmf _ \<hole> _" inline1.simps)
      apply(rewrite in "ord_spmf _ _ \<hole>" inline1.simps)
      apply(clarsimp simp add: map_spmf_bind_spmf bind_map_spmf intro!: ord_spmf_bind_reflI split: generat.split)
      apply(rule conjI)
      subgoal
         apply clarsimp
         apply(rule step.IH[THEN spmf.leq_trans])
         apply(rewrite in "ord_spmf _ \<hole> _" inline1.simps)
         apply(rewrite inline_sel)
         apply(simp add: bind_map_spmf)
         done
      subgoal by(clarsimp intro!: ord_spmf_bind_reflI split: generat.split sum.split simp add: o_def inline1_bind_gpv bind_rpv_def step.IH)
      done
  qed
qed

lemma inline1_inline_conv_inline2':
  "inline1 (\<lambda>(s2, s1) c2. map_gpv (\<lambda>((r, s2), s1). (r, s2, s1)) id (inline callee1 (callee2 s2 c2) s1)) gpv (s2, s1) =
   map_spmf (map_sum id (\<lambda>(x, rpv'', rpv', rpv). (x, \<lambda>r. bind_gpv (rpv'' r)
       (\<lambda>(r1, s1). map_gpv (\<lambda>((r2, s2), s1). (r2, s2, s1)) id (inline callee1 (rpv' r1) s1)), rpv)))
     (inline2 gpv s2 s1)"
  (is "?lhs = ?rhs")
proof(rule spmf.leq_antisym)
  show "ord_spmf op = ?lhs ?rhs"
  proof(induction arbitrary: gpv s2 s1 rule: inline1_fixp_induct)
    case (step inline1') show ?case
      by(rewrite inline2.simps)(auto simp add: map_spmf_bind_spmf o_def inline_sel gpv.map_sel bind_map_spmf id_def[symmetric] gpv.map_id map_gpv_bind_gpv split_def intro!: ord_spmf_bind_reflI step.IH[THEN spmf.leq_trans] split: generat.split sum.split)
  qed simp_all
  show "ord_spmf op = ?rhs ?lhs"
  proof(induction arbitrary: gpv s2 s1 rule: inline2_fixp_induct)
   case (step inline2')
   show ?case
     apply(rewrite in "ord_spmf _ _ \<hole>" inline1.simps)
     apply(clarsimp simp add: map_spmf_bind_spmf bind_rpv_def o_def gpv.map_sel bind_map_spmf inline_sel map_gpv_bind_gpv id_def[symmetric] gpv.map_id split_def split: generat.split sum.split intro!: ord_spmf_bind_reflI)
     apply(rule spmf.leq_trans[OF spmf.leq_trans, OF _ step.IH])
     apply(auto simp add: split_def id_def[symmetric] intro!: ord_spmf_reflI)
     done
  qed simp_all
qed

context includes gpv'.lifting begin

lift_definition inline2' :: "('a, 'c2, 'r2) gpv \<Rightarrow> 's2 \<Rightarrow> 's1
  \<Rightarrow> ('a \<times> ('s2 \<times> 's1) + 'c \<times> ('r1 \<times> 's1, 'c, 'r) rpv \<times> ('r2 \<times> 's2, 'c1, 'r1) rpv \<times> ('r2 \<Rightarrow> ('a, 'c2, 'r2, 'foo) gpv')) spmf"
is "inline2" .

lemma inline1'_inline_conv_inline2: 
  "inline1' callee1 (gpv'_of_gpv (inline callee2 gpv s2)) s1 = 
   map_spmf (map_sum (\<lambda>(x, (s2, s1)). ((x, s2), s1))
     (\<lambda>(x, rpv'', rpv', rpv). (x, rpv'', \<lambda>r1. gpv'_of_gpv (bind_gpv (rpv' r1) (\<lambda>(r2, s2). inline callee2 (gpv_of_gpv' (rpv r2)) s2)))))
   (inline2' gpv s2 s1)"
by transfer(rule inline1_inline_conv_inline2)

lemma inline1'_inline_conv_inline2':
  "inline1' (\<lambda>(s2, s1) c2. map_gpv (\<lambda>((r, s2), s1). (r, s2, s1)) id (inline callee1 (callee2 s2 c2) s1)) (gpv'_of_gpv gpv) (s2, s1) =
   map_spmf (map_sum id (\<lambda>(x, rpv'', rpv', rpv). (x, \<lambda>r. bind_gpv (rpv'' r)
       (\<lambda>(r1, s1). map_gpv (\<lambda>((r2, s2), s1). (r2, s2, s1)) id (inline callee1 (rpv' r1) s1)), rpv)))
     (inline2' gpv s2 s1)"
by transfer(rule inline1_inline_conv_inline2')

lemma bind_gpv'_assoc:
  "bind_gpv' (bind_gpv gpv f) g = cast_gpv' (bind_gpv' gpv (\<lambda>x. cast_gpv' (bind_gpv' (f x) g)))"
by transfer(rule bind_gpv_assoc)

end

lemma inline'_assoc:
  "cast_gpv' (inline' callee1 (gpv'_of_gpv (inline callee2 gpv s2)) s1) =
   map_gpv' (\<lambda>(r, s2, s1). ((r, s2), s1)) id id (inline' (\<lambda>(s2, s1) c2. map_gpv (\<lambda>((r, s2), s1). (r, s2, s1)) id (inline callee1 (callee2 s2 c2) s1)) (gpv'_of_gpv gpv) (s2, s1))"
apply(coinduction arbitrary: s2 s1 gpv rule: gpv'.coinduct_upto)
apply simp
apply(subst inline'_sel)
apply(subst gpv'.map_sel)
apply(subst inline'_sel)
apply(subst inline1'_inline_conv_inline2)
apply(subst inline1'_inline_conv_inline2')
apply(unfold spmf.map_comp o_def case_sum_map_sum spmf_rel_map generat.rel_map)
apply(rule rel_spmf_reflI)
apply(split sum.split)
apply(clarsimp simp add: split_beta bind_gpv'_assoc gpv_of_gpv'_map_gpv' o_def rel_fun_def intro!: gpv'.cong_bind_gpv'[OF refl])
apply(subst bind_gpv_assoc[symmetric])
apply(subst bind_gpv_assoc[symmetric])
apply(simp add: split_beta)
apply(subst map_gpv'_id_bind_gpv')
apply(simp add: o_def)
-- \<open>We have to reassociate the return tuple such that type types for @{term bind_gpv} are right\<close>
apply(subgoal_tac "bind_gpv (aa xa) (\<lambda>x. inline callee1 (ab (fst x)) (snd x)) = 
  map_gpv (\<lambda>(r2, s2, s1). ((r2, s2), s1)) id (map_gpv (\<lambda>((r2, s2), s1). (r2, s2, s1)) id (bind_gpv (aa xa) (\<lambda>x. inline callee1 (ab (fst x)) (snd x))))")
 apply(erule forw_subst[where P="\<lambda>x. gpv'.v1.congclp _ (gpv'_of_gpv (bind_gpv x _)) _"])
 apply(subst bind_map_gpv)
 apply(subst map_gpv_id_bind_gpv)
 apply(subst gpv'_of_gpv_bind_gpv)
 apply(rule gpv'.cong_bind_gpv')
  apply(simp add: o_def)
 apply(clarsimp simp add: rel_fun_def)
 apply(rule gpv'.cong_base)
 apply(rule exI conjI)+
 apply(rule refl)
 apply simp
apply(simp add: gpv.map_comp o_def split_def id_def[symmetric] gpv.map_id)
done

lemma inline_assoc:
  "inline callee1 (inline callee2 gpv s2) s1 =
   map_gpv (\<lambda>(r, s2, s1). ((r, s2), s1)) id (inline (\<lambda>(s2, s1) c2. map_gpv (\<lambda>((r, s2), s1). (r, s2, s1)) id (inline callee1 (callee2 s2 c2) s1)) gpv (s2, s1))"
apply(subst (2 4) inline_def)
apply(subst gpv_of_gpv'_map_gpv'[symmetric])
apply(subst inline'_assoc[symmetric])
apply simp
done

end

lemma set_inline2_lift_spmf1: "set_spmf (inline2 (\<lambda>s x. lift_spmf (p s x)) callee gpv s s') \<subseteq> range Inl"
apply(induction arbitrary: gpv s s' rule: inline2_fixp_induct)
subgoal by(rule cont_intro ccpo_class.admissible_leI)+
apply(auto simp add: set_bind_spmf o_def bind_UNION split: generat.split_asm sum.split_asm dest!: in_set_inline1_lift_spmf1)
apply blast
done

lemma in_set_inline2_lift_spmf1: "y \<in> set_spmf (inline2 (\<lambda>s x. lift_spmf (p s x)) callee gpv s s') \<Longrightarrow> \<exists>r s s'. y = Inl (r, s, s')"
by(drule set_inline2_lift_spmf1[THEN subsetD]) auto

lemma interaction_bound_inline1':
  assumes "(\<And>s x. interaction_bound (callee s x) \<le> q)"
  shows "interaction_bound gpv \<le> p
  \<Longrightarrow> set_spmf (inline1 callee gpv s) \<subseteq> {Inr (out, c', rpv) | out c' rpv. 
        (\<forall>input. eSuc (interaction_bound (c' input)) \<le> q) \<and> 
        (\<forall>x. eSuc (interaction_bound (rpv x)) \<le> p)}
      \<union> range Inl"
proof(induction arbitrary: gpv s rule: inline1_fixp_induct)
  { case adm show ?case by(rule cont_intro ccpo_class.admissible_leI)+ }
  { case bottom show ?case by simp }
  case (step inline1')
  have *: "\<And>out c input. IO out c \<in> set_spmf (the_gpv gpv) \<Longrightarrow> eSuc (interaction_bound (c input)) \<le> p"
    by(erule interaction_bound_IO[THEN order_trans])(rule step.prems)
  have **: "\<And>out c s x input. IO out c \<in> set_spmf (the_gpv (callee s x)) \<Longrightarrow> eSuc (interaction_bound (c input)) \<le> q"
    using interaction_bound_IO assms by(rule order_trans)
  show ?case
    by(auto 6 4 simp add: set_bind_spmf bind_UNION del: subsetI intro!: UN_least step.IH assms * ** order_trans[OF ile_eSuc *] split: generat.split)
qed

lemma interaction_bound_inline1:
  "\<lbrakk> Inr (out, c', rpv) \<in> set_spmf (inline1 callee gpv s);
     interaction_bound gpv \<le> p; \<And>s x. interaction_bound (callee s x) \<le> q \<rbrakk>
  \<Longrightarrow> eSuc (interaction_bound (c' input)) \<le> q \<and> eSuc (interaction_bound (rpv x)) \<le> p"
using interaction_bound_inline1'[where callee=callee and q=q and gpv=gpv and p=p and s=s] by auto
    
lemma interaction_bounded_by_inline1:
  "\<lbrakk> Inr (out, c', rpv) \<in> set_spmf (inline1 callee gpv s);
     interaction_bounded_by gpv p; \<And>s x. interaction_bounded_by (callee s x) q \<rbrakk>
  \<Longrightarrow> q \<noteq> 0 \<and> interaction_bounded_by (c' input) (q - 1) \<and> p \<noteq> 0 \<and> interaction_bounded_by (rpv x) (p - 1)"
unfolding interaction_bounded_by.simps
apply(drule (1) interaction_bound_inline1[where input=input and x=x], assumption)
apply(cases p q rule: nat.exhaust[case_product nat.exhaust])
apply(simp_all add: zero_enat_def[symmetric] eSuc_enat[symmetric])
done

lemma interaction_bounded_by_inline [interaction_bound]: (* TODO: augment with invariants and types *)
  fixes callee :: "'s \<Rightarrow> 'call \<Rightarrow> ('ret \<times> 's, 'call', 'ret') gpv"
  and gpv :: "('a, 'call, 'ret) gpv"
  assumes p: "interaction_bounded_by gpv p"
  and q: "\<And>s x. interaction_bounded_by (callee s x) q"
  shows "interaction_bounded_by (inline callee gpv s) (p * q)"
proof -
  fix x
  def gpv' \<equiv> "Done (x, s) :: ('ret \<times> 's, 'call', 'ret') gpv" 
    and r \<equiv> "0 :: nat" and rpv \<equiv> "(\<lambda>_. gpv) :: ('a, 'call, 'ret) rpv"
  with p have "interaction_bounded_by gpv' r" "\<And>y. interaction_bounded_by (rpv y) p"
    by simp_all
  hence "interaction_bound (bind_gpv gpv' (\<lambda>(y, s). inline callee (rpv y) s)) \<le> enat (r + p * q)"
  proof(coinduction arbitrary: gpv' rpv p r rule: interaction_bound_coinduct)
    case [rule_format]: (interaction_bound out c input gpv' rpv p r)
    from \<open>IO out c \<in> set_spmf (the_gpv (gpv' \<bind> (\<lambda>(y, s). inline callee (rpv y) s)))\<close>
    obtain generat where generat: "generat \<in> set_spmf (the_gpv gpv')"
      and *: "IO out c \<in> set_spmf (if is_Pure generat then the_gpv (case result generat of (y, s) \<Rightarrow> inline callee (rpv y) s)
               else return_spmf (IO (output generat) (\<lambda>input. continuation generat input \<bind> (\<lambda>(y, s). inline callee (rpv y) s))))"
      by(clarsimp simp add: set_bind_spmf)
    show ?case
    proof(cases generat)
      case (Pure ys)
      obtain y s where ys [simp]: "ys = (y, s)" by(cases ys)
      from Pure * have "IO out c \<in> set_spmf (the_gpv (inline callee (rpv y) s))" by simp
      then obtain c' rpv' where c'rpv: "Inr (out, c', rpv') \<in> set_spmf (inline1 callee (rpv y) s)"
        and c: "c = (\<lambda>input. c' input \<bind> (\<lambda>(x, s). inline callee (rpv' x) s))"
        by(clarsimp simp add: inline_sel split: sum.split_asm)
      from interaction_bounded_by_inline1[OF c'rpv interaction_bound(2) q]
      have "q \<noteq> 0" and c': "interaction_bounded_by (c' input) (q - 1)"
        and "p \<noteq> 0" and rpv': "\<And>x. interaction_bounded_by (rpv' x) (p - 1)" by blast+
      let ?n = "(q - 1) + (p - 1) * q"
      have "eSuc ?n \<le> r + p * q" using \<open>q \<noteq> 0\<close> \<open>p \<noteq> 0\<close>
        by(clarsimp simp add: eSuc_enat gr0_conv_Suc)
      then show ?thesis unfolding c using c' rpv' q by blast
    next
      case (IO out' c')
      with * have c: "c = (\<lambda>input. c' input \<bind> (\<lambda>(y, s). inline callee (rpv y) s))"
        and [simp]: "out' = out" by simp_all
      from \<open>interaction_bounded_by gpv' r\<close> generat IO
      have "r > 0" and c': "interaction_bounded_by (c' input) (r - 1)" 
        by(auto dest: interaction_bounded_by_IO)
      let ?n = "(r - 1) + p * q"
      have "eSuc ?n \<le> r + p * q" using \<open>r > 0\<close> by(clarsimp simp add: gr0_conv_Suc eSuc_enat)
      then show ?thesis unfolding c using c' interaction_bound(2) by blast
    qed
  qed
  then show ?thesis unfolding gpv'_def rpv_def r_def interaction_bounded_by.simps by simp
qed

context
  fixes \<I> :: "('call, 'ret) \<I>"
  and \<I>' :: "('call', 'ret') \<I>"
  and callee :: "'s \<Rightarrow> 'call \<Rightarrow> ('ret \<times> 's, 'call', 'ret') gpv"
  assumes results: "\<And>s x. x \<in> outs_\<I> \<I> \<Longrightarrow> results_gpv \<I>' (callee s x) \<subseteq> responses_\<I> \<I> x \<times> UNIV"
begin

lemma inline1_in_sub_gpvs_callee:
  assumes "Inr (out, callee', rpv') \<in> set_spmf (inline1 callee gpv s)"
  and WT: "\<I> \<turnstile>g gpv \<surd>"
  shows "\<exists>call\<in>outs_\<I> \<I>. \<exists>s. \<forall>x \<in> responses_\<I> \<I>' out. callee' x \<in> sub_gpvs \<I>' (callee s call)"
proof -
  from WT
  have "set_spmf (inline1 callee gpv s) \<subseteq> {Inr (out, callee', rpv') | out callee' rpv'.
    \<exists>call\<in>outs_\<I> \<I>. \<exists>s. \<forall>x \<in> responses_\<I> \<I>' out. callee' x \<in> sub_gpvs \<I>' (callee s call)} \<union> range Inl"
    (is "?concl (inline1 callee) gpv s")
  proof(induction arbitrary: gpv s rule: inline1_fixp_induct)
    case adm show ?case by(intro cont_intro ccpo_class.admissible_leI)
    case bottom show ?case by simp
    case (step inline1')
    { fix out c
      assume IO: "IO out c \<in> set_spmf (the_gpv gpv)" 
      from step.prems IO have out: "out \<in> outs_\<I> \<I>" by(rule WT_gpvD)
      { fix x s'
        assume Pure: "Pure (x, s') \<in> set_spmf (the_gpv (callee s out))"
        then have "(x, s') \<in> results_gpv \<I>' (callee s out)" by(rule results_gpv.Pure)
        with out have "x \<in> responses_\<I> \<I> out" by(auto dest: results)
        with step.prems IO have "\<I> \<turnstile>g c x \<surd>" by(rule WT_gpvD)
        hence "?concl inline1' (c x) s'" by(rule step.IH)
      } moreover {
        fix out' c'
        assume "IO out' c' \<in> set_spmf (the_gpv (callee s out))"
        hence "\<forall>x\<in>responses_\<I> \<I>' out'. c' x \<in> sub_gpvs \<I>' (callee s out)"
          by(auto intro: sub_gpvs.base)
        then have "\<exists>call\<in>outs_\<I> \<I>. \<exists>s. \<forall>x\<in>responses_\<I> \<I>' out'. c' x \<in> sub_gpvs \<I>' (callee s call)"
          using out by blast
      } moreover note calculation }
    then show ?case using step.prems
      by(auto del: subsetI simp add: set_bind_spmf bind_UNION intro!: UN_least split: generat.split)
  qed
  thus ?thesis using assms by fastforce
qed

lemma inline1_in_sub_gpvs:
  assumes "Inr (out, callee', rpv') \<in> set_spmf (inline1 callee gpv s)"
  and "(x, s') \<in> results_gpv \<I>' (callee' input)"
  and "input \<in> responses_\<I> \<I>' out"
  and "\<I> \<turnstile>g gpv \<surd>"
  shows "rpv' x \<in> sub_gpvs \<I> gpv"
proof -
  from \<open>\<I> \<turnstile>g gpv \<surd>\<close>
  have "set_spmf (inline1 callee gpv s) \<subseteq> {Inr (out, callee', rpv') | out callee' rpv'.
    \<forall>input \<in> responses_\<I> \<I>' out. \<forall>(x, s')\<in>results_gpv \<I>' (callee' input). rpv' x \<in> sub_gpvs \<I> gpv}
    \<union> range Inl" (is "?concl (inline1 callee) gpv s" is "_ \<subseteq> ?rhs gpv s")
  proof(induction arbitrary: gpv s rule: inline1_fixp_induct)
    case adm show ?case by(intro cont_intro ccpo_class.admissible_leI)
    case bottom show ?case by simp
  next
    case (step inline1')
    { fix out c
      assume IO: "IO out c \<in> set_spmf (the_gpv gpv)" 
      from step.prems IO have out: "out \<in> outs_\<I> \<I>" by(rule WT_gpvD)
      { fix x s'
        assume Pure: "Pure (x, s') \<in> set_spmf (the_gpv (callee s out))"
        then have "(x, s') \<in> results_gpv \<I>' (callee s out)" by(rule results_gpv.Pure)
        with out have "x \<in> responses_\<I> \<I> out" by(auto dest: results)
        with step.prems IO have "\<I> \<turnstile>g c x \<surd>" by(rule WT_gpvD)
        hence "?concl inline1' (c x) s'" by(rule step.IH)
        also have "\<dots> \<subseteq> ?rhs gpv s'" using IO Pure
          by(fastforce intro: sub_gpvs.cont dest: WT_gpv_OutD[OF step.prems] results[THEN subsetD, OF _ results_gpv.Pure])
        finally have "set_spmf (inline1' (c x) s') \<subseteq> \<dots>" .
      } moreover {
        fix out' c' input x s'
        assume "IO out' c' \<in> set_spmf (the_gpv (callee s out))"
          and "input \<in> responses_\<I> \<I>' out'" and "(x, s') \<in> results_gpv \<I>' (c' input)"
        then have "c x \<in> sub_gpvs \<I> gpv" using IO
          by(auto intro!: sub_gpvs.base dest: WT_gpv_OutD[OF step.prems] results[THEN subsetD, OF _ results_gpv.IO])
      } moreover note calculation }
    then show ?case
      by(auto simp add: set_bind_spmf bind_UNION intro!: UN_least split: generat.split del: subsetI)
  qed
  with assms show ?thesis by fastforce
qed

context
  assumes WT: "\<And>x s. x \<in> outs_\<I> \<I> \<Longrightarrow> \<I>' \<turnstile>g callee s x \<surd>"
begin

lemma WT_gpv_inline1:
  assumes "Inr (out, rpv, rpv') \<in> set_spmf (inline1 callee gpv s)"
  and "\<I> \<turnstile>g gpv \<surd>"
  shows "out \<in> outs_\<I> \<I>'" (is "?thesis1")
  and "input \<in> responses_\<I> \<I>' out \<Longrightarrow> \<I>' \<turnstile>g rpv input \<surd>" (is "PROP ?thesis2")
  and "\<lbrakk> input \<in> responses_\<I> \<I>' out; (x, s') \<in> results_gpv \<I>' (rpv input) \<rbrakk> \<Longrightarrow> \<I> \<turnstile>g rpv' x \<surd>" (is "PROP ?thesis3")
proof -
  from \<open>\<I> \<turnstile>g gpv \<surd>\<close>
  have "set_spmf (inline1 callee gpv s) \<subseteq> {Inr (out, rpv, rpv') | out rpv rpv'. out \<in> outs_\<I> \<I>'} \<union> range Inl"
  proof(induction arbitrary: gpv s rule: inline1_fixp_induct)
    { case adm show ?case by(intro cont_intro ccpo_class.admissible_leI) }
    { case bottom show ?case by simp }
    case (step inline1')
    { fix out c
      assume IO: "IO out c \<in> set_spmf (the_gpv gpv)" 
      from step.prems IO have out: "out \<in> outs_\<I> \<I>" by(rule WT_gpvD)
      { fix x s'
        assume Pure: "Pure (x, s') \<in> set_spmf (the_gpv (callee s out))"
        then have "(x, s') \<in> results_gpv \<I>' (callee s out)" by(rule results_gpv.Pure)
        with out have "x \<in> responses_\<I> \<I> out" by(auto dest: results)
        with step.prems IO have "\<I> \<turnstile>g c x \<surd>" by(rule WT_gpvD)
      } moreover {
        fix out' c'
        from out have "\<I>' \<turnstile>g callee s out \<surd>" by(rule WT)
        moreover assume "IO out' c' \<in> set_spmf (the_gpv (callee s out))"
        ultimately have "out' \<in> outs_\<I> \<I>'" by(rule WT_gpvD) 
      } moreover note calculation }
    then show ?case 
      by(auto del: subsetI simp add: set_bind_spmf bind_UNION intro!: UN_least split: generat.split intro!: step.IH[THEN order_trans])
  qed
  then show ?thesis1 using assms by auto

  assume "input \<in> responses_\<I> \<I>' out"
  with inline1_in_sub_gpvs_callee[OF \<open>Inr _ \<in> _\<close>] \<open>\<I> \<turnstile>g gpv \<surd>\<close>
  obtain out' s where "out' \<in> outs_\<I> \<I>" 
    and *: "rpv input \<in> sub_gpvs \<I>' (callee s out')" by auto
  from \<open>out' \<in> _\<close> have "\<I>' \<turnstile>g callee s out' \<surd>" by(rule WT)
  then show "\<I>' \<turnstile>g rpv input \<surd>" using * by(rule WT_sub_gpvsD)

  assume "(x, s') \<in> results_gpv \<I>' (rpv input)"
  with \<open>Inr _ \<in> _\<close> have "rpv' x \<in> sub_gpvs \<I> gpv"
    using \<open>input \<in> _\<close> \<open>\<I> \<turnstile>g gpv \<surd>\<close> by(rule inline1_in_sub_gpvs)
  with \<open>\<I> \<turnstile>g gpv \<surd>\<close> show "\<I> \<turnstile>g rpv' x \<surd>" by(rule WT_sub_gpvsD)
qed

lemma WT_gpv_inline:
  assumes "\<I> \<turnstile>g gpv \<surd>"
  shows "\<I>' \<turnstile>g inline callee gpv s \<surd>"
using assms
proof(coinduction arbitrary: gpv s rule: WT_gpv_coinduct_bind)
  case (WT_gpv out c gpv)
  from \<open>IO out c \<in> _\<close> obtain callee' rpv'
    where Inr: "Inr (out, callee', rpv') \<in> set_spmf (inline1 callee gpv s)"
    and c: "c = (\<lambda>input. callee' input \<bind> (\<lambda>(x, s). inline callee (rpv' x) s))"
    by(clarsimp simp add: inline_sel split: sum.split_asm)
  from Inr \<open>\<I> \<turnstile>g gpv \<surd>\<close> have ?out by(rule WT_gpv_inline1)
  moreover have "?cont TYPE('ret \<times> 's)" (is "\<forall>input\<in>_. _ \<or> _ \<or> ?case' input")
  proof(rule ballI disjI2)+
    fix input
    assume "input \<in> responses_\<I> \<I>' out"
    with Inr \<open>\<I> \<turnstile>g gpv \<surd> \<close>have "\<I>' \<turnstile>g callee' input \<surd>"
      and "\<And>x s'. (x, s') \<in> results_gpv \<I>' (callee' input) \<Longrightarrow> \<I> \<turnstile>g rpv' x \<surd>"
      by(blast intro: WT_gpv_inline1)+
    then show "?case' input" by(subst c)(auto 4 4)
  qed
  ultimately show "?case TYPE('ret \<times> 's)" ..
qed

end

context
  fixes gpv :: "('a, 'call, 'ret) gpv"
  assumes gpv: "lossless_gpv \<I> gpv" "\<I> \<turnstile>g gpv \<surd>"
begin

lemma lossless_spmf_inline1:
  assumes lossless: "\<And>s x. x \<in> outs_\<I> \<I> \<Longrightarrow> lossless_spmf (the_gpv (callee s x))"
  shows "lossless_spmf (inline1 callee gpv s)"
using gpv
proof(induction arbitrary: s rule: lossless_WT_gpv_induct)
  case (lossless_gpv p)
  show ?case using \<open>lossless_spmf p\<close>
    apply(subst inline1_unfold)
    apply(auto split: generat.split intro: lossless lossless_gpv.hyps dest: results[THEN subsetD, rotated, OF results_gpv.Pure] intro: lossless_gpv.IH)
    done
qed

lemma lossless_gpv_inline1:
  assumes *: "Inr (out, rpv, rpv') \<in> set_spmf (inline1 callee gpv s)"
  and **: "input \<in> responses_\<I> \<I>' out"
  and lossless: "\<And>s x. x \<in> outs_\<I> \<I> \<Longrightarrow> lossless_gpv \<I>' (callee s x)"
  shows "lossless_gpv \<I>' (rpv input)"
proof -
  from inline1_in_sub_gpvs_callee[OF * gpv(2)] **
  obtain out' s where "out' \<in> outs_\<I> \<I>" and ***: "rpv input \<in> sub_gpvs \<I>' (callee s out')" by blast
  from \<open>out' \<in> _\<close> have "lossless_gpv \<I>' (callee s out')" by(rule lossless)
  thus ?thesis using *** by(rule lossless_sub_gpvsD)
qed

lemma lossless_results_inline1:
  assumes "Inr (out, rpv, rpv') \<in> set_spmf (inline1 callee gpv s)"
  and "(x, s') \<in> results_gpv \<I>' (rpv input)"
  and "input \<in> responses_\<I> \<I>' out"
  shows "lossless_gpv \<I> (rpv' x)"
proof -
  from assms gpv(2) have "rpv' x \<in> sub_gpvs \<I> gpv" by(rule inline1_in_sub_gpvs)
  with gpv(1) show "lossless_gpv \<I> (rpv' x)" by(rule lossless_sub_gpvsD)
qed

end

lemmas lossless_inline1[rotated 2] = lossless_spmf_inline1 lossless_gpv_inline1 lossless_results_inline1

lemma lossless_inline[rotated]:
  fixes gpv :: "('a, 'call, 'ret) gpv"
  assumes gpv: "lossless_gpv \<I> gpv" "\<I> \<turnstile>g gpv \<surd>"
  and lossless: "\<And>s x. x \<in> outs_\<I> \<I> \<Longrightarrow> lossless_gpv \<I>' (callee s x)"
  shows "lossless_gpv \<I>' (inline callee gpv s)"
using gpv
proof(induction arbitrary: s rule: lossless_WT_gpv_induct_strong)
  case (lossless_gpv p)
  have lp: "lossless_gpv \<I> (GPV p)" by(rule lossless_sub_gpvsI)(auto intro: lossless_gpv.hyps)
  moreover have wp: "\<I> \<turnstile>g GPV p \<surd>" by(rule WT_sub_gpvsI)(auto intro: lossless_gpv.hyps)
  ultimately have "lossless_spmf (the_gpv (inline callee (GPV p) s))"
    by(auto simp add: inline_sel intro: lossless_spmf_inline1 lossless_gpv_lossless_spmfD[OF lossless])
  moreover {
    fix out c input
    assume IO: "IO out c \<in> set_spmf (the_gpv (inline callee (GPV p) s))"
      and "input \<in> responses_\<I> \<I>' out"
    from IO obtain callee' rpv
      where Inr: "Inr (out, callee', rpv) \<in> set_spmf (inline1 callee (GPV p) s)"
      and c: "c = (\<lambda>input. callee' input \<bind> (\<lambda>(x, y). inline callee (rpv x) y))"
      by(clarsimp simp add: inline_sel split: sum.split_asm)
    from Inr \<open>input \<in> _\<close> lossless lp wp have "lossless_gpv \<I>' (callee' input)" by(rule lossless_inline1)
    moreover {
      fix x s'
      assume "(x, s') \<in> results_gpv \<I>' (callee' input)"
      with Inr have "rpv x \<in> sub_gpvs \<I> (GPV p)" using \<open>input \<in> _\<close> wp by(rule inline1_in_sub_gpvs)
      hence "lossless_gpv \<I>' (inline callee (rpv x) s')" by(rule lossless_gpv.IH)
    } ultimately have "lossless_gpv \<I>' (c input)" unfolding c by clarsimp
  } ultimately show ?case by(rule lossless_gpvI)
qed

end

definition id_oracle :: "'s \<Rightarrow> 'call \<Rightarrow> ('ret \<times> 's, 'call, 'ret) gpv"
where "id_oracle s x = Pause x (\<lambda>x. Done (x, s))"

lemma inline1_id_oracle:
  "inline1 id_oracle gpv s =
   map_spmf (\<lambda>generat. case generat of Pure x \<Rightarrow> Inl (x, s) | IO out c \<Rightarrow> Inr (out, \<lambda>x. Done (x, s), c)) (the_gpv gpv)"
by(subst inline1.simps)(auto simp add: id_oracle_def map_spmf_conv_bind_spmf intro!: bind_spmf_cong split: generat.split)

lemma inline_id_oracle [simp]: "inline id_oracle gpv s = map_gpv (\<lambda>x. (x, s)) id gpv"
by(coinduction arbitrary: gpv s)(auto 4 3 simp add: inline_sel inline1_id_oracle spmf_rel_map gpv.map_sel o_def generat.rel_map intro!: rel_spmf_reflI rel_funI split: generat.split)

subsection {* Running GPVs *}

type_synonym ('call, 'ret, 's) callee = "'s \<Rightarrow> 'call \<Rightarrow> ('ret \<times> 's) spmf"

context fixes callee :: "('call, 'ret, 's) callee" notes [[function_internals]] begin

partial_function (spmf) exec_gpv :: "('a, 'call, 'ret) gpv \<Rightarrow> 's \<Rightarrow> ('a \<times> 's) spmf"
where
  "exec_gpv c s =
   the_gpv c \<bind>
     case_generat (\<lambda>x. return_spmf (x, s))
     (\<lambda>out c. callee s out \<bind> (\<lambda>(x, y). exec_gpv (c x) y))"

abbreviation run_gpv :: "('a, 'call, 'ret) gpv \<Rightarrow> 's \<Rightarrow> 'a spmf"
where "run_gpv gpv s \<equiv> map_spmf fst (exec_gpv gpv s)"

lemma exec_gpv_fixp_induct [case_names adm bottom step]:
  assumes "ccpo.admissible (fun_lub lub_spmf) (fun_ord (ord_spmf op =)) (\<lambda>f. P (\<lambda>c s. f (c, s)))"
  and "P (\<lambda>_ _. return_pmf None)"
  and "\<And>exec_gpv. P exec_gpv \<Longrightarrow> 
     P (\<lambda>c s. the_gpv c \<bind> case_generat (\<lambda>x. return_spmf (x, s)) (\<lambda>out c. callee s out \<bind> (\<lambda>(x, y). exec_gpv (c x) y)))"
  shows "P exec_gpv"
using assms(1)
by(rule exec_gpv.fixp_induct[unfolded curry_conv[abs_def]])(simp_all add: assms(2-))

lemma exec_gpv_fixp_induct_strong [case_names adm bottom step]:
  assumes "ccpo.admissible (fun_lub lub_spmf) (fun_ord (ord_spmf op =)) (\<lambda>f. P (\<lambda>c s. f (c, s)))"
  and "P (\<lambda>_ _. return_pmf None)"
  and "\<And>exec_gpv'. \<lbrakk> \<And>c s. ord_spmf op = (exec_gpv' c s) (exec_gpv c s); P exec_gpv' \<rbrakk>
    \<Longrightarrow> P (\<lambda>c s. the_gpv c \<bind> case_generat (\<lambda>x. return_spmf (x, s)) (\<lambda>out c. callee s out \<bind> (\<lambda>(x, y). exec_gpv' (c x) y)))"
  shows "P exec_gpv"
using assms
by(rule spmf.fixp_strong_induct_uc[where P="\<lambda>f. P (curry f)" and U=case_prod and C=curry, OF exec_gpv.mono exec_gpv_def, simplified curry_case_prod, simplified curry_conv[abs_def] fun_ord_def split_paired_All prod.case case_prod_eta, OF refl]) blast

lemma exec_gpv_fixp_induct_strong2 [case_names adm bottom step]:
  assumes "ccpo.admissible (fun_lub lub_spmf) (fun_ord (ord_spmf op =)) (\<lambda>f. P (\<lambda>c s. f (c, s)))"
  and "P (\<lambda>_ _. return_pmf None)"
  and "\<And>exec_gpv'.
    \<lbrakk> \<And>c s. ord_spmf op = (exec_gpv' c s) (exec_gpv c s); 
      \<And>c s. ord_spmf op = (exec_gpv' c s) (the_gpv c \<bind> case_generat (\<lambda>x. return_spmf (x, s)) (\<lambda>out c. callee s out \<bind> (\<lambda>(x, y). exec_gpv' (c x) y)));
      P exec_gpv' \<rbrakk>
    \<Longrightarrow> P (\<lambda>c s. the_gpv c \<bind> case_generat (\<lambda>x. return_spmf (x, s)) (\<lambda>out c. callee s out \<bind> (\<lambda>(x, y). exec_gpv' (c x) y)))"
  shows "P exec_gpv"
using assms
by(rule spmf.fixp_induct_strong2_uc[where P="\<lambda>f. P (curry f)" and U=case_prod and C=curry, OF exec_gpv.mono exec_gpv_def, simplified curry_case_prod, simplified curry_conv[abs_def] fun_ord_def split_paired_All prod.case case_prod_eta, OF refl]) blast+

end

lemma exec_gpv_conv_inline1:
  "exec_gpv callee gpv s = map_spmf projl (inline1 (\<lambda>s c. lift_spmf (callee s c) :: (_, unit, unit) gpv) gpv s)"
by(induction arbitrary: gpv s rule: parallel_fixp_induct_2_2[OF partial_function_definitions_spmf partial_function_definitions_spmf exec_gpv.mono inline1.mono exec_gpv_def inline1_def, unfolded lub_spmf_empty, case_names adm bottom step])
  (auto simp add: map_spmf_bind_spmf o_def spmf.map_comp bind_map_spmf split_def intro!: bind_spmf_cong[OF refl] split: generat.split)

lemma exec_gpv_simps:
  "exec_gpv callee gpv s =
   the_gpv gpv \<bind>
     case_generat (\<lambda>x. return_spmf (x, s))
     (\<lambda>out rpv. callee s out \<bind> (\<lambda>(x, y). exec_gpv callee (rpv x) y))"
by(fact exec_gpv.simps)

lemma exec_gpv_lift_spmf [simp]:
  "exec_gpv callee (lift_spmf p) s = bind_spmf p (\<lambda>x. return_spmf (x, s))"
by(simp add: exec_gpv_conv_inline1 spmf.map_comp o_def map_spmf_conv_bind_spmf)

lemma exec_gpv_Done [simp]: "exec_gpv callee (Done x) s = return_spmf (x, s)"
by(simp add: exec_gpv_conv_inline1)

lemma exec_gpv_Fail [simp]: "exec_gpv callee Fail s = return_pmf None"
by(simp add: exec_gpv_conv_inline1)

lemma if_distrib_exec_gpv [if_distribs]:
  "exec_gpv callee (if b then x else y) s = (if b then exec_gpv callee x s else exec_gpv callee y s)"
by simp

lemmas exec_gpv_fixp_parallel_induct [case_names adm bottom step] =
  parallel_fixp_induct_2_2[OF partial_function_definitions_spmf partial_function_definitions_spmf exec_gpv.mono exec_gpv.mono exec_gpv_def exec_gpv_def, unfolded lub_spmf_empty]

context includes lifting_syntax begin

lemma exec_gpv_parametric [transfer_rule]:
  fixes A :: "'a \<Rightarrow> 'b \<Rightarrow> bool"
  and CALL :: "'call \<Rightarrow> 'call' \<Rightarrow> bool"
  and S :: "'s \<Rightarrow> 's' \<Rightarrow> bool"
  shows
  "((S ===> CALL ===> rel_spmf (rel_prod (op = :: 'ret \<Rightarrow> _) S)) ===> rel_gpv A CALL ===> S ===> rel_spmf (rel_prod A S))
  exec_gpv exec_gpv"
apply(rule rel_funI)+
apply(unfold spmf_rel_map exec_gpv_conv_inline1)
apply(rule rel_spmf_mono_strong)
 apply(erule inline1_parametric[THEN rel_funD, THEN rel_funD, THEN rel_funD, rotated])
  prefer 3
  apply(drule in_set_inline1_lift_spmf1)+
  apply fastforce
 subgoal by simp
subgoal premises [transfer_rule] by transfer_prover
done

end

lemma exec_gpv_bind: "exec_gpv callee (c \<bind> f) s = exec_gpv callee c s \<bind> (\<lambda>(x, s') \<Rightarrow> exec_gpv callee (f x) s')"
by(auto simp add: exec_gpv_conv_inline1 inline1_bind_gpv map_spmf_bind_spmf o_def bind_map_spmf intro!: bind_spmf_cong[OF refl] dest: in_set_inline1_lift_spmf1)

lemma exec_gpv_map_gpv_id:
  "exec_gpv oracle (map_gpv f id gpv) \<sigma> = map_spmf (apfst f) (exec_gpv oracle gpv \<sigma>)"
proof(rule sym)
  def gpv' \<equiv> "map_gpv f id gpv"
  have [transfer_rule]: "rel_gpv (\<lambda>x y. y = f x) op = gpv gpv'"
    unfolding gpv'_def by(simp add: gpv.rel_map gpv.rel_refl)
  have "rel_spmf (rel_prod (\<lambda>x y. y = f x) op =) (exec_gpv oracle gpv \<sigma>) (exec_gpv oracle gpv' \<sigma>)"
    by transfer_prover
  thus "map_spmf (apfst f) (exec_gpv oracle gpv \<sigma>) = exec_gpv oracle (map_gpv f id gpv) \<sigma>"
    unfolding spmf_rel_eq[symmetric] gpv'_def spmf_rel_map by(rule rel_spmf_mono) clarsimp
qed

lemma exec_gpv_Pause [simp]:
  "exec_gpv callee (Pause out f) s = callee s out \<bind> (\<lambda>(x, s'). exec_gpv callee (f x) s')"
by(simp add: inline1_Pause map_spmf_bind_spmf bind_map_spmf o_def exec_gpv_conv_inline1 split_def)

lemma exec_gpv_inline:
  fixes callee :: "'s \<Rightarrow> 'c \<Rightarrow> ('r \<times> 's) spmf"
  and gpv :: "'s' \<Rightarrow> 'c' \<Rightarrow> ('r' \<times> 's', 'c, 'r) gpv"
  shows "exec_gpv callee (inline gpv c' s') s =
    map_spmf (\<lambda>(x, s', s). ((x, s'), s)) (exec_gpv (\<lambda>(s', s) y. map_spmf (\<lambda>((x, s'), s). (x, s', s)) (exec_gpv callee (gpv s' y) s)) c' (s', s))"
    (is "?lhs = ?rhs")
proof -
  have "?lhs = map_spmf projl (map_spmf (map_sum (\<lambda>(x, s2, y). ((x, s2), y))
        (\<lambda>(x, rpv'' :: ('r \<times> 's, unit, unit) rpv, rpv', rpv). (x, rpv'', \<lambda>r1. bind_gpv (rpv' r1) (\<lambda>(r2, y). inline gpv (rpv r2) y))))
      (inline2 (\<lambda>s c. lift_spmf (callee s c)) gpv c' s' s))"
    unfolding exec_gpv_conv_inline1 by(simp add: inline1_inline_conv_inline2)
  also have "\<dots> = map_spmf (\<lambda>(x, s', s). ((x, s'), s)) (map_spmf projl (map_spmf (map_sum id
        (\<lambda>(x, rpv'' :: ('r \<times> 's, unit, unit) rpv, rpv', rpv). (x, \<lambda>r. bind_gpv (rpv'' r) (\<lambda>(r1, s1). map_gpv (\<lambda>((r2, s2), s1). (r2, s2, s1)) id (inline (\<lambda>s c. lift_spmf (callee s c)) (rpv' r1) s1)), rpv)))
      (inline2 (\<lambda>s c. lift_spmf (callee s c)) gpv c' s' s)))"
   unfolding spmf.map_comp by(rule map_spmf_cong[OF refl])(auto dest!: in_set_inline2_lift_spmf1)
  also have "\<dots> = ?rhs" unfolding exec_gpv_conv_inline1
    by(subst inline1_inline_conv_inline2'[symmetric])(simp add: spmf.map_comp split_def inline_lift_spmf1 map_lift_spmf)
  finally show ?thesis .
qed

locale callee_invariant_base =
  fixes callee :: "'s \<Rightarrow> 'a \<Rightarrow> ('b \<times> 's) spmf"
  and I :: "'s \<Rightarrow> bool"

locale callee_invariant = callee_invariant_base
  +
  assumes callee_invariant: "\<And>s x y s'. \<lbrakk> (y, s') \<in> set_spmf (callee s x); I s \<rbrakk> \<Longrightarrow> I s'"
begin

lemma exec_gpv_invariant':
  "I s \<Longrightarrow> set_spmf (exec_gpv callee gpv s) \<subseteq> {(x, s'). I s'}"
proof(induction arbitrary: gpv s rule: exec_gpv_fixp_induct)
  case adm show ?case by(intro cont_intro ccpo_class.admissible_leI)
  case bottom show ?case by simp
  case step thus ?case 
    by(auto simp add: set_bind_spmf bind_UNION intro!: UN_least del: subsetI split: generat.split dest!: callee_invariant)
qed

lemma exec_gpv_invariant:
  "\<lbrakk> (x, s') \<in> set_spmf (exec_gpv callee gpv s); I s \<rbrakk> \<Longrightarrow> I s'"
by(drule exec_gpv_invariant') blast

lemma interaction_bounded_by_exec_gpv_count':
  fixes count
  assumes bound: "interaction_bounded_by gpv n"
  and count: "\<And>s x y s'. \<lbrakk> (y, s') \<in> set_spmf (callee s x); I s \<rbrakk> \<Longrightarrow> count s' \<le> Suc (count s)"
  and I: "I s"
  shows "set_spmf (exec_gpv callee gpv s) \<subseteq> {(x, s'). count s' \<le> n + count s}"
using bound I
proof(induction arbitrary: gpv s n rule: exec_gpv_fixp_induct)
  case adm show ?case by(intro cont_intro ccpo_class.admissible_leI)
  case bottom show ?case by simp
  case (step exec_gpv')
  { fix out c input s'
    assume out: "IO out c \<in> set_spmf (the_gpv gpv)"
      and input: "(input, s') \<in> set_spmf (callee s out)"
    from step.prems out have "n > 0"
      and bound': "interaction_bounded_by (c input) (n - 1)"
      by(auto dest: interaction_bounded_by_contD)
    from input \<open>I s\<close> have "I s'" by(rule callee_invariant)
    with bound' have "set_spmf (exec_gpv' (c input) s') \<subseteq> {(x, s''). count s'' \<le> n - 1 + count s'}"
      by(rule step.IH)
    also have "\<dots> \<subseteq> {(x, s''). count s'' \<le> n + count s}" using \<open>n > 0\<close> count[OF input \<open>I s\<close>]
      by(cases n rule: nat.exhaust)(auto)
    also note calculation }
  then show ?case
    by(clarsimp simp add: set_bind_spmf bind_UNION del: subsetI intro!: UN_least split: generat.split)
qed

lemma interaction_bounded_by_exec_gpv_count:
  fixes count
  assumes bound: "interaction_bounded_by gpv n"
  and xs': "(x, s') \<in> set_spmf (exec_gpv callee gpv s)"
  and count: "\<And>s x y s'. \<lbrakk> (y, s') \<in> set_spmf (callee s x); I s \<rbrakk> \<Longrightarrow> count s' \<le> Suc (count s)"
  and I: "I s"
  shows "count s' \<le> n + count s"
using bound count I by(rule interaction_bounded_by_exec_gpv_count'[THEN subsetD, OF _ _ _ xs', unfolded mem_Collect_eq prod.case])

lemma interaction_bounded_by_exec_gpv_bad':
  fixes bad
  assumes bound: "interaction_bounded_by gpv n"
  and I: "I s"
  and not_bad: "\<not> bad s"
  and prob_bad: "\<And>s x. \<lbrakk> I s; \<not> bad s \<rbrakk> \<Longrightarrow> spmf (map_spmf (bad \<circ> snd) (callee s x)) True \<le> k"
  shows "ennreal (spmf (map_spmf (bad \<circ> snd) (exec_gpv callee gpv s)) True) \<le> ennreal k * n"
proof -
  from prob_bad[OF I not_bad] have k_nonneg: "0 \<le> k" by (meson order_trans pmf_nonneg) 
  show ?thesis using bound I not_bad
  proof(induction arbitrary: n s gpv rule: exec_gpv_fixp_induct)
    case adm show ?case by(intro cont_intro ccpo_class.admissible_leI)
    case bottom show ?case using k_nonneg by(cases n)simp_all
    case (step exec_gpv')
    let ?M = "restrict_space (measure_spmf (the_gpv gpv)) {x. \<not> is_Pure x}"
    let ?O = "\<lambda>generat. measure_spmf (callee s (output generat))"
    have "ennreal (spmf (map_spmf (bad \<circ> snd) (the_gpv gpv \<bind> case_generat (\<lambda>x. return_spmf (x, s)) (\<lambda>out c. callee s out \<bind> (\<lambda>(x, y). exec_gpv' (c x) y)))) True)
      = \<integral>\<^sup>+ generat. ennreal (spmf (map_spmf (bad \<circ> snd) (case generat of Pure x \<Rightarrow> return_spmf (x, s)
                     | IO out c \<Rightarrow> callee s out \<bind> (\<lambda>(x, y). exec_gpv' (c x) y))) True)
       \<partial>measure_spmf (the_gpv gpv)"
      by(simp add: map_spmf_bind_spmf ennreal_spmf_bind)
    also have "\<dots> = \<integral>\<^sup>+ generat. ennreal (spmf (bind_spmf (callee s (output generat)) (\<lambda>(x, s'). map_spmf (bad \<circ> snd) (exec_gpv' (continuation generat x) s'))) True) \<partial>?M"
      by(auto simp add: nn_integral_restrict_space  \<open>\<not> bad s\<close> map_spmf_bind_spmf split_def o_def intro: nn_integral_cong split: generat.split)
    also have "\<dots> = \<integral>\<^sup>+ generat. \<integral>\<^sup>+ (x, s'). ennreal (spmf (map_spmf (bad \<circ> snd) (exec_gpv' (continuation generat x) s')) True) \<partial>?O generat \<partial>?M"
      by(simp add: ennreal_spmf_bind split_def)
    also have "\<dots> \<le> \<integral>\<^sup>+ generat. \<integral>\<^sup>+ (x, s'). (if bad s' then 1 else ennreal k * (n - 1)) \<partial>?O generat \<partial>?M"
    proof(clarsimp simp add: AE_restrict_space_iff not_is_Pure_conv intro!: nn_integral_mono_AE split del: if_split)
      fix out c x s'
      assume IO: "IO out c \<in> set_spmf (the_gpv gpv)"
        and xs': "(x, s') \<in> set_spmf (callee s out)"
      from xs' \<open>I s\<close> have "I s'" by(rule callee_invariant)
      show "ennreal (spmf (map_spmf (bad \<circ> snd) (exec_gpv' (c x) s')) True) \<le> (if bad s' then 1 else ennreal k * (n - 1))"
      proof(cases "bad s'")
        case False
        from step.prems(1) IO have "interaction_bounded_by (c x) (n - 1)"
          by(blast dest: interaction_bounded_by_contD)
        then show ?thesis using \<open>I s'\<close> False step.IH[of "c x" "n - 1" s'] by(simp add: o_def)
      qed(simp add: pmf_le_1)
    qed
    also have "\<dots> = \<integral>\<^sup>+ generat. \<integral>\<^sup>+ b. indicator {True} b + ennreal (k * (n - 1)) * indicator {False} b \<partial>measure_spmf (map_spmf (bad \<circ> snd) (callee s (output generat))) \<partial>?M"
      (is "_ = \<integral>\<^sup>+ generat. \<integral>\<^sup>+ _. _ \<partial>?O' generat \<partial>_")
      by(auto intro!: nn_integral_cong simp add: ennreal_mult[symmetric] ennreal_of_nat_eq_real_of_nat k_nonneg)
    also have "\<dots> = \<integral>\<^sup>+ generat. (\<integral>\<^sup>+ b. indicator {True} b \<partial>?O' generat) + ennreal k * (n - 1) * \<integral>\<^sup>+ b. indicator {False} b \<partial>?O' generat \<partial>?M"
      by(subst nn_integral_add)(simp_all add: k_nonneg nn_integral_cmult o_def ennreal_of_nat_eq_real_of_nat ennreal_mult)
    also have "\<dots> = \<integral>\<^sup>+ generat. spmf (map_spmf (bad \<circ> snd) (callee s (output generat))) True + k * (n - 1) * spmf (map_spmf (bad \<circ> snd) (callee s (output generat))) False \<partial>?M"
      by(simp del: nn_integral_map_spmf add: emeasure_spmf_single ennreal_mult k_nonneg ennreal_of_nat_eq_real_of_nat)
    also have "\<dots> \<le> \<integral>\<^sup>+ generat. ennreal k * n \<partial>?M"
    proof(rule nn_integral_mono_AE; clarsimp simp add: AE_restrict_space_iff not_is_Pure_conv ennreal_of_nat_eq_real_of_nat ennreal_mult[symmetric] k_nonneg ennreal_plus[symmetric] simp del: One_nat_def ennreal_plus)
      fix out c
      assume "IO out c \<in> set_spmf (the_gpv gpv)"
      hence "n > 0" using step.prems(1) by(blast dest: interaction_bounded_by_contD)
      have "spmf (map_spmf (bad \<circ> snd) (callee s out)) True \<le> k" (is "?o True \<le> _")
        using \<open>I s\<close> \<open>\<not> bad s\<close> by(rule prob_bad)
      hence "?o True \<le> k" by(simp del: o_apply)
      hence "?o True + k * (n - 1) * ?o False \<le> k + k * (n - 1) * 1"
        by(rule add_mono)(rule mult_left_mono, simp_all add: pmf_le_1 k_nonneg)
      also have "\<dots> \<le> k * n" using `n > 0`
        by(cases n)(auto simp add: zero_enat_def gr0_conv_Suc eSuc_enat[symmetric] field_simps)
      finally show "?o True + k * (n - 1) * ?o False \<le> k * n" .
    qed
    also have "\<dots> \<le> k * n * 1"
      by(simp add: k_nonneg emeasure_restrict_space measure_spmf.emeasure_eq_measure space_restrict_space measure_spmf.subprob_measure_le_1 mult_left_mono ennreal_of_nat_eq_real_of_nat ennreal_mult del: mult_1_right)
    finally show ?case by(simp del: o_apply add: ennreal_of_nat_eq_real_of_nat ennreal_mult k_nonneg)
  qed
qed

end

interpretation oi_True: callee_invariant callee "\<lambda>_. True" for callee
by unfold_locales (simp_all)

inductive WT_callee :: "('call, 'ret) \<I> \<Rightarrow> ('call \<Rightarrow> ('ret \<times> 's) spmf) \<Rightarrow> bool" ("(_) \<turnstile>c/ (_) \<surd>" [100, 0] 99)
  for \<I> callee
where
  WT_callee:
  "\<lbrakk> \<And>call ret s. \<lbrakk> call \<in> outs_\<I> \<I>; (ret, s) \<in> set_spmf (callee call) \<rbrakk> \<Longrightarrow> ret \<in> responses_\<I> \<I> call \<rbrakk>
  \<Longrightarrow> \<I> \<turnstile>c callee \<surd>"

lemmas WT_calleeI = WT_callee
hide_fact WT_callee

lemma WT_calleeD: "\<lbrakk> \<I> \<turnstile>c callee \<surd>; (ret, s) \<in> set_spmf (callee out); out \<in> outs_\<I> \<I> \<rbrakk> \<Longrightarrow> ret \<in> responses_\<I> \<I> out"
by(rule WT_callee.cases)

lemma WT_callee_top [intro!, simp]: "\<I>_top \<turnstile>c callee \<surd>"
by(rule WT_calleeI) simp

lemma (in callee_invariant) lossless_exec_gpv:
  assumes gpv: "lossless_gpv \<I> gpv"
  and callee: "\<And>s out. \<lbrakk> out \<in> outs_\<I> \<I>; I s \<rbrakk> \<Longrightarrow> lossless_spmf (callee s out)"
  and WT_gpv: "\<I> \<turnstile>g gpv \<surd>"
  and WT_callee: "\<And>s. I s \<Longrightarrow> \<I> \<turnstile>c callee s \<surd>"
  and I: "I s"
  shows "lossless_spmf (exec_gpv callee gpv s)"
using gpv WT_gpv I
proof(induction arbitrary: s rule: lossless_WT_gpv_induct)
  case (lossless_gpv gpv)
  show ?case using lossless_gpv.hyps lossless_gpv.prems
    by(subst exec_gpv.simps)(fastforce split: generat.split simp add: callee intro!: lossless_gpv.IH intro: WT_calleeD[OF WT_callee] elim!: callee_invariant)
qed

lemma lossless_exec_gpv:
  "\<lbrakk> lossless_gpv \<I> gpv;  \<And>s out. out \<in> outs_\<I> \<I> \<Longrightarrow> lossless_spmf (callee s out);
     \<I> \<turnstile>g gpv \<surd>; \<And>s. \<I> \<turnstile>c callee s \<surd> \<rbrakk>
  \<Longrightarrow> lossless_spmf (exec_gpv callee gpv s)"
by(rule oi_True.lossless_exec_gpv; simp)

lemma (in callee_invariant) in_set_spmf_exec_gpv_into_results_gpv:
  assumes *: "(x, s') \<in> set_spmf (exec_gpv callee gpv s)"
  and WT_gpv : "\<I> \<turnstile>g gpv \<surd>"
  and WT_callee: "\<And>s. I s \<Longrightarrow> \<I> \<turnstile>c callee s \<surd>"
  and I: "I s"
  shows "x \<in> results_gpv \<I> gpv"
proof -
  have "set_spmf (exec_gpv callee gpv s) \<subseteq> results_gpv \<I> gpv \<times> UNIV"
    using WT_gpv I
  proof(induction arbitrary: gpv s rule: exec_gpv_fixp_induct)
    { case adm show ?case by(intro cont_intro ccpo_class.admissible_leI) }
    { case bottom show ?case by simp }
    case (step exec_gpv')
    { fix out c ret s'
      assume IO: "IO out c \<in> set_spmf (the_gpv gpv)"
        and ret: "(ret, s') \<in> set_spmf (callee s out)"
      from step.prems(1) IO have "out \<in> outs_\<I> \<I>" by(rule WT_gpvD)
      with WT_callee[OF \<open>I s\<close>] ret have "ret \<in> responses_\<I> \<I> out" by(rule WT_calleeD)
      with step.prems(1) IO have "\<I> \<turnstile>g c ret \<surd>" by(rule WT_gpvD)
      moreover from ret \<open>I s\<close> have "I s'" by(rule callee_invariant)
      ultimately have "set_spmf (exec_gpv' (c ret) s') \<subseteq> results_gpv \<I> (c ret) \<times> UNIV"
        by(rule step.IH)
      also have "\<dots> \<subseteq> results_gpv \<I> gpv \<times> UNIV" using IO \<open>ret \<in> _\<close>
        by(auto intro: results_gpv.IO)
      finally have "set_spmf (exec_gpv' (c ret) s') \<subseteq> results_gpv \<I> gpv \<times> UNIV" . }
    then show ?case using step.prems
      by(auto simp add: set_bind_spmf bind_UNION intro!: UN_least del: subsetI split: generat.split intro: results_gpv.Pure)
  qed
  thus "x \<in> results_gpv \<I> gpv" using * by blast+
qed

end
